//////////////////////////////////////////////////////////////////////////
// Contest.cpp
// (C) Max I. Fomitchev, 1999

#include <time.h> 
#include <stdlib.h> 
#include <stdio.h> 
#include <string.h> 
#include <windows.h>
#include "amd3dx.h"
// Uncomment if using emulation
//#include "macros.h"

// Inverse filter calculation using Toeplitz recursion
// d - The input wavelet
// w - The desired output
// f - The output inverse filter, d*f = w
// m - the length of the filters (<= 128)
void Toeplitz(float* d, float* w, float* f, int m)
{
	float a[128];	// a(n + 1)
	float a1[128];	// a(n)
	float r[128];	// Autocorrelation of d
	float g[128];	// Cross correlation of d and the desired output
					// (a spike with the lag l)
	
	// Compute g(i) and r(i)
	for ( int i = 0; i < m; i++ )
	{
		g[i] = r[i] = 0;
		for ( int k = 0; k + i < m; k++ )
		{
			g[i] += d[k]*w[k + i];
			r[i] += d[k]*d[k + i];
		}
	}

	// Initialization: n = 0
	a[0] = 1;
	f[0] = g[0]/r[0]; 
	float k, q, alfa = r[0], beta, gamma;

	// Iteration steps: n = 1, 2... N
	for ( int n = 1; n < m; n++ )
	{
		// Compute beta(n + 1), gamma(n + 1)
		beta = gamma = 0;
		for ( int j = 0; j < n; j++ )
		{
			beta += a[j]*r[n - j];
			gamma += f[j]*r[n - j];
		}

		// Compute k(n)
		k = -beta/alfa;
		// Compute alfa(n + 1)
		alfa += k*beta;
		// Compute q(n)
		q = (g[n] - gamma)/alfa;

		// Compute a(n + 1)
		for ( j = 1; j < n; j++ )
			a[j] += k*a1[n - j];
		a[n] = k*a[0];
		// Copy in reversed order!
		memcpy(a1, a, (n + 1)*sizeof(float));

		// Compute f(n + 1)
		for ( j = 0; j < n; j++ )
			f[j] += q*a[n - j];
		f[n] = q*a[0];
	}

	// Debug
	/*for ( i = 0; i < m; i++ )
		printf("f[%i] = %g\n", i, f[i]);*/
}

// Inverse filter calculation using Toeplitz recursion
// d - The input wavelet
// w - The desired output
// f - The output inverse filter, d*f = w
// m - the length of the filters (<= 128)
void Toeplitz2(float* d, float* w, float* f, int m)
{
	static float a[128] = {1};	// a(n + 1)
	static float a1[128];		// a(n)
	static float r[128];		// Autocorrelation of d
	static float g[128];		// Cross correlation d and w

	// Compute g(i) and r(i)
	for ( int i = 0; i < m - 1; i++ )
	{
		g[i] = r[m - i - 1] = 0;
		for ( int k = 0; k + i < m; k++ )
		{
			g[i] += d[k]*w[k + i];
			r[m - i - 1] += d[k]*d[k + i]; // r(i) is reversed!
		}
	}
	g[m - 1] = d[0]*w[m - 1];
	r[0] = d[0]*d[m - 1];

	// Initialization: n = 0
	f[0] = g[0]/r[m - 1];
	float alfa = r[m - 1];

	// First iteration step: n = 1
	// Compute beta(2), gamma(2)
	float beta = r[m - 2], gamma = f[0]*r[m - 2];

	// Compute k(n)
	float k = -beta/alfa;
	// Compute alfa(n + 1)
	alfa += k*beta;
	// Compute q(n)
	float q = (g[1] - gamma)/alfa;

	// Compute a(n + 1)
	a[1] = k;
	// Copy in reversed order!
	a1[1] = k;
	a1[2] = a[0];

	// Compute f(n + 1)
	f[0] += q*k;
	f[1] = q;

	// Iteration steps: n = 2...m
	for ( int n = 2; n < m; n++ )
	{
		// Compute beta(n + 1), gamma(n + 1)
		beta = gamma = 0;
		for ( int j = 0; j < n; j++ )
		{
			beta += a[j]*r[m - n + j - 1];
			gamma += f[j]*r[m - n + j - 1];
		}

		// Compute k(n)
		k = -beta/alfa;
		// Compute alfa(n + 1)
		alfa += k*beta;
		// Compute q(n)
		q = (g[n] - gamma)/alfa;

		// Compute a(n + 1)
		for ( j = 0; j < n; j++ )
			a[j] += k*a1[j];
		a[0] = 1;
		a[n] = k;
		// Copy in reversed order!
		for ( j = 0; j < n + 1; j++ )
			a1[n + 1 - j] = a[j];

		// Compute f(n + 1)
		for ( j = 0; j < n; j++ )
			f[j] += q*a1[j + 1];
		f[n] = q;
	}

	// Debug
	for ( i = 0; i < m; i++ )
		printf("f[%i] = %g\n", i, f[i]);
}

// Inverse filter calculation using Toeplitz recursion
// d - The input wavelet
// w - The desired output
// f - The output inverse filter, d*f = w
// m - the length of the filters (<= 128)
void Toeplitz3(float* d, float* w, float* f, int m)
{
	static float a[128] = {1};	// a(n + 1)
	static float a1[128];		// a(n)
	static float r[128];		// Autocorrelation of d
	static float g[128];		// Cross correlation d and w
	static float alfa, beta, gamma, k, q;

	// Compute g(i) and r(i)
	__asm  {
		femms

		xor		edx,edx				// edx = i
		mov		ecx,m				// ecx = m

		lea		eax,g				// [eax] -> g
		dec		ecx					// ecx = m - 1
M:
		mov		ebx,-2				// ebx = k
		mov		edi,d				// [edi] -> d

		mov		esi,w				// [esi] -> w

		pxor	mm2,mm2				// g[i] = 0
		pxor	mm3,mm3				// r[m - i - 1] = 0
M1:
		movq	mm0,[edi]			// mm0 = d[k]   | d[k+1]

		movq	mm1,[edi+edx]		// mm1 = d[k+i] | d[k+i+1]
		add		ebx,2				// k++

		movq	mm4,[esi+edx]		// mm4 = w[k+i] | w[k+i+1]
		pfmul	(mm1,mm0)			// mm4 = d[k]*d[k+i] | d[k+1]*d[k+i+1]

		add		esi,8				// w++
		pfmul	(mm4,mm0)			// mm4 = d[k]*w[k+i] | d[k+1]*w[k+i+1]

		add		edi,8				// d++
		pfadd	(mm2,mm1)			// mm2 = r[m - i - 1]

		pfadd	(mm3,mm4)			// mm3 = g[i]
		cmp		ebx,ecx				// k < m - i ?

		jl		M1					// Even part
		jne		SkipOdd

		// Odd part
		movd	mm0,[edi]			// mm0 = d[k]

		movd	mm1,[edi+edx]		// mm1 = d[k+i]
		
		movd	mm4,[esi+edx]		// mm4 = w[k+i]
		pfmul	(mm1,mm0)			// mm1 = d[k]*d[k+i]

		pfmul	(mm4,mm0)			// mm4 = d[k]*w[k+i]
		pfadd	(mm2,mm1)			// mm2 = r[m - i - 1]

		pfadd	(mm3,mm4)			// mm3 = g[i]
SkipOdd:
		pfacc	(mm2,mm2)			// mm2 = r[m - i - 1]
		pfacc	(mm3,mm3)			// mm3 = g[i]

		add		edx,4
		movd	[eax],mm3			// g[i] = mm3

		movd	r[ecx*4],mm2		// r[m - i - 1] = mm2
		add		eax,4

		loop	M

		// The last odd part
		mov		edi,d				// [edi] -> d

		mov		esi,w				// [esi] -> w
		movd	mm0,[edi]			// mm0 = d[0]

		movd	mm1,[edi+edx]		// mm1 = d[k+i]

		movd	mm4,[esi+edx]		// mm4 = w[k+i]
		pfmul	(mm1,mm0)			// mm1 = d[0]*d[m - 1]

		pfmul	(mm4,mm0)			// mm4 = d[0]*w[m - 1]
		movd	r,mm1				// r[0] = mm1

		movd	[eax],mm4			// g[m - 1] = mm4
		// Initialization: n = 0
		mov		ecx,m

		movd	mm0,r[ecx*4 - 4]	// mm0 = r[m - 1]

		movd	alfa,mm0			// alfa = r[m - 1]
		pfrcp	(mm1,mm0)

		movd	mm2,alfa			// mm2 = alfa
		pfrcpit1(mm0,mm1)

		mov		edx,f				// [edx] ->f
		pfrcpit2(mm0,mm1)			// mm0 = 1/r[m - 1]

		movq	mm5,mm0				// mm5 = 1/r[m - 1] = 1/alfa
		movd	mm4,g

		pfmul	(mm0,mm4)			// mm0 = g[0]/r[m - 1]

		movq	mm6,mm0				// mm6 = f[0]
		movd	[edx],mm0			// f[0] = g[0]/r[m - 1]

		// First iteration step: n = 1
		// Compute beta(2), gamma(2)
		movd	mm1,r[ecx*4 - 8]	// mm1 = r[m - 2]
		pxor	mm3,mm3

		movd	beta,mm1			// beta = r[m - 2]
		pfsub	(mm3,mm1)			// mm3 = -beta

		pfmul	(mm0,mm1)			// mm0 = f[0]*r[m - 2]

		movd	gamma,mm0			// gamma = f[0]*r[m - 2]
		// Compute k(n)
		pfmul	(mm3,mm5)			// mm3 = -beta/alfa

		movd	k,mm3				// k = -beta/alfa
		// Compute a(n + 1)
		movd	a[4],mm3			// a[1] = k

		// Copy in reversed order!
		movd	a1[4],mm3			// a1[1] = k
		movd	mm4,a				// mm4 = a[0]
		movd	a1[8],mm4			// a1[2] = a[0]

		// Compute alfa(n + 1)
		pfmul	(mm1,mm3)			// mm3 = k*beta
		pfadd	(mm2,mm1)			// mm2 = alfa + k*beta
		movd	alfa,mm2			// alfa += k*beta

		// Compute q(n)
		movd	mm1,g[4]			// mm1 = g[1]
		pfsub	(mm1,mm0)			// mm1 = g[1] - gamma
		pfrcp	(mm4,mm2)
		pfrcpit1(mm2,mm4)
		pfrcpit2(mm2,mm4)			// mm4 = 1/alfa
		pfmul	(mm1,mm2)			// mm1 = (g[1] - gamma)/alfa
		movd	q,mm1				// q = (g[1] - gamma)/alfa
		movd	[edx+4],mm1			// f[1] = q

		// Compute f(n + 1)
		pfmul	(mm1,mm3)			// mm1 = q*k
		pfadd	(mm1,mm6)			// mm1 = q*k + f[0]
		movd	[edx],mm1			// f[0] += q*k

		// Iteration steps: n = 2...m
		mov		ebx,2				// ebx = n
Iter:
		// Compute beta(n + 1), gamma(n + 1)
		mov		ecx,ebx			// ecx = j
		lea		edi,a			// edi -> a

		mov		eax,m
		pxor	mm3,mm3			// mm3 = beta

		pxor	mm4,mm4			// mm4 = gamma
		sub		eax,ecx

		mov		edx,f			// edx -> f
		add		ecx,1

		shr		ecx,1
		lea		eax,r[eax*4 - 4]// eax -> r[m - n - 1]

		sub		edx,8
M2:
		movq	mm0,[edi]		// mm0 = a[j]
		add		edx,8			// f++

		movq	mm2,[eax]		// mm2 = r[m - n + j - 1]
		add		edi,8			// a++

		movq	mm1,[edx]		// mm1 = f[j]
		pfmul	(mm0,mm2)		// mm0 = a[j]*r[m - n + j - 1]

		pfmul	(mm1,mm2)		// mm1 = f[j]*r[m - n + j - 1]
		add		eax,8			// r++

		pfadd	(mm3,mm0)		// beta += a[j]*r[m - n + j - 1]
		pfadd	(mm4,mm1)		// gamma += f[j]*r[m - n + j - 1]

		loop	M2

		pfacc	(mm3,mm3)
		pfacc	(mm4,mm4)

		movd	beta,mm3		// mm3 = beta

		movd	gamma,mm4		// mm4 = gamma
		// Compute k(n)
		movd	mm0,alfa		// mm0 = alfa

		pxor	mm1,mm1
		movq	mm5,mm0			// mm4 = alfa

		pfsub	(mm1,mm3)		// mm1 = -beta
	    pfrcp	(mm2,mm0)

		pfrcpit1(mm0,mm2)
		mov		ecx,ebx			// ecx = j

		pfrcpit2(mm0,mm2)		// mm0 = 1/alfa
		mov		eax,ebx

		pfmul	(mm1,mm0)		// mm1 = -beta1/alfa
		movd	mm2,g[eax*4]	// mm2 = g[n]

		movd	k,mm1
		// Compute alfa(n + 1)
		pfmul	(mm3,mm1)		// mm2 = k*beta

		pfadd	(mm5,mm3)		// mm4 = alfa + k*beta
		add		ecx,1

		movd	alfa,mm5		// alfa += k*beta
		// Compute q(n), a(n + 1)
		pfsub	(mm2,mm4)		// mm2 = g[n] - gamma

	    pfrcp	(mm4,mm5)
		shr		ecx,1

		pfrcpit1(mm5,mm4)
		lea		edi,a1			// edi -> a1

		pfrcpit2(mm5,mm4)		// mm4 = 1/alfa
		movd	mm1,k			// mm1 = k

		pfmul	(mm2,mm5)		// mm2 = (g[n] - gamma)/alfa
		lea		edx,a			// edx -> a

		movd	q,mm2			// q = (g[n] - gamma)/alfa
		punpckldq mm1,mm1
M3:
		movq	mm0,[edi]		// mm0 = a1[j]    | a1[j + 1]

		pfmul	(mm0,mm1)		// mm0 = k*a1[j]  | k*a1[j + 1]
		movq	mm2,[edx]		// mm2 = a[j]     | a[j + 1]

		pfadd	(mm0,mm2)		// mm0 = k*a1[j] + a[j] | q*a1[j + 1] + a[j + 1]

		movq	[edx],mm0		// a[j] += k*a1[j], a[j + 1] += k*a1[j + 1]
		add		edi,8			// a1++

		add		edx,8			// a++
		loop	M3

		movd	[edx],mm1		// a[n] = k
		mov		a,0x3f800000	// a[0] = 1

		// Copy in reversed order!
		mov		ecx,ebx
		lea		esi,a			// edi -> a

		inc		ecx
		lea		edx,a1[ecx*4]	// edx -> a1[n + 1]
Move:
		lodsd					// eax = a[j]
		mov		[edx],eax		// a1[n + 1 - j] = a[j]
		sub		edx,4
		loop	Move

		// Compute f(n + 1)
		mov		ecx,ebx			// ecx = j
		lea		edi,a1[4]		// edi -> a1[1]

		movd	mm1,q			// mm1 = q
		add		ecx,1

		mov		edx,f			// edx -> f
		shr		ecx,1

		punpckldq mm1,mm1
		sub		edx,8
M4:
		movq	mm0,[edi]		// mm0 = a1[j + 1]    | a1[j + 2]
		add		edx,8			// f++

		pfmul	(mm0,mm1)		// mm0 = q*a1[j + 1]  | q*a1[j + 2]

		pfaddm	(mm0,edx,0)		// mm0 = q*a1[j + 1] + f[j] | q*a1[j + 2] + f[j + 1]

		movq	[edx],mm0		// f[j] += q*a1[j + 1], f[j + 1] += q*a1[j + 2]
		add		edi,8			// a1++

		loop	M4

		movd	[edx+8],mm1		// f[n] = q
		inc		ebx

		cmp		ebx,m
		jl		Iter

		emms
	}
	// Debug
	for ( int i = 0; i < m; i++ )
		printf("f[%i] = %g\n", i, f[i]);
}

///////////////////////////////////////////////////////////////////////
#define N 1
#define M 10
void main()
{
	static float d[128], w[128], f[128];
	clock_t t;

	srand(100);
	for ( int i = 0; i < M; i++ )
	{
		d[i] = ((float)rand())/RAND_MAX*10;
		w[i] = ((float)rand())/RAND_MAX*10;
	}

	t = clock();
	for ( i = 0; i < N; i++ )
	{
		memset(f, 0, sizeof(f));
		Toeplitz2(d, w, f, M);
	}
	t = clock() - t;
	printf("%g seconds\n", float(t)/CLOCKS_PER_SEC/N);

	// Uncomment if using emulation
	//_InstallAMD3D();
	t = clock();
	for ( i = 0; i < N; i++ )
	{
		memset(f, 0, sizeof(f));
		Toeplitz3(d, w, f, M);
	}
	t = clock() - t;
	printf("%g seconds\n", float(t)/CLOCKS_PER_SEC/N);
	// Uncomment if using emulation
	//_AMD3DHandler();
}
