/*******************************************************************************
 *
 * (c) Copyright Laurence Marrie 1999
 * All Rights Reserved
 *
 * Purpose: Node class for asl_set
 *
 * Notes:	(1) The capacity of the skip pointer array is not stored since it
 *				can be inferred from height_
 *			(2) The level of the first skip pointer is 0.
 *				Successor and predecessor pointers are both assigned a level 
 *				of -1. The height of a node with 0 skip pointers is 0.
 *			(3) A wrapper class is used to allocate the END node, so that T 
 *				does not need a default constructor.
 *
 ******************************************************************************/

#ifndef ASL_NODE_H
#define ASL_NODE_H

template <class T>
class asl_node
{
public:

	// used for all nodes except the END node
	asl_node( T const & key )
		: height_(0), right_(0), left_(0), next_(0), key_(key) {}

	~asl_node() 						{ delete [] next_; }

	T const & 			key() const		{ return key_; }
	unsigned char	 	height() const	{ return height_; }

	// return successor node
	asl_node * 			right()			{ return right_; }
	asl_node const * 	right() const	{ return right_; }

	// return predecessor node
	asl_node * 			left() 			{ return left_; }
	asl_node const * 	left() const	{ return left_; }

	// return successor node at level i, where -1 <= i < height_
	asl_node * 			next( int i )
									{ return ( i == -1 ) ? right_ : next_[i]; }
	asl_node const *	next( int i ) const
									{ return ( i == -1 ) ? right_ : next_[i]; }

	// "skip pointer" only version of next (exists only for efficiency); 
	// return successor node at level i, where 0 <= i < height_
	asl_node * 			next_sp( int i ) 		{ return next_[i]; }
	asl_node const * 	next_sp( int i ) const 	{ return next_[i]; }

	// return a pointer to the predecessor node at level i, 
	// where -1 <= i < height_
	asl_node const * prev( int i ) const
	{
		if ( i == -1 )
		{
			return left_;
		}
		else
		{
			asl_node const * temp = next(i-1);
			// The loop executes at most 3 times, since in the worst case
			// there are 3 nodes all of the same height in a row.
			while ( temp->height_ < i+1 )		
				temp = temp->next(i-1);

			return temp;
		}
	}

	// return a pointer to the predecessor node at level i, 
	// where -1 <= i < height_
	asl_node * prev( int i )
				{	return const_cast<asl_node*>( 	// just call const version
						static_cast<asl_node const*>( this )->prev(i) ); }

	void setRight( asl_node<T> * n )		{ right_ 	= n; }
	void setLeft( asl_node<T> * n )			{ left_ 	= n; }
	void setNext( int i, asl_node<T> * n )	{ next_[i]	= n; }

	// change the current height, without resizing skip array
	void raiseWithoutResizing( asl_node<T> * n ) 
											{ next_[height_] = n; height_++; }
	void lowerWithoutResizing()				{ height_--; }

	// increment the current height, releasing the current skip array and 
	// replacing it with 'next'
	asl_node** 
	raiseWithResizing( asl_node** next, asl_node<T> * n )
	{	
		for ( int i = 0; i < height_; i++ )
			next[i] = next_[i];
			
		next[height_] = n;
		height_++;

		asl_node ** temp 	= next_;
		next_ 				= next;

		return temp;
	}

	// decrement the current height, releasing the current skip array and 
	// replacing it with 'next'
	asl_node** 
	lowerWithResizing( asl_node** next )
	{
		height_--;
		for ( int i = 0; i < height_; i++ )
			next[i] = next_[i];

		asl_node ** temp 	= next_;
		next_ 				= next;
		return temp;
	}

	// A specialized interchange method.
	// Requires that this.height() > 0 && this.next_ != 0, and 
	// node.height() == 0 && node.next_ == 0
	//
	void 
	swapSkipArrays( asl_node<T> & node )
	{
		// update predecessor ptrs
		for ( int i = height_-1; i >= 0; i-- )
			prev(i)->setNext( i, &node );

		node.height_ 	= height_;
		node.next_		= next_;

		height_		= 0; 	// not necessary; node is due to be deleted
		next_		= 0;	// necessary (so don't delete in dtor)
	}

	friend class asl_end_node<T>;

private:

	asl_node *		right_;		// successor node
	asl_node *		left_;		// predecessor node
	asl_node**		next_;		// array of skip pointers; 0 if height is 0

	T 				key_;

	unsigned char 	height_;	// how many skip pointers are in use?
								// (unsigned char saves a few bytes over int)
};

template <class T>
class asl_end_node_base
{
public:
	~asl_end_node_base() { operator delete( end_ );	}

protected:
	asl_end_node_base() 
		:	end_( static_cast<asl_node<T>*>(
					operator new( sizeof(asl_node<T>) ) ) )	{}

	asl_node<T> * end_;
};

template <class T>
class asl_end_node : public asl_end_node_base<T>
{
public:
	asl_end_node()
	{
		end_->height_ 	= 1;
		end_->next_ 	= new asl_node<T>* [ end_->height_ ];

		// initialize pointers
		end_->next_[0] 	= end_;
		end_->right_ 	= end_;
		end_->left_		= end_;
	}

	~asl_end_node()
	{
		// release memory for all nodes in set
		//
		asl_node<T> const * n = end_->right();
		while ( n != end_ )
		{
			n = n->right();
			delete n->left();
		}

		delete [] end_->next_; 
	}

	void swap( asl_end_node & n )			{ ::swap( end_, n.end_ ); }

	// operators for treating this class like a pointer
	//
	operator asl_node<T> const * () const	{ return end_; }

	asl_node<T> *		operator->()		{ return end_; }
	asl_node<T> const *	operator->() const	{ return end_; }

	asl_node<T> &		operator*()			{ return *end_; }
	asl_node<T> const &	operator*() const	{ return *end_; }

private:
	asl_end_node( asl_end_node const & );
	asl_end_node & operator=( asl_end_node const & );
};

#endif
