/*******************************************************************************
 *
 * (c) Copyright Laurence Marrie 1999
 * All Rights Reserved
 *
 * Purpose: Alternating Skip List implementation of a dynamic set.
 *
 * Notes:	The element type, T, must:
 *				(a)	Have a copy constructor
 *				(b) Support comparison using the '<' operator
 *				(c) Have a non-throwing destructor
 ******************************************************************************/

#ifndef ASL_SET_H
#define ASL_SET_H

#include "asl_iterator.h"
#include "asl_free_list.h"

template <class T>
class asl_set
{
public:
	typedef asl_iterator<T> iterator;

	asl_set();
	asl_set( asl_set const & asl );

	// number of entries
	size_t 		size() const			{ return size_; }

	// support forward iteration
	iterator	begin() const			{ return end_->right(); }
	iterator	end() const				{ return end_; }

	// support reverse iteration
	iterator	rbegin() const			{ return end_->left(); }
	iterator	rend() const			{ return end_; }

	// search; return end() if key not found
	iterator 	find( T const & key ) const;

	// search using a "hint"; return end() if key not found
	iterator	find( iterator const & it, T const & key ) const;

	// add key; return success or failure
	bool 		insert( T const & key );

	// add key, search using a "hint"; return success or failure
	bool		insert( iterator const & it, T const & key );

	// remove key
	void 		erase( T const & key );
	void 		erase( iterator it );

	// exchange the contents of two sets
	void		swap( asl_set & asl );

	asl_set & 	operator=( asl_set const & asl );

protected:

	// Top-down search procedure starting at the given node and level.
	//
	asl_node<T> const * 
	find_helper( int level, asl_node<T> const * node, T const & key ) const
	{
		// Note that "node->height() == level+1" implicitly tests that
		// "node != end_". The latter test can be used instead, but the former 
		// also reduces key comparisons.
		//
		while ( --level != -1 )
		{
			node = node->next_sp( level );

			// move right
			if ( node != end_ && node->key() < key )
			{
				node = node->next_sp( level );
				if ( node->height() == level+1 && node->key() < key )
					node = node->next_sp( level );
			}
			
			--level;
			node = node->next_sp( level );

			// move left
			if ( node != end_ && !(node->key() < key) )
			{
				node = node->next_sp( level );
				if ( node->height() == level+1 && !(node->key() < key) )
					node = node->next_sp( level );
			}
		}

		node = node->right();
		if ( node != end_ && node->key() < key )
		{
			node = node->right();
			if ( node->height() == level+1 && node->key() < key )
				node = node->right();
		}

		return node;
	}

	asl_node<T> const *
	find_with_hint_helper( iterator const & it, T const & key ) const
	{
		asl_node<T> const * hint = it.node_;
		int level;

		if ( hint != end_ )
		{
			if ( hint->key() < key )
			{
				// move right, climbing levels in steps of 2 until it is no longer 
				// possible to stay left of the key position

				level = ( hint->height() > 1 ) ? 2 : 0;
				asl_node<T> const * right = hint->next(level-1);
			
				while ( right != end_ && right->key() < key )
				{
					hint = right;
					if ( level + 2 <= hint->height() )
						level += 2;

					right = ( level == 0 ) 
								? hint->right() : hint->next_sp(level-1);
				}
			}
			else if ( key < hint->key() )
			{
				// move left, climbing levels in steps of 2 until left of the 
				// key position

				level = 0;
				while ( hint != end_ && !(hint->key() < key) )
				{
					hint = ( level == 0 ) ? hint->left() : hint->next_sp(level);
					if ( level + 2 < hint->height() )
						level += 2;
				}
			}
		}
		else
		{
			// caller passed end() instead of a valid iterator; use regular
			// top-down search in this case
			level = end_->height() - 1;
		}

		// if the hint node contains the search key just return, else
		// search top-down
		return ( hint->key() > key || hint->key() < key )
					? find_helper( level, hint, key ) : hint;
	}

	bool 
	insert_helper( asl_node<T> const * node, T const & key )
	{
		if ( node == end_ || key < node->key() )
		{
			// Perform all allocations first. If an exception is thrown 
			// the structure will remain gap invariant.

			list_.cleanupBeforeRaising();

			// Create new node and link it in between 'first' and
			// 'third'.
			asl_node<T> * first		= const_cast<asl_node<T>*>(node)->left();
			asl_node<T> * second 	= new asl_node<T>( key );
			asl_node<T> * third		= const_cast<asl_node<T>*>( node );

			second->setRight( third );
			second->setLeft( first );
			first->setRight( second );
			third->setLeft( second );

			size_++;

			int level = -1;
			int nextLevel;

			do
			{
				nextLevel = level + 1;

				// If we now have 3 in a row, raise the middle node	
				if ( third->height() == first->height() )
				{
					if ( first != third )
					{
						second 	= first->next(level);
						first 	= third->next(level);
						third	= first->next_sp(nextLevel);

						// raise and link in
						list_.raiseNode( *second, third );
						first->setNext( nextLevel, second );
						level = nextLevel;
					}
				}
				else if ( third->height() > first->height() )
				{
					if ( third->next_sp(nextLevel)->next(level) != first )
					{
						first	= third;
						second 	= third->next_sp(nextLevel)
											->next(level)->next(level);
						third	= first->next_sp(nextLevel);

						// raise and link in
						list_.raiseNode( *second, third );
						first->setNext( nextLevel, second );
						level = nextLevel;
					}
				}
				else // third->height() < first->height()
				{
					if ( third->next(level)->height() == third->height() )
					{
						first	= third->next(level)->next(level);
						second 	= third;
						third	= first->next_sp(nextLevel);
						
						// raise and link in
						list_.raiseNode( *second, third );
						first->setNext( nextLevel, second );
						level = nextLevel;
					}
				}
			}
			while ( level == nextLevel );

			// Raise the end node if necessary
			list_.updateEndNodeAfterRaising( *end_ );

			return true;
		}

		return false;
	}

	asl_free_list<T>	list_;		// "free list" of skip pointer arrays
	asl_end_node<T>		end_;		// the END node
	size_t				size_;		// number of nodes in the set
};

#ifndef ASL_LINK_TIME_INSTANTIATION
#	include "asl_set.cpp"
#endif

#endif
