/*******************************************************************************
 *
 * (c) Copyright Laurence Marrie 1999
 * All Rights Reserved
 *
 * Purpose:	Maintain a "free list" of skip pointer arrays, in increasing order
 *			of size. This serves to (a) reduce the number of calls to the 
 *			memory manager and (b) provide a mechanism for the insertion
 *			algorithm to handle memory allocation failures (c) raise and lower
 *			node heights.
 *
 * Notes:	(1)	The implemention actually uses an array rather than a list.
 *			(2) The size of each skip pointer array is stored as well;
 *				this simplifies testing for whether a skip pointer array needs 
 *				resizing.
 *			(3) The first entry always has size 0 and 0 skip array pointer.
 *
 ******************************************************************************/

#ifndef ASL_FREE_LIST_H
#define ASL_FREE_LIST_H

#include "asl_node.h"

template <class T>
struct asl_free_skip_array
{
	unsigned char 	size_;
	asl_node<T>**	next_;
};

template <class T>
class asl_free_list
{
public:
	asl_free_list();
	~asl_free_list();

	void swap( asl_free_list & afl );
	
	/*
	 * Methods called by the insertion procedure.
	 */

	// Called before adding a new node. Pre-allocates as much memory as needed 
	// (in the worst case) to restore gap invariance.
	void cleanupBeforeRaising();

	// Increase the node height by 1 (resize skip pointer array if necessary)
	void 
	raiseNode( asl_node<T> & node, asl_node<T> * n )
	{
		if ( node.height() == arrays_[raiseIndex_].size_ )
		{
			arrays_[raiseIndex_].next_ 
				= node.raiseWithResizing( arrays_[raiseIndex_+1].next_, n );
			raiseIndex_++;
		}
		else
			node.raiseWithoutResizing( n );
	}

	// Increase the height of the END node by 2 if necessary
	void 
	updateEndNodeAfterRaising( asl_node<T> & endNode )
	{
		if ( endNode.next_sp( endNode.height()-1 ) != &endNode )
		{
			if ( endNode.height() == arrays_[raiseIndex_].size_ )
			{
				arrays_[raiseIndex_].next_ 
					= endNode.raiseWithResizing( arrays_[raiseIndex_+1].next_, 
												 &endNode );
				raiseIndex_++;
			}
			else
				endNode.raiseWithoutResizing( &endNode );

			if ( endNode.height() == arrays_[raiseIndex_].size_ )
			{
				arrays_[raiseIndex_].next_ 
					= endNode.raiseWithResizing( arrays_[raiseIndex_+1].next_, 
												 &endNode );				
				raiseIndex_++;
			}
			else
				endNode.raiseWithoutResizing( &endNode );
		}
	}

	/*
	 * Methods called by the deletion procedure.
	 */

	// Decrease the node height by 1 (resize skip pointer array if necessary)
	void 
	lowerNode( asl_node<T> & node )
	{	
		if ( node.height()-1 == arrays_[lowerIndex_].size_ )
		{
			lastReleased_ = node.lowerWithResizing( lastReleased_ );
			lowerIndex_++;
		}
		else
			node.lowerWithoutResizing();
	}
	
	// raise using the skip pointer array released by the last node lowered
	void 
	raiseAfterLowering( asl_node<T> & node, asl_node<T> * n )
	{
		lastReleased_ = node.raiseWithResizing( lastReleased_, n );
		lowerIndex_--;
	}

	// Decrease the height of the END node by 2 if necessary
	void 
	updateEndNodeAfterLowering( asl_node<T> & endNode )
	{
		if ( endNode.height() > 2 && 
			 endNode.next_sp( endNode.height()-3 ) == &endNode )
		{
			endNode.lowerWithoutResizing();

			if ( endNode.height()-1 == arrays_[lowerIndex_].size_ )
			{
				lastReleased_ = endNode.lowerWithResizing( lastReleased_ );
				lowerIndex_++;
			}
			else
				endNode.lowerWithoutResizing();
		}
	}

	// Release the last skip array discarded. Called when gap invariance 
	// has been fully restored after removing a node.
	void cleanupAfterLowering();

private:

	asl_free_list( asl_free_list const & );
	asl_free_list & operator=( asl_free_list const & );

	// number of skip pointer arrays currently stored in free list
	unsigned 	entries_;

	// free list index used during insertion; if raiseIndex_ > 0,
	// freeList_[raiseIndex_].next_ is the last array removed
	unsigned 	raiseIndex_;

	// The free list index used during deletion; if lowerIndex_ > 0,
	// freeList_[lowerIndex_-1].size_ is the physical size of lastReleased_
	unsigned 		lowerIndex_;
	asl_node<T>** 	lastReleased_;	// the last skip pointer array released

	// The "free list" array.
	//
	// Node heights are stored using unsigned char, which can range from 0-255.
	// Arrays_ therefore needs to hold 9 items at most -- the corresponding 
	// array sizes are 0,1,2,4,8,16,32,64 and 128. This allows for at least 
	// 2^62 nodes (since the END node is at most 2 higher than the highest 
	// ordinary node, and the largest array in the free list always has twice 
	// the capacity of the END node array).
	//
	enum { FREE_LIST_SIZE = 9 };
	asl_free_skip_array<T> arrays_[FREE_LIST_SIZE];
};

#ifndef ASL_LINK_TIME_INSTANTIATION
#	include "asl_free_list.cpp"
#endif

#endif
