/*
** @#START#@
**
** Cage Controller Openbios boot code
**
**	(C) Copyright International Business Machines Corporation 1997,1998
**	Licensed Material  -  Program Property of I B M 
**	Refer to copyright instructions: Form G120-2083
**
**  Module:
**	enetLib.h
**
** Description:
**	I/O addresses and Bit Field definitions for the Ethernet LANTIC 83905.
**
** Functions:
**
** Dependencies:
**
** Context:
**
** Restrictions:
**
** Environment:
**	OS Open 1.6.3 (ELF) and PowerPC 403GCX JDS-Box
**
** Linkage:
**	AIX 4.2.1 with MetaWare High C/C++ 2.6u
**
** Entry points:
**
** External references:
**
** @author
**	Thomas Richter
**
** @version
**	$KW; $-FN; $-Ver; $-ChkD;
**
** History:
** Date      Author      Description
** --------------------------------------------------------------------------
** 09-Sep-97 Naegele     Created (based on Matt Tyrlik's version)
** 16-Mar-98 Richter     Modified to handle multiple devices per driver
**
** @#END#@*/

#ifndef enet_h
#define enet_h

#include <time.h>
#include <netinet/if_ether.h>
#include <sys/mbuf.h>

#define PAGE0		0x00
#define PAGE1		0x40
#define PAGE2		0x80

#define PBMASK		0xC0
#define ACC_MASK	0x01

#define PACKET_SIZE     256
#define RAM64k
#ifdef RAM64K
#define PROM_SIZE	0x100		/* PROM area 0000-0100H, FF00-FFFF */
#define RAM_SIZE	0xFE00		/* 64KB-0x200 = 254 Buffer a 256 Byte */
#define XMIT_BUFS       8		/* 8 buffers a 256 bytes for xmit */
#define RCV_BUFS        246		/* 246 buffers a 256 bytes for rcv */
#else
#define PROM_SIZE	0x4000		/* PROM area 0000-3FFFH */
#define RAM_SIZE	0x2000		/* 8KB Ram = 32 Buffers a 256 Bytes */
#define XMIT_BUFS       8		/* 8 buffers a 256 bytes for xmit */
#define RCV_BUFS        24		/* 24 buffers a 256 bytes for rcv */
#endif

/*
**  Mode Configuration Register A (MCR_A) / Read-Write
**  Note: MCR_A can only be written by a read to 0x0A and a following write
**        to 0x0A. A write to 0x0A w/o a previous read will access RBCR0.
**  D0-D2, IOAD0-IOAD2: I/O Address within system's I/O map.
**                      '0 0 0'  0x300    Controller 0
**                      '1 1 1'  0x360    Controller 1
**  D3-D5, INT0-INT2:   Interrupt Line used
**                      Pre-Req: Bit 5='1' in Hardware Config. Reg. C
**                      '0 0 0' INT0 pin active in Direct Mode
**  D6, FREAD:          Fast Read ('0')
**  D7, MEMIO:          Memory or I/O Mode
**                      '0' I/O Mode
*/
#define MCR_A           0x0A		/* configure controller (r/w) */
#define MEMIO		0x80		/* shared memory select */
#define FREAD		0x40		/* fast read select */
#define INT3		0x28		/* INT3 signal */
#define INT2		0x10		/* INT2 signal */
#define INT1		0x08		/* INT1 signal */
#define INT0		0x00		/* INT0 signal */
#define IOAD2		0x04		/* base I/O address */
#define IOAD1		0x02		/* base I/O address */
#define IOAD0		0x01		/* base I/O address */
#define IOAD300		0x00		/* base I/O address at 0x300 */
#define IOAD360		0x07		/* base I/O address at 0x360 */

/*
**  Mode Configuration Register B (MCR_B) / Read-Write
**  Note: MCR_A can only be written by a read to 0x0B and a following write
**        to 0x0B. A write to 0x0B w/o a previous read will access RBCR1.
**  D0-D1, PHYS0-PHYS1: Physical Layer Interface (w)
**                      '0 0'  Twisted pair (10BASE-T)
**  D2     GDLNK:       Good Link (r/w)
**                      '0/1' write: Link test enabled/disabled
**                      '0/1' read:  Link status bad(TPI broken)/good
**  D3     IO16CON:     IO16 Control (w)
**                      '1' generates IO16 signal after IORD or IOWR active
**                      '0' generates IO16 only on address decode
**  D4     CHRDY:       CHRDY from IORD or IOWR or from BALE (w)
**                      '0' CHRDY active after command strobe
**                      '1' CHRDY active after BALE active
**  D5     BE           Bus Error (r/w)
**                      '1/0' Read:  ERROR/NO ERROR.
**                            ERROR: LANTIC inserts wait states but system
**                                   terminates cycle w/o inserting wait states
**                      '1'   Write: Clears error bit to '0'
**  D6     BPWR         Boot PROM Write.
**                      '0' No write cycles are generated to boot PROM.
**  D7     EELOAD       EEPROM Load.
**                      '1' Enables EEPROM load algorithm
*/
#define MCR_B           0x0B	/* configure controller (r/w) */
#define EELOAD		0x80	/* EEPROM load */
#define BPWR		0x40	/* Boot PROM Write */
#define BE		0x20	/* Bus Error ('1' clears) */
#define CHRDY		0x10	/* CHARDY signal */
#define IO16CON		0x08	/* IO16 Control signal */
#define GDLNK		0x04	/* Link checking and link status */
#define PHYS1		0x02	/* Physical layer interface bit */
#define PHYS0		0x01	/* Physical layer interface bit */
#define PHYS_TPI	0x00	/* Twisted pair(10BASE-T) select */

/*
**  Command Register (CR) / Read-Write
**  D0     STP:         STOP (w)
**                      '1'  Software reset (powers up high)
**                      '0'  Exit reset (STA='1')
**  D1     STA  :       Start (w)
**                      '1'  Start NIC core (powers up low)
**  D2     TXP:         Transmit Packet (w)
**                      '1' initiate transmission of a packet
**                      Internally reset after good xmit or abort
**  D3-D5  RD0-RD2:     Remote DMA Command (w)
**                      '1 0 0' Remote read
**                      '0 1 0' Remote write
**                      '0 0 1' Abort Remote DMA (see Note 1)
**  D6-D7  PS0-PS10     Page Select (w)
**                      '0 0' Register Page 0
**                      '1 0' Register Page 1
**                      '0 1' Register Page 2
**
**  Note 1:  Read or write NIC port for a remote read/write
*/
#define CR              0x00	/* command register (r/w) */
#define PS1		0x80	/* page select bit 1 */
#define PS0		0x40	/* page select bit 0 */
#define RD2		0x20	/* remote DMA Command bit 2 */
#define RD1		0x10	/* remote DMA Command bit 1 */
				/* Physical layer interface bit */
#define PHYS_TPI	0x00	/* Twisted pair(10BASE-T) select */
#define RD0		0x08	/* remote DMA Command bit 0 */
#define TXP		0x04	/* transmit Packet */
#define STA		0x02	/* start NIC core */
#define STP		0x01	/* stop NIC core */

#define ISR             0x07	/* interrupt status (r/w) */
#define RST		0x80	/* reset status */
#define RDC		0x40	/* remote DMA complete */
#define CNT		0x20	/* counter overflow */
#define OVW		0x10	/* overwrite warning */
#define TXE		0x08	/* transmit error */
#define RXE		0x04	/* receive error */
#define PTX		0x02	/* packet transmitted */
#define PRX		0x01	/* packet received */

#define IMR             0x0F	/* interrupt mask (w) */
#define RDCE		0x40	/* remote DMA complete enable */
#define CNTE		0x20	/* counter overflow enable */
#define OVWE		0x10	/* overwrite warning enable */
#define TXEE		0x08	/* transmit error enable */
#define RXEE		0x04	/* recevie error enable */
#define PTXE		0x02	/* packet transmitted enable */
#define PRXE		0x01	/* packet received enable */

#define DCR             0x0E	/* data configuration register (w) */
#define FT1		0x40	/* fifo threshold select bit 1 */
#define FT0		0x20	/* fifo threshold select bit 0 */
#define FT2B		0x00	/* two bytes */
#define FT4B		0x20	/* 4 bytes */
#define FT8B		0x40	/* 8 bytes */
#define FT12B		0x60	/* 12 bytes */
#define ARM		0x10	/* auto initialize remote */
#define LS		0x08	/* loopback select */
#define LAS		0x04	/* long Address select */
#define BOS		0x02	/* byte order select */
#define WTS		0x01	/* word DMA transfers */

#define TCR             0x0D	/* transmit configuration reg (w) */
#define OFST		0x10	/* collision offset enable */
#define ATD		0x08	/* auto transmit disable */
#define LB1		0x04	/* loopback control 1 */
#define LB0		0x02	/* loopback control 0 */
#define LBM0		0x00	/* normal mode */
#define LBM1		0x02	/* internal NIC loopback */
#define LBM2		0x04	/* internal ENDEC loopback */
#define LBM3		0x06	/* external loopback */
#define CRC		0x01	/* cRC inhibited */

#define TSR             0x04	/* transmit status register (r) */
#define OWC		0x80	/* out of window collision */
#define CDH		0x40	/* collision detect heartbeat */
#define FU		0x20	/* fIFO underrun */
#define CRS		0x10	/* carrier Sense lost */
#define ABT		0x08	/* transmit ended */
#define COL		0x04	/* transmit collided */
#define PTXS		0x01	/* transmit without error */

#define RCR             0x0C	/* receive config reg (w) */
#define MMO		0x20	/* monitor mode */
#define PRO		0x10	/* promiscuous addressing */
#define AM		0x08	/* accept multicast */
#define AB		0x04	/* accept broadcast */
#define AR		0x02	/* accept runt packets */
#define SEP		0x01	/* save errored packets */

#define RSR             0x0C	/* receive status register (r) */
#define DFR		0x80	/* collision */
#define DIS		0x40	/* receiver disabled */
#define PHY		0x20	/* physical Address match */
#define MPA		0x10	/* missed packet */
#define FO		0x08	/* fIFO Overrun */
#define FAE		0x04	/* frame alignment error */
#define CRCE		0x02	/* cRC error */
#define PRXS		0x01	/* packet received intact */

/*
** Other reg definitions.
*/
#define TPSR            0x04	/* transmit page start (w, 2) */
#define TBCR1           0x06	/* high order byte count (w, 0) */
#define TBCR0           0x05	/* low order byte count (w, 0) */
#define PSTART          0x01	/* receive page start (w, 0) */
#define PSTOP           0x02	/* receive page stop (w, 0) */
#define BNRY            0x03	/* boundary register (r/w, 0) */
#define CLDA1           0x02	/* current local DMA reg 1 (w, 2) */
#define CLDA0           0x01	/* current local DMA reg 0 (w, 2) */
#define RSAR1           0x09	/* remote start address 1 (w, 0) */
#define RSAR0           0x08	/* remote start address 0 (w, 0) */
#define RBCR1           0x0B	/* remote byte count reg 1 (w, 0) */
#define RBCR0           0x0A	/* remote byte count reg 0 (w, 0) */
#define CRDA1           0x09	/* remote DMA address reg 1 (r, 0) */
#define CRDA0           0x08	/* remote DMA address reg 0 (r, 0) */
#define NCR             0x05	/* number of collisions (r, 0) */
#define CNTR0           0x0D	/* frame alignment error count (r, 0) */
#define CNTR1           0x0E	/* cRC errors (r, 0) */
#define CNTR2           0x0F	/* missed packets (r, 0) */
#define FIFO            0x06	/* fIFO register (r, 0) */

#define PAR0            0x01	/* physical address regs (r/w, 1) */
#define PAR1            0x02
#define PAR2            0x03
#define PAR3            0x04
#define PAR4            0x05
#define PAR5            0x06
#define CURR            0x07	/* current page register (r/w, 1) */
#define MAR0            0x08	/* multicast address regs (r/w, 1) */
#define MAR1            0x09
#define MAR2            0x0A
#define MAR3            0x0B
#define MAR4            0x0C
#define MAR5            0x0D
#define MAR6            0x0E
#define MAR7            0x0F

#define	ENET_MAX_MTU	1518	/* Max. length of ethernet packet */
#define	ENET_HWADDRLEN	6	/* Length of MAC address in bytes */

/*
** Function prototype.
*/
int boardenet_attach(int);

#endif
