/*
** @#START#@
**
**	Cage Controller OS Open library code
**
**	(C) Copyright International Business Machines Corporation 1999
**	Licensed Material  -  Program Property of I B M 
**	Refer to copyright instructions: Form G120-2083
**
** Module:
**	enetlib.c
**
** Description:
**	Native Ethernet device driver for NS83905. Supports
**	multiple devices.
**
** Functions:
**
** Dependencies:
**
** Context:
**
** Restrictions:
**
** Environment:
**	OS Open 1.6.3 and JDS-Box
**
** Linkage:
**	AIX 4.2.1 with CSet++ 3.1.4.7
**
** Entry points:
**	boardenet_attach
**
** External references:
**
** @author
**	Thomas Richter
**
** History:
** Date      Author      Description
** --------------------------------------------------------------------------
** 16-Mar-98 Richter     Modified (copied from enetLib.c 5.7 9/3/97 08:01:02 
** 23-Apr-99 Richter     Fixed msr problem, fixed alignment problem.
** 23-Aug-99 Richter     Added dual ethernet failover code
*/

/*-----------------------------------------------------------------------------+
|
|  Author:      Maciej Tyrlik.
|
|  Change Activity-
|
|  Date        Description of Change                                       BY
|  ---------   ---------------------                                       ---
|  22-Mar-96   Created                                                     MPT
|  27-Mar-96   Realized that mbufs will be freed twice in some situation   MPT
|  25-Apr-96   Fixed code blockina signal delivery during send packet	   MPT
|  03-Sep-97   Written read_port()/write_port()/write_cr() in assembler    MPT
|  XX-XXX-XX   1.00 first version after code review                        MPT
|
+-----------------------------------------------------------------------------*/

#include <machine/machparam.h>
#include <machine/endian.h>
#include <sys/types.h>
#include <ioLib.h>
#include <flih.h>
#include <pthread.h>
#include <ppcLib.h>
#include <types.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <semaphore.h>
#include <enetLib.h>
#include <sys/devLib.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/tcpipLib.h>
#include <sys/ioctl.h>
#include <sys/ksyscall.h>
#include <net/if.h>
#include <net/if_types.h>
#include <net/netisr.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>

#include <cccfg.h>
#include "enetswap.h"

/*
** Defines.
*/
#define	DIM(x)			(sizeof (x) / sizeof ((x)[0]))
#define ENET_MAC_LENGTH		(ETHERMTU + sizeof(struct ether_header))
#define ENET_MAC_MIN_LENGTH	64
#define ENET_THREAD_PRIO	(PTHREAD_PRIO_MAX_NP-5)
#define ENET_RCV_STACKSIZE	4096
#define ENET_FLIH_SIZE		4096
#define	ENET_DEV		2	/* # of devices */
#define	NUMRECV_BUFF		2	/* # of receive buffers */

/*
** Device driver data needed to operate properly prior to initialization.
** I/O register base for LANTIC 83905 Chip 0 and 1.
** Definitions for Interrupt settings.
*/
#define NIC_BASE0	0xF4000300	/* Base address for chip 0 */
#define NIC_PORT0	0xF4000310	/* Port address for chip 0 */
#define NIC_BASE1	0xF4000360
#define NIC_PORT1	0xF4000370
#define	ENET_INT0	EVENT_EXT0_NP	/* Interrupt for chip 0 */
#define	ENET_INT1	EVENT_EXT1_NP	/* Interrupt for chip 1 */
#define	ENET_INT0_CFG	EXTINT_POS_ACTIVE|EXTINT_LEVEL
					/* Positive level triggered */
#define	ENET_INT1_CFG   EXTINT_POS_ACTIVE|EXTINT_LEVEL
					/* Positive level triggered */

#define	WRITE_CR(p, value)	outbyte((unsigned char *)((p)->base + CR), value)
#define	WRITE_PORT(p, value)	outbyte((unsigned char *)(p)->port, value)
#define	READ_PORT(p)		inbyte((unsigned char *)(p)->port)

/*
** Global data for this library
*/
char *enetlib_id = "$KW; $-FN; $-Ver; $-ChkD; $-Own; $EKW;";
/* Start Teamconnect keyword expansion: $KW; */
static char *libname = "@(#) enetLib.a $-Ver; $-ChkD;";
/* Stop Teamconnect keyword expansion: $EKW; */

static struct {
    struct mbuf *ifq_head;
    struct mbuf *ifq_tail;
    sem_t enet_sem;
} my_ifq = {
     NULL, NULL
};

struct t_ioframe {			/* Send/Receive buffer structure */
    unsigned long filler;		/* Force word alignment */
    unsigned short len;			/* Length in frame (bytes) */
    unsigned char data[ENET_MAX_MTU];	/* Data from hardware ring buffers */
};

struct t_edd {				/* Ethernet device driver structure */
    unsigned long base,			/* Base address for dd instance */
		  port,			/* Port address for dd instance */
		  intr,			/* Interrupt event */
		  intr_cfg,		/* Trigger/Action level for intr. */
		  ioaddr;		/* Base I/O Address for controller */
    unsigned long _PStart_Addr,		/* Start addr of RCV buf ring */
                  _PStop_Addr;		/* Stop addr of RCV buf ring */
    unsigned char _PStart_HAddr,	/* Start page of RCV buf ring */
                  _PStop_HAddr,		/* Stop page of RCV buf ring */
                  _TPStart_HAddr,	/* Start page of Xmit buf */
                  next_packet_ptr;	/* Next packet position */
    struct t_ioframe outframe;		/* Output buffer */
    struct t_ioframe inframe[NUMRECV_BUFF];	/* Input buffers */
    struct in_addr ip_addr;		/* IP address */
    int flih_bufnum,
	read_bufnum;			/* Pointer to correct recv buffers */
    unsigned char flih_stack[ENET_FLIH_SIZE];		/* Stack pointer */
    struct arpcom arpcom;		/* ARP & Interface data */
    sem_t dev_access,			/* Access to device */
	  xmit_sem;			/* Transmitting possible */
    char linkok;			/* True if TP link bad */
}edd[ENET_DEV] = {
    { NIC_BASE0, NIC_PORT0, ENET_INT0, ENET_INT0_CFG, IOAD300 },
    { NIC_BASE1, NIC_PORT1, ENET_INT1, ENET_INT1_CFG, IOAD360 }
};


/*
** Device number of chip to use
*/
int enet_chip = -1;		/* Must be zero or one */

void net_setsoftnet();		/* Missing prototype in tcpipLib.a */


static unsigned char read_nic(unsigned long base, unsigned long reg_addr, unsigned char page)
{
    unsigned char cr;

    cr = inbyte((unsigned char *)(base + CR));
    if( (cr & PBMASK) != page ){
        cr &= 0x3F;
        outbyte((unsigned char *)(base + CR), cr|page);
    }
    return inbyte((unsigned char *)(base + reg_addr));
}

static void write_nic(unsigned long base, unsigned long reg_addr, unsigned char value, unsigned char page)
{
    unsigned char cr;

    cr = inbyte((unsigned char *)(base + CR));
    if( (cr & PBMASK) != page ){
        cr &= 0x3F;
        outbyte((unsigned char *)(base + CR), cr|page);
    }
    outbyte((unsigned char *)(base + reg_addr), value);
}

/*
** This function returns the difference in milliseconds between two
** time base values. Calculation depends on the input clock Hertz freqency.
** Multiplier values for upper value depending on clock
** 50Mhz: 85899
** 33Mhz: 128849
** 25Mhz: 171798
** 20Mhz: 214748
*/
static unsigned long tbmilisec_dif(tb_t * tb_end, tb_t * tb_start)
{
    unsigned long upper;
    unsigned long lower;

    upper = tb_end->tbu - tb_start->tbu;
    if( tb_start->tbl > tb_end->tbl )
	upper--;
    lower = tb_end->tbl - tb_start->tbl;
    return upper * 85899 + lower / 50000;
}

/*
** enet_init.  Must be called with interrupts disabled.
*/
static int enet_init(int device)
{
    struct t_edd *ep;
    unsigned long msr, nrBufPkts;
    unsigned char regMCR_A, regMCR_B;
    tb_t tb_start, tb_end;

    if( device < 0 || device >= DIM(edd) )		/* Unknown device */
	return -1;
    ep = &edd[device];
    /*
    ** Reinitialize driver structure, just be sure we use correct values
    ** when invoked for dumping memory.
    */
    memset(&ep->outframe, 0, sizeof ep->outframe);
    memset(ep->inframe, 0, sizeof ep->inframe);
    ep->flih_bufnum = ep->read_bufnum = 0;
    /*
    ** Initialize RAM Ring Buffer global variables.
    ** - NIC RAM starts at: PROM_SIZE   (0x0000-0x3FFF PROM Register area)
    ** - NIC RAM ends at:   PROM_SIZE+RAM_SIZE
    ** - XMIT Buffer Ring:  PROM_SIZE    -> _PStart_Addr-1
    ** - RCV  Buffer Ring:  _PStart_Addr -> _PStop_Addr-1
    ** Note: _PStop_Addr points always one byte higher than limit
    **       _PStop_HAddr points always one buffer(256B) higher than limit
    */
    nrBufPkts = PROM_SIZE / PACKET_SIZE;
    if( nrBufPkts > 255 )
	nrBufPkts=255;				/* MAX packets if 64kB RAM */
    ep->_TPStart_HAddr = PROM_SIZE / PACKET_SIZE;	/* A15-A8 of 1st xmit buf */
    ep->_PStart_HAddr = ep->_TPStart_HAddr + XMIT_BUFS;	/* A15-A8 of 1st rcv buf */
    ep->_PStart_Addr = ep->_PStart_HAddr * PACKET_SIZE;	/* Start addr of 1st rcv buf */
    ep->_PStop_HAddr = ep->_PStart_HAddr + RCV_BUFS;	/* A15-A8 of last rcv buf */
    ep->_PStop_Addr = ep->_PStop_HAddr * PACKET_SIZE;	/* Stop addr of rcv buf */

    /*
    ** Initialization procedure from page 1-118 of the controller data sheet.
    */
    msr = ppcAndMsr(~ppcMsrEE);		/* Disable Ext. IR */
    WRITE_CR(ep, RD2|STP);		/* Stop controller (take offline) */
    ppcMftb(&tb_start);
    do
	ppcMftb(&tb_end);
    while( tbmilisec_dif(&tb_end, &tb_start) < 3 );

    /* Initialization procedure for Mode Configuration Register A and B */
    regMCR_A = read_nic(ep->base, MCR_A, PAGE0);	/* Read MCR_A and imm. write it */
    regMCR_A = ep->ioaddr|INT0;			/* Set I/O base addr + INT 0 */
    outbyte((unsigned char *)(ep->base + MCR_A), regMCR_A);
    ppcEieio();
    regMCR_B = read_nic(ep->base, MCR_B, PAGE0);	/* Read MCR_B and imm. write it */
    regMCR_B = PHYS_TPI;
    outbyte((unsigned char *)(ep->base + MCR_B), regMCR_B);
    ppcEieio();
    regMCR_A = inbyte((unsigned char *)(ep->base + MCR_A));
    regMCR_B = inbyte((unsigned char *)(ep->base + MCR_B));

    /* Initialization LANTIC core regs */
    write_nic(ep->base, DCR, LS|FT8B, PAGE0);
    write_nic(ep->base, RBCR0, 0, PAGE0);
    write_nic(ep->base, RBCR1, 0, PAGE0);
    write_nic(ep->base, RCR, AB, PAGE0);
    write_nic(ep->base, TCR, LBM1, PAGE0);

    /*
    ** Initialize receive buffer ring.  Set aside first 8 pages as transmit
    ** pages, rest as receive ring.
    ** In a 8 K buffer, pages are numbered 0 to 31.  First
    ** receive page is page 8.  PSTOP is set to one past last valid buffer.
    */
    write_nic(ep->base, PSTART, ep->_PStart_HAddr, PAGE0);
    write_nic(ep->base, PSTOP, ep->_PStop_HAddr, PAGE0);
    write_nic(ep->base, BNRY, ep->_PStart_HAddr, PAGE0);
    ep->next_packet_ptr = ep->_PStart_HAddr + 1;
    /*
    ** Clear all bits in ISR register.  Interrupt on receive packets and buffer
    ** overflow only.
    */
    write_nic(ep->base, ISR, 0xFF, PAGE0);
    write_nic(ep->base, IMR, PRXE|OVWE|PTXE|TXEE, PAGE0);
    /*
    ** Initialize physical address and multicast address.  Set current page
    ** pointer to page start.
    */
    write_nic(ep->base, PAR0, ep->arpcom.ac_enaddr[0], PAGE1);
    write_nic(ep->base, PAR1, ep->arpcom.ac_enaddr[1], PAGE1);
    write_nic(ep->base, PAR2, ep->arpcom.ac_enaddr[2], PAGE1);
    write_nic(ep->base, PAR3, ep->arpcom.ac_enaddr[3], PAGE1);
    write_nic(ep->base, PAR4, ep->arpcom.ac_enaddr[4], PAGE1);
    write_nic(ep->base, PAR5, ep->arpcom.ac_enaddr[5], PAGE1);
    write_nic(ep->base, MAR0, 0, PAGE1);
    write_nic(ep->base, MAR1, 0, PAGE1);
    write_nic(ep->base, MAR2, 0, PAGE1);
    write_nic(ep->base, MAR3, 0, PAGE1);
    write_nic(ep->base, MAR4, 0, PAGE1);
    write_nic(ep->base, MAR5, 0, PAGE1);
    write_nic(ep->base, CURR, ep->_PStart_HAddr + 1, PAGE1);

    /* Start controller and take it out of loopback */
    WRITE_CR(ep, STA|RD2);
    write_nic(ep->base, TCR, LBM0, PAGE0);
    ppcMtmsr(msr);

    /* Check the link indicator should be done earlier */
    regMCR_B = read_nic(ep->base, MCR_B, PAGE0);
    if( !(regMCR_B & GDLNK) )
	return -1;
    return 0;
}

static int enet_ioctl(struct ifnet *ifp, unsigned long cmd, caddr_t data)
{
    struct ifaddr *ifa = (struct ifaddr *) data;
    struct arpcom *acp;
    pthread_mutex_t *s = net_splimp();
    struct sockaddr_in *sin = (struct sockaddr_in *)IA_SIN(ifa);
    unsigned long msr;
    int error = ESUCCESS;

    switch (cmd) {
    case SIOCSIFADDR:
	if (ifa->ifa_addr->sa_family == AF_INET) {
	    /*
            ** Is there any addresses for the interface.
            */
	    if (ifp->if_addrlist == NULL) {
		error = ENETUNREACH;
		break;
	    }
	    /*
            ** Enet should be already enabled.  Lets just register our receive
            ** function and our IP address.
            */
	    msr = ppcAndMsr(~ppcMsrEE);
	    if (ifp->if_flags & IFF_UP)
		ifp->if_flags |= IFF_RUNNING;
	    acp = &edd[0].arpcom;
	    acp->ac_ipaddr.s_addr = sin->sin_addr.s_addr;
	    acp = &edd[1].arpcom;
	    acp->ac_ipaddr.s_addr = sin->sin_addr.s_addr;
	    ppcMtmsr(msr);
	} else
	    error = EAFNOSUPPORT;
	break;
    case SIOCSIFFLAGS:
	/*
	** If interface is being brought down then IFF_UP is not set.  If on
	** the other hand interface is being brought up IFF_UP is set.
	*/
	msr = ppcAndMsr(~ppcMsrEE);
	if (ifp->if_flags & IFF_UP)
	    if (enet_init(0) || enet_init(1) )
		error = ENETDOWN;
	    else
		ifp->if_flags |= IFF_RUNNING;
	else
	    if (enet_init(0) || enet_init(1) )
		error = EBUSY;
	    else
		ifp->if_flags &= ~IFF_RUNNING;
	ppcMtmsr(msr);
	break;
    default:
	error = EINVAL;
    }
    net_splx(s);
    return error;
}

static int enet_send_packet(struct t_edd *ep, struct ether_header *eh, struct mbuf *m, int total)
{
    struct mbuf	*temp_m;
    unsigned long msr;
    int saved_total;
    int rc;
    unsigned long i;
    struct timespec  timeout;
    sigset_t new_mask, old_mask;

    /*
    ** Disable interrupts and serialize access to this section of code. 
    ** No two threads can transmit at the same time. Temporarily disable thread
    ** signals.
    */
    msr = ppcAndMsr(~ppcMsrEE);
    sigemptyset(&new_mask);
    pthread_sigmask(SIG_SETMASK, &new_mask, &old_mask);
    if (sem_wait(&ep->dev_access)) {
	pthread_sigmask(SIG_SETMASK, &old_mask, NULL);
	ppcMtmsr(msr);
	return EIO; 
    }

    /*
    ** Set up DMA to transfer packet to the controller chip.  Do not inline
    ** write port, compiler will optimize it in such a way that machine check
    ** will result the first time this is run.
    */
    write_nic(ep->base, RBCR0, total & 0xff, PAGE0);
    write_nic(ep->base, RBCR1, (total & 0xff00) >> 8, PAGE0);
    write_nic(ep->base, RSAR0, 0, PAGE0);
    write_nic(ep->base, RSAR1, ep->_TPStart_HAddr, PAGE0);
    WRITE_CR(ep, RD1);
    /* First 14 characters is ether_header.  */
    saved_total = total;
    for( i = 0; i < sizeof(struct ether_header); ++i )
	WRITE_PORT(ep, ((char *)eh)[i]);
    total -= sizeof(struct ether_header);

    /* Pull rest out of mbuf's */
    for (temp_m = m; temp_m; temp_m = temp_m->m_next) {
	for( i = 0; i < temp_m->m_len; i++ )
	    WRITE_PORT(ep, mtod(temp_m, u_char *)[i]);
	total -= temp_m->m_len;
    }

    /* Back fill if necessary */
    for( ; --total >= 0; )
	WRITE_PORT(ep, 0);
    /* Issue transmit command  */
    write_nic(ep->base, TPSR, ep->_TPStart_HAddr, PAGE0);
    write_nic(ep->base, TBCR0, saved_total & 0xff, PAGE0);
    write_nic(ep->base, TBCR1, (saved_total & 0xff00) >> 8, PAGE0);
    write_nic(ep->base, ISR, PTX|TXE, PAGE0);
    WRITE_CR(ep, TXP);

    /* Wait for transmit to complete (wait up to 3 seconds) */
    timeout.tv_sec = 3;
    timeout.tv_nsec = 0;
    rc = 0;
    if( sem_timedwait(&ep->xmit_sem, &timeout) ){
	enet_init(ep - edd);
	rc = EIO;
    }
    /* Was transmit succesful */
    if( (read_nic(ep->base, TSR, PAGE0) & PTXS) == 0 )
	rc = EIO;
    ppcMtmsr(msr);
    pthread_sigmask(SIG_SETMASK, &old_mask, NULL);
    sem_post(&ep->dev_access);
    return rc;
}

static int enet_output(struct ifnet *ifp, struct mbuf *m, struct sockaddr *dst)
{
    struct arpcom *acp = &edd[enet_chip].arpcom;
    struct mbuf *temp_m;
    struct in_addr idst;
    int rc;
    int total;
    struct ether_header eh;


    /*
    ** Interface down, return error.
    */
    if ((ifp->if_flags & (IFF_UP | IFF_RUNNING)) != (IFF_UP | IFF_RUNNING)) {
	m_freem(m);
	return ENETDOWN;
    }
    idst.s_addr = ((struct sockaddr_in *) dst)->sin_addr.s_addr;
    switch (dst->sa_family) {
    case AF_INET:
	/*
	** A return value of 1 indicates that eh.ether_dhos structure
	** has been filled in and the packet should be sent normally.  A 0
	** return indicates that the packet has been taken over by arp, for
	** transmission when address is resolved (if address is not resolved the
	** packet is thrown out).
	*/
	rc = enet_arpresolve(acp, m, &idst, eh.ether_dhost);
	if (!rc )
	    return 0;
	eh.ether_type = ETHERTYPE_IP;
	break;
    case AF_UNSPEC:
	/*
	** Arp frames come down here as unspecified frames (AF_UNSPEC).
	*/
	memcpy((void *) &eh, (void *) dst->sa_data, sizeof(struct ether_header));
	break;
    default:
	m_freem(m);
	return EAFNOSUPPORT;
    }
    /*
    ** Now we have destination hardware address, type of transmission and data
    ** so we can build a buffer to be transmitted.
    */
    total = sizeof(struct ether_header);
    for (temp_m = m; temp_m; temp_m = temp_m->m_next)
	if( temp_m->m_len != 0 )
	    total += temp_m->m_len;

    memcpy(eh.ether_shost, acp->ac_enaddr, ENET_HWADDRLEN);
    /* If frame is too long or too short, modify length */
    if( total > ENET_MAC_LENGTH ){
	m_freem(m);
	return EIO;
    }else
	if( total < ENET_MAC_MIN_LENGTH )
	    total = ENET_MAC_MIN_LENGTH;
    rc = enet_send_packet(&edd[enet_chip], &eh, m, total);
    m_freem(m);
    return rc;
}

static struct mbuf *ent_mbuf(struct mbuf **anchor, int *len, struct mbuf *previous, struct ifnet *ifp)
{
    struct mbuf *m;

    /* Get an mbuf */
    if( (m = (struct mbuf *)mempool_alloc(__global_mbuf_pool)) == NULL ){
	if( *anchor ){
	    m_freem(*anchor);
	    *anchor = NULL;
	}
	return NULL;
    } 
    if( *anchor == NULL ){		/* Get header mbuf */
	m->m_type = MT_DATA;
	m->m_next = (struct mbuf *)NULL;
	m->m_nextpkt = (struct mbuf *)NULL;
	m->m_data = m->m_pktdat;
	m->m_flags = M_PKTHDR;
	m->m_pkthdr.rcvif = ifp;
	m->m_pkthdr.len = 0;
	m->m_len = 0;
	*anchor = m;
	*len = MHLEN;
    } else {				/* Get next memory buffer */
	m->m_type = MT_DATA;
	m->m_next = (struct mbuf *)NULL;
	m->m_nextpkt = (struct mbuf *)NULL;
	m->m_data = m->m_dat;
	m->m_flags = 0;
	m->m_len = 0;
	previous->m_next = m;
	*len = MLEN;
    }
    /*
    ** Fix alignment problem in IP protocol stack software, which
    ** assumes a word aligned IP header. This is not the case since we
    ** prepend the ethernet header. Therefore start with ethernet header
    ** on a halfword alignment.
    */
    m->m_data += 2;
    *len -= 2;
    return m;
}

static struct mbuf *read_into_m_buf(int read_count, int *len, struct mbuf **anchor, struct mbuf *m, struct t_edd *ep)
{
    int	i;
    unsigned char ch;

    for( i = 0; i < read_count; i++ ){
	if( *len == 0 && m != NULL )
	    m = ent_mbuf(anchor, len, m, &ep->arpcom.ac_if);
	ch = READ_PORT(ep);
	if( m != NULL ){
	    mtod(m, char *)[m->m_len] = ch;
	    m->m_len++;
	    --*len;
	}
   }
   return m;
}

static int enet_rcv_packet(struct t_edd *ep)
{
    unsigned char recv_status;
    unsigned char next_packet;
    unsigned char bp;
    unsigned long count;
    unsigned long dma_count;
    unsigned long addr;
    unsigned long residual;
    int len;
    struct mbuf *m, *anchor;

    /* Receive first 4 bytes giving us the receive packet statistics. */
    write_nic(ep->base, RBCR0, 4, PAGE0);
    write_nic(ep->base, RBCR1, 0, PAGE0);
    write_nic(ep->base, RSAR0, 0, PAGE0);
    write_nic(ep->base, RSAR1, (ep->next_packet_ptr * PACKET_SIZE) >> 8, PAGE0);
    WRITE_CR(ep, RD0);
    recv_status = READ_PORT(ep);
    next_packet = READ_PORT(ep);
    count = READ_PORT(ep);
    count |= READ_PORT(ep) << 8;
    /*
    ** We already got 4 bytes.  Set up addr where the next packet resides.
    ** Check count if too big.
    */
    count -= 4;
    if( count > ENET_MAC_LENGTH || (recv_status & PRXS) == 0 ){
	enet_init(ep - edd);
	return -1;
    }
    addr = ep->next_packet_ptr * PACKET_SIZE + 4;
    /*
    ** Pull packet out of memory.  If we do not hit the end boundry handle it
    ** in one transfer.  Packet is received starting at byte 3 so that the ip
    ** packet will be word aligned and we do not have to do memmove later.
    */
    anchor = NULL;
    m = ent_mbuf(&anchor, &len, NULL, &ep->arpcom.ac_if);
    if( addr + count < ep->_PStop_Addr ){
	/* Set up DMA address and count. */
	write_nic(ep->base, RBCR0, count & 0xff, PAGE0);
	write_nic(ep->base, RBCR1, (count & 0xff00) >> 8, PAGE0);
	write_nic(ep->base, RSAR0, addr & 0xff, PAGE0);
	write_nic(ep->base, RSAR1, (addr & 0xff00) >> 8, PAGE0);
	WRITE_CR(ep, RD0);
	m = read_into_m_buf(count, &len, &anchor, m, ep);
    }else{
	/*
	** Pull out all the bytes to the end boundry then do another transfer to
	** get the rest of the packet.
	*/
	residual = addr + count - ep->_PStop_Addr;
	dma_count = count - residual;
	/* Set up DMA address, and count. */
	write_nic(ep->base, RBCR0, dma_count & 0xff, PAGE0);
	write_nic(ep->base, RBCR1, (dma_count & 0xff00) >> 8, PAGE0);
	write_nic(ep->base, RSAR0, addr & 0xff, PAGE0);
	write_nic(ep->base, RSAR1, (addr & 0xff00) >> 8, PAGE0);
	WRITE_CR(ep, RD0);
	m = read_into_m_buf(dma_count, &len, &anchor, m, ep);
	/* Set up DMA address and count (start transfer at start page). */
	write_nic(ep->base, RBCR0, residual & 0x000000FF, PAGE0);
	write_nic(ep->base, RBCR1, (residual & 0x0000FF00) >> 8, PAGE0);
	write_nic(ep->base, RSAR0, 0, PAGE0);
	write_nic(ep->base, RSAR1, ep->_PStart_HAddr, PAGE0);
	WRITE_CR(ep, RD0);
	m = read_into_m_buf(residual, &len, &anchor, m, ep);
    }
    /*
    ** Advance boundry pointer using algorighm on page 1-56 of AT/LANTIC
    ** data sheet.
    ** If "next_packet" is outside bounds signal error.
    */
    if( next_packet > ep->_PStop_HAddr || next_packet < ep->_PStart_HAddr ){
	enet_init(ep - edd);
	return -1;
    }
    ep->next_packet_ptr = next_packet;
    bp = ep->next_packet_ptr - 1;
    if( bp < ep->_PStart_HAddr )
	bp = ep->_PStop_HAddr - 1;
    write_nic(ep->base, BNRY, bp, PAGE0);
    /* Enqueue packet on the correct receive queue */
    if( anchor != NULL ){
	/*
	** m == NULL then we cound not receive entire packet!  Discard partialy
	** received packets.
	*/
	if( m == NULL ){
	    m_freem(anchor);
	    return 0;
	}
	/* Enqueue packed and post appropriate semaphore.  */
	anchor->m_nextpkt = NULL;
	anchor->m_pkthdr.len = count;
	if( my_ifq.ifq_tail == NULL)
	    my_ifq.ifq_head = anchor;
	else
	    my_ifq.ifq_tail->m_nextpkt = anchor;
	my_ifq.ifq_tail = anchor;
	sem_post(&my_ifq.enet_sem);
    }
    return count;
}
 
static void enet_int(void *arg)
{
    unsigned char isr;
    struct t_edd *ep = &edd[(int)arg];

    while (1) {
	isr = read_nic(ep->base, ISR, PAGE0);
	if( (isr & (PRX|OVW|PTX|TXE)) == 0 )
	    return;
	if( isr & PTX ){
	    write_nic(ep->base, ISR, PTX, PAGE0);
	    sem_post(&ep->xmit_sem);
	}else
	    if( isr & TXE ){
		write_nic(ep->base, ISR, TXE, PAGE0);
		sem_post(&ep->xmit_sem);
	    }else
		if( isr & PRX ){
		    if( enet_rcv_packet(ep) == -1)
			continue;
		    /*
		    ** Clear interrupt only if there are no more packets
		    ** in ring buffer
		    */
		    isr = read_nic(ep->base, CURR, PAGE1);
		    if( ep->next_packet_ptr == isr )
		        write_nic(ep->base, ISR, PRX, PAGE0);
		}else
		    /*
		    ** All the packets are removed from the buffer,
		    ** so reinitialize controller.
    		    */
		    enet_init(ep - edd);
    }
    return;
}

/*
** Send an ethernet swap message via broadcast.
*/
static int enet_swap_msg(int dev)
{
    struct t_edd *ep = &edd[dev];
    struct ether_header eh;
    struct mbuf mbuf;
    struct ether_swapcmd *es;

    memset(&mbuf, 0, sizeof mbuf);
    memset(eh.ether_dhost, 0xff, ENET_HWADDRLEN);
    memcpy(eh.ether_shost, ep->arpcom.ac_enaddr, ENET_HWADDRLEN);
    eh.ether_type = ETHERTYPE_SWAP;
    mbuf.m_type = MT_CONTROL;
    mbuf.m_next = (struct mbuf *)NULL;
    mbuf.m_nextpkt = (struct mbuf *)NULL;
    mbuf.m_data = mbuf.m_dat;
    mbuf.m_len = 2;
    mbuf.m_flags = 0;
    es = (struct ether_swapcmd *)mbuf.m_dat;
    es->sw_cmd = ETHERSWAP_USEIF;
    es->sw_ifno = dev;
    return enet_send_packet(ep, &eh, &mbuf, ENET_MAC_MIN_LENGTH);
}

/*
** Thread to watch out for the link status bit of both ethernet chips. We don't
** get an interrupt if link status goes bad.
*/
static void *enet_link_thread(void *arg)
{
    int	old_state;
    sigset_t signal_mask;
    struct t_edd *ep[ENET_DEV];
    timespec_t timewait;

    sigemptyset(&signal_mask);
    pthread_sigmask(SIG_SETMASK, &signal_mask, NULL);
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &old_state);
    timewait.tv_sec = 0;
    timewait.tv_nsec = 100000000;
    ep[0] = edd;
    ep[1] = &edd[1];
    for( ;; ){
	nanosleep(&timewait, NULL);
	ep[0]->linkok = read_nic(ep[0]->base, MCR_B, PAGE0) & GDLNK;
	ep[1]->linkok = read_nic(ep[1]->base, MCR_B, PAGE0) & GDLNK;
	if( ep[enet_chip]->linkok == 0 ){
	    if( ep[enet_chip ^ 1]->linkok ){
		if( arp("-f") )
		    s1dbprintf(25000000, 0, "arp flush failed\n");
		enet_chip ^= 1;
		if( enet_swap_msg(enet_chip) )
		    printf("swap of ethernet failed\n");
	    }

	}
    }
}

static void *enet_rcv_thread(void *arg)
{
    sigset_t signal_mask;
    struct mbuf *m;
    pthread_mutex_t *s;
    int	old_state;
    unsigned long msr;
    unsigned short fr_type;
    struct ether_swapcmd *eswap;

    sigemptyset(&signal_mask);
    pthread_sigmask(SIG_SETMASK, &signal_mask, NULL);
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &old_state);
    while( 1 ){
	while( sem_wait(&my_ifq.enet_sem) )
	    ;
	msr = ppcAndMsr(~ppcMsrEE);
	m = my_ifq.ifq_head;
	/* If there are mbus queued then process them */
	if( m != NULL ){
	    if( (my_ifq.ifq_head = m->m_nextpkt) == NULL )
		my_ifq.ifq_tail = NULL;
	    m->m_nextpkt = NULL;
	    ppcMtmsr(msr);
	    fr_type = (mtod(m, struct ether_header *))->ether_type;
	    m->m_data += sizeof(struct ether_header);    
	    m->m_len -= sizeof(struct ether_header);
	    m->m_pkthdr.len -= sizeof(struct ether_header);
	    switch( fr_type ){
	    case ETHERTYPE_IP:		/* Process IP frame */
		/*
		** Check if frame came in on active interface, drop input
		** from alternate (passive) interface.
		*/
		if( (m->m_flags & M_PKTHDR) && m->m_pkthdr.rcvif 
		&& &edd[enet_chip].arpcom.ac_if != m->m_pkthdr.rcvif ){
		    m_freem(m);
		}else
		    if( IF_QFULL(&ipintrq) ){
			IF_DROP(&ipintrq);
			m_freem(m);
		    }else{
			s = net_splimp();
			IF_ENQUEUE(&ipintrq, m);
			net_splx(s);
			schednetisr(NETISR_IP);
		    } 
		break;
	    case ETHERTYPE_ARP:		/* Process ARP frame */
		/* Find out which interface this packet came in on */
		if( (m->m_flags & M_PKTHDR) && m->m_pkthdr.rcvif )
		    if( &edd[enet_chip].arpcom.ac_if == m->m_pkthdr.rcvif )
			enet_arpinput(&edd[enet_chip].arpcom, m);
		    else
			m_free(m);
		break;
	    case ETHERTYPE_SWAP:	/* Process SWAP frame */
		eswap = mtod(m, struct ether_swapcmd *);
		if( eswap->sw_cmd == ETHERSWAP_USEIF
		&& eswap->sw_ifno != enet_chip 
		&& (eswap->sw_ifno == 0 || eswap->sw_ifno == 1) ){
		    /* Swap ethernet chips */
		    msr = ppcAndMsr(~ppcMsrEE);
		    enet_chip = eswap->sw_ifno;
		    if( arp("-f") )
			s1dbprintf(25000000, 0, "arp flush failed\n");
		    ppcMtmsr(msr);
		}
		/* fall through intended */
	    default:
		m_freem(m);
		break;
	    }
	}else
	    ppcMtmsr(msr);
    }
    return (void *)-1;
}

int boardenet_attach(int dev)
{
    struct arpcom *arp;
    struct ifnet *ifp;
    flih_t enet_flih;
    struct t_edd *ep;
    struct t_cccfg *cccfgp = OB_getcccfg();
    static int installed;

    /*
    ** For failsave ethernet device driver, fake second device.
    */
    if( installed == 1 )
	return 0;

    /*
    ** Initialize global variables. Next_packet_ptr is set by enet_init().
    */
    if( dev < 0 || dev >= DIM(edd) )		/* Unknown device */
	return -1;
    ep = &edd[dev];

    /*
    ** Initialize tcp/ip interface.  
    */
    arp = &ep->arpcom;
    memset((void *) arp, 0, sizeof *arp);
    memcpy(arp->ac_enaddr, cccfgp->netif[dev].mac, sizeof arp->ac_enaddr);
    ifp = &arp->ac_if;
    ifp->if_name = "ent";
    ifp->if_unit = dev;
    ifp->if_mtu = ETHERMTU;
    ifp->if_flags = IFF_BROADCAST | IFF_NOTRAILERS;
    ifp->if_type = IFT_ETHER;
    ifp->if_ioctl = enet_ioctl;
    ifp->if_output = enet_output;
    ifp->if_snd.ifq_maxlen = ETHERMTU;
    /*
    ** Create semaphore and start receiver thread.
    */
    if (sem_init(&ep->xmit_sem, 0, 0))
	return -1;
    if (sem_init(&ep->dev_access, 0, 1)) {
	sem_destroy(&ep->xmit_sem);
	return -1;
    }
    if( !installed ){
	struct sched_param sparam;
	pthread_attr_t ent_attr;
	pthread_t threadid;

	if( pthread_attr_init(&ent_attr) || sem_init(&my_ifq.enet_sem, 0, 0) ){
	    sem_destroy(&ep->xmit_sem);
	    sem_destroy(&ep->dev_access);
	    return -1;
	}
	sparam.sched_priority = ENET_THREAD_PRIO;
	pthread_attr_setschedparam(&ent_attr, &sparam);
	pthread_attr_setstacksize(&ent_attr, ENET_RCV_STACKSIZE);
	if (pthread_create(&threadid, &ent_attr, enet_rcv_thread, NULL)) {
	    pthread_attr_destroy(&ent_attr);
	    sem_destroy(&my_ifq.enet_sem);
	    sem_destroy(&ep->xmit_sem);
	    sem_destroy(&ep->dev_access);
	    return -1;
	}
	sparam.sched_priority = ENET_THREAD_PRIO;
	pthread_attr_setschedparam(&ent_attr, &sparam);
	pthread_attr_setstacksize(&ent_attr, 8192);
	if (pthread_create(&threadid, &ent_attr, enet_link_thread, NULL))
	    return -1;
	installed = 1;
	package_install(libname, &installed, NULL);
	enet_chip = dev;
    }
    if_attach(ifp);
    /*
    ** Install interrupt handler.
    */
    enet_flih.flih_stack = (void *) ((unsigned long) ep->flih_stack + ENET_FLIH_SIZE);
    enet_flih.flih_function = (void *) enet_int;
    enet_flih.arg = (void *)dev;
    if( int_install(edd[dev].intr, &enet_flih, NULL) != 0)
	return -1;
    enet_init(dev);
    ext_int_config(edd[dev].intr, edd[dev].intr_cfg);
    int_enable(edd[dev].intr);

    /*
    ** Install second ethernet chip
    */
    dev ^= 1;
    ep = &edd[dev];
    arp = &ep->arpcom;
    memcpy(arp->ac_enaddr, cccfgp->netif[dev].mac, sizeof arp->ac_enaddr);
    ifp = &arp->ac_if;
    ifp->if_name = "ent";
    ifp->if_unit = dev;
    ifp->if_mtu = ETHERMTU;
    ifp->if_flags = IFF_BROADCAST | IFF_NOTRAILERS | IFF_UP | IFF_RUNNING;
    ifp->if_type = IFT_ETHER;
    ifp->if_ioctl = enet_ioctl;
    ifp->if_output = enet_output;
    ifp->if_snd.ifq_maxlen = ETHERMTU;
    if (sem_init(&ep->xmit_sem, 0, 0))
	return -1;
    if (sem_init(&ep->dev_access, 0, 1)) {
	sem_destroy(&ep->xmit_sem);
	return -1;
    }
    enet_flih.flih_stack = (void *) ((unsigned long) ep->flih_stack + ENET_FLIH_SIZE);
    enet_flih.flih_function = (void *) enet_int;
    enet_flih.arg = (void *)dev;
    if( int_install(edd[dev].intr, &enet_flih, NULL) != 0)
	return -1;
    enet_init(dev);
    ext_int_config(edd[dev].intr, edd[dev].intr_cfg);
    int_enable(edd[dev].intr);
    return 0;
}
