//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  IEEE-1394 OHCI Driver for Linux
//
//  2.2.9 KERNEL
//
//  Original Author: Jeffrey S. Akgul
//
//  Modified by Bill Alexander (www.geocities.com/billalexander)
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifdef MODULE
#include <linux/module.h>
#include <linux/version.h>
#endif

#ifndef __KERNEL__
#define __KERNEL__
#endif

#if LINUX_VERSION_CODE < 0x020100
#define ioremap vremap
#define iounmap vfree
#endif

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/stat.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/io.h>
#include <linux/tty.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/timer.h>
#include <linux/wrapper.h>
#include <asm/irq.h>
#include <asm/dma.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/init.h>
#include <linux/poll.h>
#include <asm/uaccess.h>

//-------------------------------------------
// VMA OPERATIONS
//-------------------------------------------
#include <asm/pgtable.h>

//-------------------------------------------
// My 1394/ohci definitions
//-------------------------------------------
#include "ohci.h"

//-------------------------------------------
// My debug definitions
//-------------------------------------------
#define DEBUG
#define _DBG_DRV_
#define _DBG_LINUX_
#define _DBG_DRV_MAIN_MODULE_
#include "dbg.h"

//-------------------------------------------
//  PROTOTYPES
//-------------------------------------------
int init_module(void);
void cleanup_module(void);
int device_init(void);
int device_attach(int irq);
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static int device_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static int device_mmap(struct file *, struct vm_area_struct *);
static void ohci_intr(int irq, void *dev_id, struct pt_regs * regs);
int device_dma_mmap(struct file *, struct vm_area_struct *);
static int device_dma_open(struct inode *, struct file *);
static int device_dma_release(struct inode *, struct file *);
// static ssize_t device_read(struct file *, char *, size_t, loff_t *);
// static ssize_t device_write(struct file *, const char *, size_t, loff_t *);
extern struct file_operations ohci_fops;
extern struct file_operations ohci_dma_fops;

//-------------------------------------------
//  MACROS
//-------------------------------------------
#define READREG(regname)  (readl  ((unsigned char *)VIR_addr+regname))
#define WREG(regname,data)(writel (data,(unsigned char *)VIR_addr+regname))
#ifndef NodeID
#define NodeID 0xE8
#endif
#define NODEID() (READREG(NodeID) & 0xFFFFFFFF)
#define PACKETBUFFERSIZE 1024
#define TYPE(dev) (MINOR(dev) >> 4)
#define NUM(dev) (MINOR(dev) & 0xf)
#define OHCI_MAX_TYPE 2

//-------------------------------------------
// GLOBAL DATA STRUCTURES USED BY THE DRIVER
//-------------------------------------------
int Mapping_errno = 0;
unsigned long off;
unsigned long phys_add;
unsigned long vsize;
unsigned long VIR_addr;
unsigned long dmabuf;
static int ohci_major = OHCI_MAJOR_DEV_NUM;

static struct pci_dev *dev;

struct ohci_data
{
    unsigned short  irq;
    u32             AT_programmbuffer[100];
    u32             *ATRQ_program_small;
};

static struct ohci_data *ohcidata;

// File operations when an application calls open() 
// for special files /dev/ohci0 and /dev/ohci1
// NOTE:  root must create these special files
// via mknod.  For example:
//		mknod /dev/ohci0 c 254 0
//		mknod /dev/ohci1 c 254 1
//  The app then does an open:
//		fh0 = open ("/dev/ohci0",O_RDWR);
//		fh1 = open ("/dev/ohci1",O_RDWR);
static struct file_operations *ohci_fop_array[]=
{
    &ohci_fops,    /* type 0 */
    &ohci_dma_fops /* type 1 */
};

static struct file_operations ohci_fops =
{
    NULL,           //1 llseek
    NULL,           // device_read 
    NULL,           // device_write 
    NULL,           //4 readdir
    NULL,           //5 poll
    device_ioctl,
    device_mmap, 
    device_open,
    NULL,           // flush
    device_release,
    NULL,
};

static struct file_operations ohci_dma_fops=
{
    NULL,           //1 llseek
    NULL,           //2 read
    NULL,           //3 write
    NULL,           //4 readdir
    NULL,           //5 poll
    device_ioctl,
    device_dma_mmap, 
    device_dma_open,
    NULL,           // flush
    device_dma_release,
    NULL,
};

u32 AT_DMA_program_small_ROM[8] =
{
    0x120C000C,      /* output_last_imm */
    0x00000000,      /* output_more_1   */
    0x00000000,      /* output_more_2   */
    0x00000000,      /* output_more_3   */
    0x00000040,      /* output_more_4   */
    0xFFC3FFFF,      /* output_more_5   */
    0xF0000404,      /* output_more_6   */
    0x00000000       /* ouput_last      */
};

// If the OHCI baseaddress[0] is not page aligned, we
// need to communicate the fact up to the application.
// REASON:  Linux can only page align mapped memory.
int gRegsPageAlignmentOffset = 0;

// The physical address of our DMA page
int gDmaPagePhysAddr = 0;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// init_module()
//
//      Module entry point.
//
// Called by:
//
//		kernel during insmod()
//
// Entry:
//
//      NONE
//
// Exit:
//
//      0       OK.  Driver loaded.
//      ENODEV  No device was detected.  Driver failed to load.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int init_module(void)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int retcode;
    int ccode;

    // Initialize debug module
    DBGInitialize (DBG_DEFAULT);

    DBOUT (DBG_LF_ENTRY, "ohci: init_module()\n");

    // Assume failure
    retcode = ENODEV;

    // Initialize the device
    ccode = device_init();
    if (ccode)
        goto InitError;

    // Do any additional initialization here

    // OK
    retcode = 0;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohci: init_module() exit, retcode = %08X\n", retcode);
    return retcode;

InitError:
    DBOUT (DBG_ERROR, "ohci: ERROR:  Unable to initialize device\n");
    retcode = ENODEV;
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// cleanup_module
//
//      Module exit point.  Cleans up all allocated resources.
//
// Called by:
//
//		kernel during rmmod()
//
// Entry:
//
//      NONE
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void cleanup_module(void)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int i;

    DBOUT (DBG_LF_ENTRY, "ohci: cleanup_module()\n");

    if(MOD_IN_USE)
    {
        DBOUT (DBG_ERROR, "ohci: ERROR: SHOULD NOT HAPPEN! Device busy, but removing anyway!\n");
    }

    // Free memory
    if (VIR_addr)
    {
        iounmap((void *)(VIR_addr & PAGE_MASK));
        free_page(VIR_addr);
    }
    if (ohcidata)
    {
        free_irq(ohcidata->irq,NULL);
        kfree(ohcidata);
    }

    // Free DMA
    for (   i=MAP_NR (dmabuf);
            i<= MAP_NR(dmabuf+PAGE_SIZE-1); 
            i++)
    {
        mem_map_unreserve(i);
    }
    if (dmabuf)
        free_pages(dmabuf,1);

    // Unregister driver
    if (ohci_major)
        unregister_chrdev(ohci_major,"ohci");

    DBOUT (DBG_LF_EXIT, "ohci: cleanup_module(), device unloaded\n");

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// device_init()
//
//      Finds the device and initializes it.
//
// Called by:
//
//		init_module()
//
// Entry:
//
//      NONE
//
// Exit:
//
//      0       OK.  Driver loaded.
//      ENODEV  No device was detected.  Driver failed to load.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int device_init(void)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int retcode;
    unsigned short vendor_id = 0x104c;
    unsigned short device_id = 0x8009;
    int base;
    unsigned char irq;
    int ccode;

    DBOUT (DBG_LF_ENTRY, "ohci: device_init()\n");

    // Assume failure
    retcode = ENODEV;

    // Is PCI bus present?
    if (!pci_present())
        goto PciNotPresent;

    // Find the OHCI board
    dev = pci_find_device(vendor_id, device_id, dev);
    if(!dev)
        goto DevNotFound;
    DBOUT1 (DBG_LF_INTERNAL, "ohci: vendor_id: 0x%04x \n", dev->vendor);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: device id: 0x%04x \n", dev->device);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: PCI Bus  : 0x%04x \n", dev->bus->number);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: device fn: 0x%04x \n", dev->devfn);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: master   : 0x%04x \n", dev->master);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: irq no   : 0x%04x \n", dev->irq);

    // Read device configuration space to get base addresses, etc.
    ccode = pci_read_config_word(dev,PCI_VENDOR_ID,&vendor_id);
    if (ccode)
        goto PciReadError;
    ccode = pci_read_config_word(dev,PCI_DEVICE_ID,&device_id);
    if (ccode)
        goto PciReadError;
    ccode = pci_read_config_dword(dev,PCI_BASE_ADDRESS_0,&base);
    if (ccode)
        goto PciReadError;
    ccode = pci_read_config_byte(dev,PCI_INTERRUPT_LINE,&irq);
    if (ccode)
        goto PciReadError;
    DBOUT  (DBG_LF_INTERNAL, "ohci: Device Config space read\n");
    DBOUT1 (DBG_LF_INTERNAL, "ohci: vendor: 0x%x\n",dev->vendor);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: base  : 0x%8lx\n",dev->base_address[0]);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: irq   : 0x%4x\n",dev->irq);

    // Calculate mem map alignment offet
    gRegsPageAlignmentOffset = dev->base_address[0] & ~PAGE_MASK;

    // Attach the device
    ccode = device_attach((int)irq);
    if (ccode)
        goto AttachError;

    // OK
    retcode = 0;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohci: device_init() exit, retcode = %08X\n", retcode);
    return retcode;

PciNotPresent:
    DBOUT (DBG_ERROR, "ohci: ERROR: PCI bus not present!\n");
    return -ENODEV;
    goto Exit;

DevNotFound:
    DBOUT (DBG_ERROR, "ohci: ERROR: Device not found!\n");
    return -ENODEV;
    goto Exit;

PciReadError:
    DBOUT1 (DBG_ERROR, "ohci: ERROR: Device found, but unable to read PCI space for device!  ccode = %d.\n", ccode);
    return -ENODEV;
    goto Exit;

AttachError:
    DBOUT1 (DBG_ERROR, "ohci: ERROR: Device found, but unable to allocate system resources for device!  ccode = %d\n", ccode);
    return -ENODEV;
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// device_attach()
//
//      Allocates system resources for the device.  
//
// Called by:
//
//		device_init()
//
// Entry:
//
//      NONE
//
// Exit:
//
//      0       OK.  Driver loaded.
//      ENODEV  No device was detected.  Driver failed to load.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int device_attach(int irq)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int retcode;
    int ccode;
    extern struct ohci_data *ohcidata;
    unsigned long page_base, page_offs, page_remapped;
    unsigned long arsize;
    __u32 base;
    __u32 mask;
    int i;

    DBOUT (DBG_LF_ENTRY, "ohci: device_attach()\n");

    // Assume failure
    retcode = ENODEV;

    // Allocate card control info structure
    ohcidata = kmalloc(sizeof(struct ohci_data), GFP_KERNEL);
    if(!ohcidata)
        goto InfoAllocError;

    // Allocate DMA buffer memory
    dmabuf   = __get_dma_pages(GFP_KERNEL,1);
    if (!dmabuf)
        goto DmaBufAllocError;
    DBOUT1 (DBG_LF_INTERNAL, "ohci: DMA Memory allocated = %08X\n", dmabuf);

    // Map DMA pages
    for (i=MAP_NR (dmabuf);
         i<= MAP_NR(dmabuf+PAGE_SIZE-1);
         i++)
    {
        mem_map_reserve(i);
    }

    // Get the physical address of the DMA buffer
    gDmaPagePhysAddr = virt_to_phys ((void *)dmabuf);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: DMA Memory physical address = %08X\n", gDmaPagePhysAddr);

    // Remap ohci register memory
    base = dev->base_address[0];
    cli();
    pci_write_config_dword(dev,PCI_BASE_ADDRESS_0,~0x0);
    pci_read_config_dword(dev,PCI_BASE_ADDRESS_0,&mask);
    pci_write_config_dword(dev,PCI_BASE_ADDRESS_0,base);
    sti();
    mask &= 0xFFFFF000;
    arsize = (~mask) + 1;
    page_base     = ((unsigned long) base) & PCI_BASE_ADDRESS_MEM_MASK;
    page_offs     = ((unsigned long) base) - page_base;
    page_remapped = (unsigned long)ioremap(page_base,arsize);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: page_base: 0x%08lx \n",page_base);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: page_offs: 0x%08lx \n",page_offs);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: page_remapped: 0x%08lx \n", page_remapped);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: arsize       : 0x%08lx \n", arsize);
    VIR_addr = (unsigned long) (page_remapped ? (page_remapped + page_offs) : 0UL);
    if (!VIR_addr)
        goto MemMapError;
    DBOUT (DBG_LF_INTERNAL, "ohci: MAPPING SUCCESS!!!\n");
    DBOUT1 (DBG_LF_INTERNAL, "ohci: VIR_addr: 0x%08lx \n", VIR_addr);

    // Install the interrupt handler
    ccode = request_irq(irq, ohci_intr, SA_INTERRUPT, "ohci", NULL);
    if (ccode)
        goto IrqReqError;
    ohcidata->irq = irq;
    DBOUT1 (DBG_LF_INTERNAL, "ohci: ohcidata->irq: 0x%x \n",ohcidata->irq);

    // Enable interrupts
    ccode = READREG(0x088);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: INTR MASK(reset): 0x%x\n", ccode);
    ccode = READREG(0x080);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: INTR SET(reset) : 0x%x\n", ccode);
 
    // INTERRUPTS: 1000 0111 0010 1011;;0000 0000 0000 0000
    WREG(0x088, 0x00000005);
    WREG(0x084, 0xFFFFFFFF);
    DBOUT (DBG_LF_INTERNAL, "ohci: INTR: CLEARED ALL\n");
    ccode = READREG(0x088);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: INTR MASK: 0x%x \n", ccode);
    ccode = READREG(0x080);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: INTR SET: 0x%x \n", ccode);

    // Register as a character driver
    ccode = register_chrdev(ohci_major,"ohci", &ohci_fops);
    if (ccode)
        goto RegisterError;
    DBOUT1 (DBG_LF_INTERNAL, "ohci: Major : %d\n", ohci_major);
    DBOUT1 (DBG_LF_INTERNAL, "ohci: Node ID: 0x%x\n", READREG(0x0E8));

    // OK
    retcode = 0;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohci: device_attach() exit, retcode = %08X\n", retcode);
    return retcode;

InfoAllocError:
    DBOUT (DBG_ERROR, "ohci: Failed to allocate card information structure\n");
    retcode = ENODEV;
    goto Exit;

DmaBufAllocError:
    DBOUT (DBG_ERROR, "ohci: Failed to allocate DMA buffer memory\n");
    retcode = ENODEV;
    goto Exit;

MemMapError:
    DBOUT (DBG_ERROR, "ohci: Cannot map memmory mapped I/O region\n");
    retcode = -1;
    goto Exit;

IrqReqError:
    DBOUT1 (DBG_ERROR, "ohci: Request irq %d failed\n", irq);
    retcode = -1;
    goto Exit;

RegisterError:
    DBOUT1 (DBG_ERROR, "ohci: Cannot register char driver, major = %08X\n", ohci_major);
    retcode = -EIO;
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// device_open()
//
//      Entry point to open the ohci device
//
// Called by:
//
//		kernel during open() on /dev/ohci0
//
// Entry:
//
//      inode	inode for this device
//
//      file	file handle for this device
//
// Exit:
//
//      0       OK
//      !0      OK
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static int
device_open(struct inode *inode, struct file *file)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int retcode;
    int num;
    
    DBOUT (DBG_LF_ENTRY, "ohci: device_open()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: inode = %08X\n", inode);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: file = %08X\n", file);

    // Determine which device to open (/dev/ohci0 or /dev/ohci1)
    num = NUM(file->f_dentry->d_inode->i_rdev);
    if (!num)
    {
        // Incrment use count
        MOD_INC_USE_COUNT;
        retcode = 0;
    }
    else
    {
        // Setup return info
        file->f_op = ohci_fop_array[num];
        retcode = file->f_op->open(inode,file);
    }
    DBOUT1 (DBG_LF_INTERNAL, "ohci: num: %d \n",num);

    DBOUT1 (DBG_LF_EXIT, "ohci: device_open(), retcode = %08X\n", retcode);
    return retcode;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// device_release()
//
//      Entry point to release the ohci device
//
// Called by:
//
//		kernel during close() on /dev/ohci0
//
// Entry:
//
//      inode	inode for this device
//
//      file	file handle for this device
//
// Exit:
//
//      0       OK
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static int device_release(struct inode *inode, struct file *file)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int retcode;
    
    DBOUT (DBG_LF_ENTRY, "ohci: device_release()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: inode = %08X\n", inode);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: file = %08X\n", file);

    // Assume default
    retcode = 0;

    // Decrement use count
    MOD_DEC_USE_COUNT;

    DBOUT (DBG_LF_EXIT, "ohci: device_release(), device released\n");
    return retcode;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// device_ioctl()
//
//      General IOCTL handler for the ohci device
//
// Called by:
//
//		kernel during ioctl() on /dev/ohci0 and /dev/ohci1
//
// Entry:
//
//      inode	inode for this device
//
//      file	file handle for this device
//
//      cmd		the ioctl command to be performed
//
//      arg		argument for this command.  Interpretation of 
//				this is command dependent
//
// Exit:
//
//      0       OK.
//      !0      error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static int
device_ioctl(struct inode *inode, struct file *file,
           unsigned int cmd, unsigned long arg)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    extern struct ohci_data *ohcidata;
    int retcode;
    int i;

    DBOUT (DBG_LF_ENTRY, "ohci: device_ioctl()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: inode = %08X\n", inode);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: file = %08X\n", file);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: cmd = %08X\n", cmd);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: arg = %08X\n", arg);

    // Assume error
    retcode = -1;
    i = 0;

    // Process the command
    switch(cmd)
    {

        case OHCIQUADSEND: // send 1 quadlet packet

            DBOUT (DBG_LF_INTERNAL, "ohci: OHCIQUADSEND\n");

            // Copy Descriptor Program
            (unsigned int)ohcidata->ATRQ_program_small = (unsigned int)
                (&ohcidata->AT_programmbuffer + 0x10) & 0xfffffff0;
            while(i<8)
            {
                ohcidata->ATRQ_program_small[i] = AT_DMA_program_small_ROM[i];
                printk("ATRQpgmsm[%d]: %x\n",i,ohcidata->ATRQ_program_small[i]);
                i++;
            };
            DBOUT1 (DBG_LF_INTERNAL, "ohci: &od->ATRQ_pgm_small[0] = %08lx \n",
              (unsigned long)&ohcidata->ATRQ_program_small[0]);

            // Execute DMA Program
            WREG(0x0184,0xFFFFFFFF);
            DBOUT1 (DBG_LF_INTERNAL, "ohci: CONTEXT_SET[180]: %08lx\n",(unsigned long)READREG(0x0180));
            WREG(0x018C, ((unsigned long)virt_to_phys(&ohcidata->ATRQ_program_small[0])|0x2));
            DBOUT1 (DBG_LF_INTERNAL, "ohci: &ATRQ_pgm_sm: %08lx\n",(unsigned long)&ohcidata->ATRQ_program_small[0]);
            DBOUT1 (DBG_LF_INTERNAL, "ohci: CONTEXT_ADDR_REG[18C]: %08lx\n", (unsigned long)READREG(0x018C));

            break;

        case OHCIINFO: 
            DBOUT (DBG_LF_INTERNAL, "ohci: Hello OHCIINFO IOCTL BLOCK..\n");
            break;

        case OHCISEND: 
            WREG(0x0180, 0x00008000);
            DBOUT1 (DBG_LF_INTERNAL, "ohci: OHCISEND[180] : %08lx \n", (unsigned long)READREG(0x0180));
            break;

        case OHCIGETREGPAGEALIGNOFFSET:
            DBOUT (DBG_LF_INTERNAL, "ohci: OHCIGETREGALIGNOFFSET\n");
            DBOUT3 (DBG_LF_INTERNAL, "ohci: copy from kerenl addr %08X value %08X to user addr %08X\n", &gRegsPageAlignmentOffset, gRegsPageAlignmentOffset, arg);
            copy_to_user ((void *)arg, (void *)&gRegsPageAlignmentOffset, sizeof(gRegsPageAlignmentOffset));
            break;

        case OHCIGETDMAPAGEPHYSADDR:
            DBOUT (DBG_LF_INTERNAL, "ohci: OHCIGETDMAPAGEPHYSADDR\n");
            copy_to_user ((void *)arg, (void *)&gDmaPagePhysAddr, sizeof(gDmaPagePhysAddr));
            break;

        default:
            goto InvalidCommand;
            break;
    }

    // OK
    retcode = 0;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohci: device_ioctl() exit, retcode = %08X\n", retcode);
    return retcode;

InvalidCommand:
    DBOUT1 (DBG_ERROR, "ohci: ERROR: Invalid command %08X in device_ioctl()\n", cmd);
    retcode = -EINVAL;
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// device_mmap()
//
//      Memory map entry point
//
// Called by:
//
//		kernel during mmap() on /dev/ohci0
//
// Entry:
//
//      file	file handle for this device
//
//      vma		virtual memory area structure
//
// Exit:
//
//      0       OK.
//      !0      error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static int device_mmap (struct file *file, struct vm_area_struct *vma)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int retcode;
    unsigned long MaxMapSize;

    DBOUT (DBG_LF_ENTRY, "ohci: device_mmap()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: file = %08X\n", file);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: vma = %08X\n", vma);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: vma->vm_start = %08X\n", vma->vm_start);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: vma->vm_end = %08X\n", vma->vm_end);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: vma->vm_offset = %08X\n", vma->vm_offset);

    // Validate parameters
    if (gRegsPageAlignmentOffset)
        MaxMapSize = 4096 *2;
    else
        MaxMapSize = 4096;
    if ((vma->vm_end - vma->vm_start + vma->vm_offset) > MaxMapSize)
        goto InvalidParam;

    // Recalculate offset
    vma->vm_offset += dev->base_address[0];
    vma->vm_offset &= PAGE_MASK;
    DBOUT1(DBG_LF_INTERNAL, "ohci: new vma->vm_offset = %08X\n", vma->vm_offset);

    // Ignore page alignment issue since app will take this into
    // consideration by reading the page alignment variable
    // and adding it to the returned address, gRegsPageAlignmentOffset.
//  if (vma->vm_offset & ~PAGE_MASK)
//      goto BoundaryError;

    // Do some 486 adjustments
    #ifdef __i486__
    if ((int)vma->vm_offset >= (int)high_memory)
        pgprot_val (vma->vm_page_prot) |= _PAGE_PCD;
    #endif

    // Remap page range
    if (remap_page_range (  vma->vm_start,
                            vma->vm_offset,
                            (vma->vm_end - vma->vm_start),
                            vma->vm_page_prot))
        goto RemapError;

    // OK
    retcode = 0;

Exit:

    DBOUT1 (DBG_LF_EXIT, "ohci: device_mmap() exit, retcode = %08X\n", retcode);
    return retcode;

InvalidParam:
    DBOUT1 (DBG_ERROR, "ohci: ERROR: Invalid parameter: vma = %08X\n", vma);
    retcode = -EINVAL;
    goto Exit;

//BoundaryError:
//  DBOUT1 (DBG_ERROR, "ohci: ERROR: value not on page boundary, vma->vm_offset = %08X\n", vma->vm_offset);
//  retcode = -ENXIO;
//  goto Exit;

RemapError:
    DBOUT (DBG_ERROR, "ohci: ERROR: cannot remap page range\n");
    retcode = -EAGAIN;
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ohci_intr()
//
//      ISR for the ohci device
//
// Called by:
//
//		kernel during an interrupt
//
// Entry:
//
//      irq		IRQ handle
//
//      dev_id	interrupting device
//
//      pt_regs	processor registers
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static void ohci_intr(int irq, void *dev_id, struct pt_regs *regs)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    struct timeval tv;

    DBOUT (DBG_LF_ENTRY, "ohci: ohci_intr()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: irq = %08X\n", irq);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: dev_id = %08X\n", dev_id);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: regs = %08X\n", regs);

    DBOUT1(DBG_LF_INTERNAL, "ohci: INTR MASK: 0x%x\n", READREG(0x088));

    /* INTERRUPTS: 1000 0111 0010 1011;;0000 0000 0000 0000 */
    // printk("-- INTR SETTING \n");
    // WREG(0x088, 0x072B0000);
    // WREG(0x088, 0x00000001);
    // printk("--INTR MASK: 0x%x \n", READREG(0x088));

    // WREG(0x084, 0xFFFFFFFF);
    DBOUT(DBG_LF_INTERNAL, "ohci: INTR: CLEARED ALL\n");

    /* TURN-OFF INTERRUPTS */
    WREG(0x084, 0x00000005);
    DBOUT(DBG_LF_INTERNAL, "ohci: CLEARED INTERRUPT\n");
    DBOUT1(DBG_LF_INTERNAL, "ohci: INTR SET[80]: 0x%x \n", READREG(0x080));

    do_gettimeofday(&tv);

    DBOUT (DBG_LF_EXIT, "ohci: ohci_intr() exit\n");

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// device_dma_mmap()
//
//      handled dma memory map
//
// Called by:
//
//		kernel during mmap() on /dev/ohci1
//
// Entry:
//
//      file	file handle for this device
//
//      vma		virtual memory area structure
//
// Exit:
//
//      0       OK.
//      !0      error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int device_dma_mmap(struct file *file, struct vm_area_struct * vma)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int retcode;
    int ccode;
    extern unsigned long dmabuf;

    DBOUT (DBG_LF_ENTRY, "ohci: device_dma_mmap()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: file = %08X\n", file);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: vma = %08X\n", vma);

    // Assume failure
    retcode = -1;

    // Validate parameters
    if(dmabuf & ~PAGE_MASK)
        goto InvalidParameter;

    // Remap
    ccode = remap_page_range (
                vma->vm_start,
                virt_to_phys((unsigned long *)dmabuf),
                vma->vm_end - vma->vm_start,
                vma->vm_page_prot);
    if (ccode)
        goto RemapError;

    // Increment count
    file->f_dentry->d_inode->i_count++;

    // OK
    retcode = 0;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohci: device_dma_mmap() exit, retcode = %08X\n", retcode);
    return retcode;

InvalidParameter:
    DBOUT (DBG_ERROR, "ohci: ERROR: DMABUF value no on page boundary");
    retcode = -ENXIO;
    goto Exit;

RemapError:
    DBOUT (DBG_ERROR, "ohci: ERROR: Remap failed!\n");
    retcode = -EAGAIN;
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// device_dma_open()
//
//      opens the DMA
//
// Called by:
//
//		kernel during open() on /dev/ohci1
//
// Entry:
//
//      inode	inode for this device
//
//      file	file handle for this device
//
// Exit:
//
//      0       OK.
//      !0      error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int device_dma_open(struct inode * inode, struct file *file)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int retcode;

    DBOUT (DBG_LF_ENTRY, "ohci: device_dma_open()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: inode = %08X\n", inode);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: file = %08X\n", file);

    // Assume failure
    retcode = -1;

    // Increment count
    MOD_INC_USE_COUNT;

    // DO WHATEVER ELSE

    // OK
    retcode = 0;

    DBOUT1 (DBG_LF_EXIT, "ohci: device_dma_open(), retcode = %08X\n", retcode);
    return retcode;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// device_dma_release()
//
//      opens the DMA
//
// Called by:
//
//		kernel during close() on /dev/ohci1
//
// Entry:
//
//      inode	inode for this device
//
//      file	file handle for this device
//
// Exit:
//
//      0       OK.
//      !0      error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int device_dma_release(struct inode * inode, struct file *file)
{
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugThisRoutine

    int retcode;

    DBOUT (DBG_LF_ENTRY, "ohci: device_dma_open()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: inode = %08X\n", inode);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohci: file = %08X\n", file);

    // Assume failure
    retcode = -1;

    // Decrement count
    MOD_DEC_USE_COUNT;

    // DO WHATEVER ELSE

    // OK
    retcode = 0;

    DBOUT1 (DBG_LF_EXIT, "ohci: device_dma_release(), retcode = %08X\n", retcode);
    return retcode;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}
