//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  IEEE-1394 OHCI Test Application for Linux
//
//  2.2.9 KERNEL
//
//  Original Author: Bill Alexander (www.geocities.com/SiliconValley/Haven/4824)
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

//-------------------------------------------
// Standard include files
//-------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <asm/io.h>
#include <linux/ioctl.h>
#include <linux/errno.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <sys/times.h>
#include <syslog.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <ctype.h>

//-------------------------------------------
// My debug definitions
//-------------------------------------------
// Uncomment the DEBUG definition if you want debug traces
//#define DEBUG
#define _DBG_APP_
#define _DBG_LINUX_
#define _DBG_APP_MAIN_MODULE_
#include "dbg.h"

//-------------------------------------------
// My 1394/ohci definitions
//-------------------------------------------
#define _APP_MAIN_MODULE_
#include "ohci.h"

//-------------------------------------------
// Global variables
//-------------------------------------------

// Program path
char gProgramPath[500];

// Handle to ohci.o module
int gOhciFd;
int gOhciDmaFd;

// OHCI Register Memory Map pointer - so we can directly access the hardware from here
QUADLET *gOhciRegMemMap;
__u32 gOhciRegMemMapAlignmentOffset;
__u32 gOhciRegMemMapLength;

// OHCI DMA Memory Map pointer - so we can program DMA directly from here
QUADLET *gOhciDmaMemMap;
__u32 gOhciDmaMemMapLength;
__u32 gOhciDmaMemPhysAddr;

// Structure to organize DMA memory map
typedef struct MyDmaStruct
{
    // 1st 2K

    // Self IDs from bus resets
    SelfIDReceive   SelfIds;
    BYTE            Pad[(2*1024) - sizeof(SelfIDReceive)];

    // 2nd 2K

    // Dynamic DMA memory allocation starting point
    BYTE            DynamicMemory[2*1024];

} MyDmaT, *PMyDmaT;

// Dynamic DMA Memory Management structure
typedef struct DmaBlockHandleStruct *PDmaBlockHandleT;
typedef struct DmaBlockHandleStruct
{
    PDmaBlockHandleT    Next;
    BOOLEAN             IsFree;
    __u32               BlockAddr;
    __u32               BlockSize;
    __u32               RealAllocedSize;
    __u32               PhysAddr;
} DmaBlockHandleT;

// Dynamic DMA Memory Managment Lists
PDmaBlockHandleT gDynamicDmaMemoryAllocedHead;
PDmaBlockHandleT gDynamicDmaMemoryFreeHead;

// DMA Node
typedef struct DmaNodeStruct *PDmaNodeT;
typedef struct DmaNodeStruct
{
    PDmaNodeT           Next;
    PDmaBlockHandleT    DescBlkHandle;
    PDmaBlockHandleT    DataBlkHandle;
} DmaNodeT;

// Async request transmit DMA
PDmaNodeT gAsyncRequestTransmitDmaList;

// Async response transmit DMA
PDmaNodeT gAsyncResponseTransmitDmaList;

// Async request receive DMA
PDmaNodeT gAsyncRequestReceiveDmaList;

// Async response receive DMA
PDmaNodeT gAsyncResponseReceiveDmaList;

// Isoch transmit DMAs
#define MAX_ISOCH_DMA_CONTEXTS  32
PDmaNodeT gIsochTransmitDmaList[MAX_ISOCH_DMA_CONTEXTS];

// Isoch receive DMAs
PDmaNodeT gIsochReceiveDmaList[MAX_ISOCH_DMA_CONTEXTS];

// Configuration ROM Buffer
#define CONFIG_ROM_MAX_SIZE 0x2024
QUADLET gConfigRomBuf[CONFIG_ROM_MAX_SIZE/sizeof(QUADLET)];
__u32 gConfigRomBufSize = 0;

// Command variables and structures
typedef struct ExitCommandStruct
{
    int                 Reserved;
} ExitCommandT, *PExitCommandT,
  HelpCommandT, *PHelpCommand,
  CommentCommandT, *PCommentCommand,
  InitCardCommandT, *PInitCardCommandT,
  InitSelfIdCommandT, *PInitSelfIdCommandT,
  BusResetCommandT, *PBusResetCommandT,
  InitArDmaCommandT, *PInitArDmaCommandT,
  InitAtDmaCommandT, *PInitAtDmaCommandT;

typedef struct ReadCommandStruct
{
    int                 Offset;
    int                 ValueRead;
} ReadCommandT, *PReadCommandT,
  ReadDmaMemCommandT, *PReadDmaMemCommandT;

typedef struct WriteCommandStruct
{
    int                 Offset;
    int                 ValueToWrite;
} WriteCommandT, *PWriteCommandT,
  WriteDmaMemCommandT, *PWriteDmaMemCommandT;

#define MAX_ECHO_BUFFER     256

typedef struct EchoCommandStruct
{
    char                EchoBuffer[MAX_ECHO_BUFFER];
} EchoCommandT, *PEchoCommandT;

#ifdef DEBUG
typedef struct DebugFlagsCommandStruct
{
    BOOLEAN             PromptForValue;
    int                 DebugFlagsValue;
} DebugFlagsCommandT, *PDebugFlagsCommandT;
#endif

typedef struct ReadQuadletCommandStruct
{
    __u32           NodeId;
    __u32           AddrOffsetHigh;
    __u32           AddrOffsetLow;
} ReadQuadletCommandT, *PReadQuadletCommandT;

typedef struct DumpRomCommandStruct
{
    __u32           NodeId;
} DumpRomCommandT, *PDumpRomCommandT;

#define PAUSE_SECONDS		1
#define PAUSE_MILLISECONDS	2
#define PAUSE_MICROSECONDS	3
typedef struct PauseCommandStruct
{
    __u32           SecondsType;
	__u32			Seconds;
} PauseCommandT, *PPauseCommandT;

typedef enum CommandTypeEnum
{
    CMD_EXIT,
    CMD_READ,
    CMD_WRITE,
    CMD_HELP,
    CMD_COMMENT,
    CMD_ECHO,
    CMD_READ_DMA_MEM,
    CMD_WRITE_DMA_MEM,
    CMD_INIT_CARD,
    CMD_INIT_SELF_ID,
    CMD_BUS_RESET,
#ifdef DEBUG
    CMD_DEBUG_FLAGS,
#endif
    CMD_INIT_AR_DMA,
    CMD_INIT_AT_DMA,
    CMD_READ_QUADLET,
    CMD_DUMP_ROM,
    CMD_PAUSE,
} CommandTypeT, *PCommandTypeT;

typedef struct CommandAssociationStruc
{
    char *          CmdString;
    CommandTypeT    CmdType;
} CommandAssociationT, *PCommandAssociationT;

CommandAssociationT gCmdAssoc[] = 
{
    {"E",                   CMD_EXIT},
    {"EXIT",                CMD_EXIT},
    {"Q",                   CMD_EXIT},
    {"QUIT",                CMD_EXIT},
    {"R",                   CMD_READ},
    {"READ",                CMD_READ},
    {"RD",                  CMD_READ},
    {"W",                   CMD_WRITE},
    {"WR",                  CMD_WRITE},
    {"WRITE",               CMD_WRITE},
    {"?",                   CMD_HELP},
    {"H",                   CMD_HELP},
    {"HELP",                CMD_HELP},
    {"#",                   CMD_COMMENT},
    {";",                   CMD_COMMENT},
    {"ECHO",                CMD_ECHO},
    {"RDM",                 CMD_READ_DMA_MEM},
    {"WDM",                 CMD_WRITE_DMA_MEM},
    {"InitCard",            CMD_INIT_CARD},
    {"InitSelfId",          CMD_INIT_SELF_ID},
    {"InitArDma",           CMD_INIT_AR_DMA},
    {"InitAtDma",           CMD_INIT_AT_DMA},
    {"ReadQuadlet",         CMD_READ_QUADLET},
    {"RQ",                  CMD_READ_QUADLET},
    {"BusReset",            CMD_BUS_RESET},
    {"DumpRom",             CMD_DUMP_ROM},
    {"DR",                  CMD_DUMP_ROM},
    {"PAUSE",               CMD_PAUSE},
#ifdef DEBUG
    {"DebugFlags",          CMD_DEBUG_FLAGS},
#endif
};

typedef struct CommandStuct
{
    CommandTypeT                CmdType;
    union
    {
        ExitCommandT            ExitCmd;
        ReadCommandT            ReadCmd;
        WriteCommandT           WriteCmd;
        CommentCommandT         CommentCmd;
        HelpCommandT            HelpCmd;
        EchoCommandT            EchoCmd;
        ReadDmaMemCommandT      ReadDmaMemCmd;
        WriteDmaMemCommandT     WriteDmaMemCmd;
        InitCardCommandT        InitCardCmd;
        InitSelfIdCommandT      InitSelfIdCmd;
        InitArDmaCommandT       InitArDmaCmd;
        InitAtDmaCommandT       InitAtDmaCmd;
        ReadQuadletCommandT     ReadQuadletCmd;
        DumpRomCommandT         DumpRomCmd;
        BusResetCommandT        BusResetCmd;
        PauseCommandT			PauseCmd;
#ifdef DEBUG
        DebugFlagsCommandT      DebugFlagsCmd;
#endif
    }u;
} CommandT, *PCommandT;

//-------------------------------------------
// Interrupt Event Thread Variables
//-------------------------------------------
int gEventProcessPid;
int gSharedMemId;
#define SHARED_MEM_KEY 0x1234
typedef struct ParentPausedOnEventStruct
{
    int     Signal;
    QUADLET EventWaitingFor;
} ParentPausedOnEventT, *PParentPausedOnEventT;
typedef struct SharedMemStruct
{
    BOOLEAN                 Exit;
    BOOLEAN                 IgnoreEvents;
    ParentPausedOnEventT    ParentPausedOnEvent;
} SharedMemT, *PSharedMemT;
PSharedMemT gSharedMem;

//-------------------------------------------
// Function Prototypes
//-------------------------------------------
int main (int argc, char *argv[]);
BOOLEAN ParseCommandLine (int argc, char **argv);
void Usage (void);
void EventProcess (void);
void HandleBusReset (void);
void DoBusReset (BOOLEAN WantToBeRoot);
void SignalEvent (__u32 sig);
void ParentSignalHandler (int sig);
BOOLEAN PauseOnEvent (__u32 Event, __u32 TimeOutSec);
void DisplayMenuAndExecute (void);
void DisplayMenu (void);
BOOLEAN GetCommand (PCommandT cmd);
BOOLEAN ParseCommandString (char *CmdStr, PCommandT cmd);
BOOLEAN ExecuteCommand (PCommandT cmd);
#ifdef DEBUG
void DisplayModifyDebugFlags (BOOLEAN PromptForValue, int DebugFlagsValue);
#endif
BOOLEAN InitCard (void);
BOOLEAN InitSelfId (void);
BOOLEAN BusReset (void);
BOOLEAN InitArDma (void);
BOOLEAN InitAtDma (void);
BOOLEAN ReadQuadlet (
            __u32 NodeId, 
            __u32 AddrOffsetHigh, 
            __u32 AddrOffsetLow,
            QUADLET *QuadletValue);
BOOLEAN DmaAllocateBlock (__u32 BlockSize, PDmaBlockHandleT *DmaBlockHandle);
BOOLEAN DmaCreateBlock (__u32 BlockAddr, __u32 BlockSize, __u32 PhysAddr);
BOOLEAN DmaFreeBlock (PDmaBlockHandleT DmaBlockHandle);
BOOLEAN DmaCoalesceWithNextBlock (PDmaBlockHandleT DmaBlockHandle);
BOOLEAN DmaAllocDmaNode (
            __u32 DescriptorSize,
            __u32 DataSize,
            PDmaNodeT *DmaNode);
BOOLEAN DmaAddDmaNodeToList (
            PDmaNodeT *DmaListHead,
            PDmaNodeT DmaNode);
BOOLEAN DmaAllocAndAddDmaNodeToList (
            PDmaNodeT *DmaListHead,
            __u32 DescriptorSize,
            __u32 DataSize,
            PDmaNodeT *DmaNode);
BOOLEAN DmaFreeDmaNode (
            PDmaNodeT DmaNode);
BOOLEAN DmaRemoveDmaNodeFromList (
            PDmaNodeT *DmaListHead,
            PDmaNodeT DmaNode);
BOOLEAN DmaRemoveAndFreeDmaNodeFromList (
            PDmaNodeT *DmaListHead,
            PDmaNodeT DmaNode);
BOOLEAN DmaDestroyNodeList (PDmaNodeT DmaListHead);
__u32 GetNextTLabel (void);
void HostToNetU32Buffer (__u32 *Buffer, int Size);
void NetToHostU32Buffer (__u32 *Buffer, int Size);
BOOLEAN ReadConfigRom (
            __u32 NodeId, 
            QUADLET *ConfigRomBuf, 
            __u32 ConfigRomBufMaxSize, 
            __u32 *BytesRead);
BOOLEAN DumpConfigRom (
            QUADLET *CfgRomBuf, 
            __u32 CfgRomBufSize);
BOOLEAN DecodeConfigRom (
            QUADLET * CfgRomBuf, 
            __u32 CfgRomBufSize);
void DumpQuadletAsChars (
            QUADLET q);
char * KeyToString (
            PDirEntryT DirEntry, 
            __u32 UnitSpecId, 
            __u32 UnitSwVersion);
char * ImmediateValueToString (
            PDirEntryT DirEntry, 
            __u32 UnitSpecId, 
            __u32 UnitSwVersion);
void MyPause (__u32 SecondsType, __u32 Seconds);

int ioctl (int fd, int cmd, ...);

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// main()
//
//      Entry point to the program
//
// Entry:
//
//      argc    number of arguments in argv
//
//      argv    array of strings representing command line arguments
//
// Exit:
//
//      0   success
//      !0  error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
int main (int argc, char *argv[])
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    int retcode;
    int ccode;
    BOOLEAN bDrvPreviouslyLoaded;
    BOOLEAN bDrvLoaded;
    BOOLEAN ccode2;
    int i;

    // Initialize debug module
    DBGInitialize (DBG_DEFAULT);

    DBOUT (DBG_LF_ENTRY, "ohcitest: main()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: argc = %08X\n", argc);
    #ifdef DEBUG
    {
        int i;
        for (i=0; i<argc; i++)
            DBOUT2(DBG_LF_ENTRY_PARAMS, "ohcitest: argv[%d] = %s\n", i, argv[i]);
    }
    #endif

    // Assume failure
    retcode = -1;
    bDrvPreviouslyLoaded = FALSE;
    bDrvLoaded = FALSE;

    // Init global variables
    gOhciFd = -1;
    gOhciDmaFd = -1;
    gOhciRegMemMap = NULL;
    gEventProcessPid = -1;
    gSharedMemId = -1;
    gSharedMem = NULL;
    gDynamicDmaMemoryFreeHead = NULL;
    gDynamicDmaMemoryAllocedHead = NULL;
    gAsyncRequestTransmitDmaList = NULL;
    gAsyncResponseTransmitDmaList = NULL;
    gAsyncRequestReceiveDmaList = NULL;
    gAsyncResponseReceiveDmaList = NULL;
    for (i=0; i<MAX_ISOCH_DMA_CONTEXTS; i++)
    {
        gIsochTransmitDmaList[i] = NULL;
        gIsochReceiveDmaList[i] = NULL;
    }

    // Parse command line
    ccode = ParseCommandLine (argc, argv);
    if (ccode2 != TRUE)
        goto InvalidCommandLine;

    // Load ohci.o
    // NOTE:  Matt Pujol says this is naughty.  You should login as root to
    // to load the module, logout, and then run the app as a normal user.  
    // FINE! unless you make changes to both driver and app and want to see
    // the results sometime this century.  THIS IS AN INITIAL DEVELOPMENT 
    // where we are experimenting with the hardware and software.  Kick-butt
    // super programmers do everything as root, don't they?!  In production 
    // software, you would never do this.
	ccode = getuid();
	if (ccode != 0)
		goto NotRoot;
    ccode = system ("insmod ohci.o");
    if (ccode)
    {
        printf ("Unable to load ohci.o module.  Assuming already loaded.\n");
        if (errno)
            printf ("Last error = %08x = %s\n", errno, strerror(errno));
        bDrvPreviouslyLoaded = TRUE;
    }
    else
    {
		char str[80];
		sprintf (str, "mknod /dev/ohci0 c %d 0", OHCI_MAJOR_DEV_NUM);
        system (str);
        sprintf (str, "mknod /dev/ohci1 c %d 1", OHCI_MAJOR_DEV_NUM);
        system (str);
        bDrvPreviouslyLoaded = FALSE;
    }
    bDrvLoaded = TRUE;
        
    // Open the ohci module
    gOhciFd = open ("/dev/ohci0",O_RDWR);
    if (gOhciFd < 0 )
        goto OpenError;

    // Read the OHCI register memory map alignment offset
    ccode = ioctl (gOhciFd, OHCIGETREGPAGEALIGNOFFSET, &gOhciRegMemMapAlignmentOffset);
    if (ccode)
        goto AlignIoctlError;

    // Calculate the number of pages for the memory map
    if (gOhciRegMemMapAlignmentOffset)
        gOhciRegMemMapLength = 4096*2;
    else
        gOhciRegMemMapLength = 4096;

    // Map the ohci register memory into application space so we can directly
    // control the hardware from here.
    gOhciRegMemMap = mmap (0, gOhciRegMemMapLength, PROT_READ|PROT_WRITE, MAP_SHARED, gOhciFd,0);
    if ((long)gOhciRegMemMap == -1)
        goto MapRegError;

    // Adjust the mem map by the alignment offset
    gOhciRegMemMap = (QUADLET *)((__u32)gOhciRegMemMap + gOhciRegMemMapAlignmentOffset);
    printf ("OHCI Register Memory Map = %08X\n", (unsigned)gOhciRegMemMap);

    // Open the ohci dma device - trick so that we can access dma from here
    gOhciDmaFd = open ("/dev/ohci1",O_RDWR);
    if (gOhciDmaFd < 0 )
        goto OpenDmaError;

    // Map the ohci dma memory into application space
    gOhciDmaMemMapLength = 4096;
    gOhciDmaMemMap = mmap (0, gOhciDmaMemMapLength, PROT_READ|PROT_WRITE, MAP_SHARED, gOhciDmaFd,0);
    if ((long)gOhciDmaMemMap == -1)
        goto MapDmaError;
    printf ("OHCI DMA Memory Map = %08X\n", (unsigned)gOhciDmaMemMap);

    // Get the physical address of the ohci DMA memory page
    ccode = ioctl (gOhciFd, OHCIGETDMAPAGEPHYSADDR, &gOhciDmaMemPhysAddr);
    if (ccode)
        goto PhysDmaIoctlError;
    printf ("OHCI DMA Memory Physical Address = %08X\n", (unsigned)gOhciDmaMemPhysAddr);

    // Initialize Dymanic DMA allocation
    DmaCreateBlock (
        (__u32)&((PMyDmaT)gOhciDmaMemMap)->DynamicMemory, 
        2*1024, 
        gOhciDmaMemPhysAddr+offsetof(MyDmaT,DynamicMemory));

    // Catch the signal from the child event process
    signal (SIGUSR1, ParentSignalHandler);

    // Allocate shared memory to allow parent/child communication
    gSharedMemId = shmget (SHARED_MEM_KEY, sizeof(SharedMemT), 0666 | IPC_CREAT);
    if (gSharedMemId == -1)
        goto SharedMemAllocError;

    // Attach to the shared memory
    // NOTE:  Attach does not carry over a fork().  The child process must attach again.
    gSharedMem = shmat (gSharedMemId, (void *)0, 0);
    if (gSharedMem == NULL)
        goto SharedMemAttachError;

    // Allow the child to process events once it gets started
    gSharedMem->Exit = FALSE;
    gSharedMem->IgnoreEvents = FALSE;

    // Start the Child Event Process to poll the IntEvent register
    gEventProcessPid = fork ();
    if (gEventProcessPid == 0)      // Child process start here!  Are we the child?
    {
		// CHILD PROCESS!

        // Attach to the shared memory
        // NOTE:  Attach does not carry over a fork().  The child process must attach again.
        gSharedMem = shmat (gSharedMemId, (void *)0, 0);
        if (gSharedMem == NULL)
            goto ChildSharedMemAttachError;

        // Raise the priority of the child by a notch
        setpriority (PRIO_PROCESS, getpid(), getpriority(PRIO_PROCESS,getpid())-1);

        // Process all events
        EventProcess ();

        // Child process will exit
        exit (0);
    }
    if (gEventProcessPid == -1)
        goto ForkError;

    // Display the menu and execute commands
    DisplayMenuAndExecute ();

    // OK
    retcode = 0;

Exit:

    // Deallocate resources
    signal (SIGUSR1, SIG_DFL);
    if (gEventProcessPid > 0)
    {
        // Tell child to exit
        gSharedMem->Exit = TRUE;
        sleep (1);

        // Can use the kill signal instead of shared memory
        // signal.
//      kill (gEventProcessPid, SIGKILL);
    }
    if (gAsyncRequestTransmitDmaList != NULL)
        DmaDestroyNodeList (gAsyncRequestTransmitDmaList);
    if (gAsyncResponseTransmitDmaList != NULL)
        DmaDestroyNodeList (gAsyncResponseTransmitDmaList);
    if (gAsyncRequestReceiveDmaList != NULL)
        DmaDestroyNodeList (gAsyncRequestReceiveDmaList);
    if (gAsyncResponseReceiveDmaList != NULL)
        DmaDestroyNodeList (gAsyncResponseReceiveDmaList);
    if (gSharedMem != NULL)
        shmdt (gSharedMem);
    if (gSharedMemId)
        shmctl (gSharedMemId, IPC_RMID, 0);
    if (gOhciRegMemMap != NULL && (long)gOhciRegMemMap != -1)
    {
        // Adjust the mem map by the alignment offset
        gOhciRegMemMap = (QUADLET *)((__u32)gOhciRegMemMap - gOhciRegMemMapAlignmentOffset);

        // Unmap register space from user space
        munmap (gOhciRegMemMap, gOhciRegMemMapLength);

        // Unmap dma space from user space
        munmap (gOhciDmaMemMap, gOhciDmaMemMapLength);
    }
    if (gOhciFd != -1)
        close (gOhciFd);
    if (gOhciDmaFd != -1)
        close (gOhciDmaFd);
    if (bDrvLoaded == TRUE && bDrvPreviouslyLoaded == FALSE)
    {
        system ("rmmod ohci");
        system ("rm -f /dev/ohci0");
        system ("rm -f /dev/ohci1");
    }

    DBOUT1 (DBG_LF_EXIT, "ohcitest: main() exit, retcode = %08X\n", retcode);
    return retcode;

InvalidCommandLine:
    Usage ();
    retcode = -1;
    goto Exit;

NotRoot:
    printf ("ERROR: you are not root.  Your user id = %d (should be 0)\n", ccode);
    retcode = -1;
    goto Exit;

OpenError:
    printf ("ERROR: Open of device returned %d\n", gOhciFd);
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -ENODEV;
    goto Exit;

AlignIoctlError:
    printf ("ERROR: IOCTL OHCIGETREGPAGEALIGNOFFSET to device returned %d\n", ccode);
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -ENODEV;
    goto Exit;

MapRegError:
    printf ("ERROR: Unable to map OHCI regsiter memory into application space\n");
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -ENODEV;
    goto Exit;

OpenDmaError:
    printf ("ERROR: Open of DMA device returned %d\n", gOhciDmaFd);
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -ENODEV;
    goto Exit;

MapDmaError:
    printf ("ERROR: Unable to map OHCI DMA memory into application space\n");
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -ENODEV;
    goto Exit;

PhysDmaIoctlError:
    printf ("ERROR: IOCTL OHCIGETDMAPAGEPHYSADDR to device returned %d\n", ccode);
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -ENODEV;
    goto Exit;

SharedMemAllocError:
    printf ("ERROR: Unable to allocate shared memory, return code = %d\n", gSharedMemId);
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -1;
    goto Exit;

SharedMemAttachError:
    printf ("ERROR: Unable to attach to shared memory\n");
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -1;
    goto Exit;

ChildSharedMemAttachError:
    printf ("ERROR: Child process was unable to attach to shared memory\n");
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -1;
    goto Exit;

ForkError:
    printf ("ERROR: Unable to fork a child process, return code = %d\n", gEventProcessPid);
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    retcode = -1;
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ParseCommandLine
//
//      Parses the command line for options
//
// Entry:
//
//      argc    number of arguments in argv
//
//      argv    array of strings representing command line arguments
//
// Exit:
//
//      TRUE    Command line valid and successfully parsed
//      FALSE   Invalid command line        
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN ParseCommandLine (int argc, char **argv)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    int curArg;
    char *p;

    DBOUT (DBG_LF_ENTRY, "ohcitest: ParseCommandLine()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: argc = %08X\n", argc);
    #ifdef DEBUG
    {
        int i;
        for (i=0; i<argc; i++)
            DBOUT2(DBG_LF_ENTRY_PARAMS, "ohcitest: argv[%d] = %s\n", i, argv[i]);
    }
    #endif

    // Assume failure
    retcode = FALSE;

    // Parse each argument one at a time
    for (curArg = 0, p = argv[curArg]; curArg < argc; curArg++, p++)
    {
        DBOUT2 (DBG_LF_INTERNAL, "curArg = %d, %s\n", curArg, p);

        // Parse current argument
        switch (p[0])
        {
            case '-':
                switch (p[1])
                {
                    default:
                        goto InvalidCommandLine;
                        break;
                }
                break;
            default:
                switch (curArg)
                {
                    case 0:
                        strcpy (gProgramPath, p);
                        break;
                    default:
                        goto InvalidCommandLine;
                        break;
                }
                break;
        }
    }

    // OK
    retcode = TRUE;

Exit:

    DBOUT1 (DBG_LF_EXIT, "ohcitest: ParseCommandLine() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

InvalidCommandLine:
    printf ("ERROR: Invalid command line\n");
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags

}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Usage
//
//      Print the usage of the program
//
// Entry:
//
//      NONE
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void Usage (void)
{
    printf ("\n");
    printf ("\n");
    printf ("Usage:\n");
    printf ("\n");
    printf ("ohcitest\n");
    printf ("\n");
    printf ("\n");
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// EventProcess()
//
//      Called exclusively by the child process to poll the IntEvent register.
//      This is the equivalent of an ISR in a device driver.
//
// Entry:
//
//      NONE
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void EventProcess (void)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    QUADLET intEvent;
    time_t lastWorkDone;
    time_t curTime;
    QUADLET intEventsToProcess;
    #define EVENT_BANNER "**************** EVENT OCCURRED *****************\nEvent ="

    DBOUT (DBG_LF_ENTRY, "ohcitest: EventProcess()\n");

    // Determine which events we really want to process
    intEventsToProcess = 
        reqTxComplete_BIT |
        respTxComplete_BIT |
        ARRQ_BIT |
        ARRS_BIT |
        RQPkt_BIT |
        RSPkt_BIT |
        isochTx_BIT |
        isochRx_BIT |
        postedWriteErr_BIT |
        lockRespErr_BIT |
        selfIDComplete_BIT |
        busReset_BIT |
        phy_BIT |
//      cycleSynch_BIT |
//      cycle64Seconds_BIT |
//      cycleLost_BIT |
//      cycleInconsistent_BIT |
        unrecoverableError_BIT |
//      cycleTooLong_BIT |
        phyRegRcvd_BIT |
        vendorSpecific_BIT;

    // Process events until parent tells us to stop
    lastWorkDone = time (NULL);
    curTime = time (NULL);
    while (gSharedMem->Exit == FALSE)
    {
        // Are we allowed to process events?
        if (gSharedMem->IgnoreEvents == FALSE)
        {
            // Read the current IntEvent register
            intEvent = REG_READ (IntEventSet);

            // Mask out events we don't want
            intEvent &= intEventsToProcess;
        }
        else
        {
            // No events this time around
            intEvent = 0;
        }

        // Process all events
        while (intEvent != 0)
        {
            DBOUT1 (DBG_HF_INTERNAL, "ohcitest: intEvent = %08X\n", intEvent);

            // Process events
            if (((IntEventSet_REG *)&intEvent)->fld.ARRQ)
            {
                printf ("%s ARRQ\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (ARRQ_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.ARRS)
            {
                printf ("%s ARRS\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (ARRS_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.busReset)
            {
                printf ("%s busReset\n", EVENT_BANNER);

                // Process busReset event
                HandleBusReset ();

                // Tell parent about the event
                SignalEvent (busReset_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.cycle64Seconds)
            {
                printf ("%s cycle64Seconds\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (cycle64Seconds_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.cycleInconsistent)
            {
                printf ("%s cycleInconsistent\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (cycleInconsistent_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.cycleLost)
            {
                printf ("%s cycleLost\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (cycleLost_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.cycleSynch)
            {
                printf ("%s cycleSynch\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (cycleSynch_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.cycleTooLong)
            {
                printf ("%s cycleTooLong\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (cycleTooLong_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.isochRx)
            {
                printf ("%s isochRx\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (isochRx_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.isochTx)
            {
                printf ("%s isochTx\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (isochTx_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.lockRespErr)
            {
                printf ("%s lockRespErr\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (lockRespErr_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.phy)
            {
                printf ("%s phy\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (phy_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.phyRegRcvd)
            {
                printf ("%s phyRegRcvd\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (phyRegRcvd_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.postedWriteErr)
            {
                printf ("%s postedWriteErr\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (postedWriteErr_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.reqTxComplete)
            {
                printf ("%s reqTxComplete\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (reqTxComplete_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.respTxComplete)
            {
                printf ("%s respTxComplete\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (respTxComplete_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.RQPkt)
            {
                printf ("%s RQPkt\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (RQPkt_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.RSPkt)
            {
                printf ("%s RSPkt\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (RSPkt_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.selfIDComplete)
            {
                printf ("%s selfIDComplete\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (selfIDComplete_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.unrecoverableError)
            {
                printf ("%s unrecoverableError\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (unrecoverableError_BIT);
            }
            if (((IntEventSet_REG *)&intEvent)->fld.vendorSpecific)
            {
                printf ("%s vendorSpecific\n", EVENT_BANNER);

                // Tell parent about the event
                SignalEvent (vendorSpecific_BIT);
            }

            // Clear events
            REG_WRITE (IntEventClear, intEvent);

            // Update timer
            lastWorkDone = time (NULL);

            // Read the current IntEvent register
            intEvent = REG_READ (IntEventSet);

            // Mask out events we don't want
            intEvent &= intEventsToProcess;
        }

        // Determine how long to wait before waking up
        curTime = time (NULL);
        DBOUT1 (DBG_HF_INTERNAL, "time diff = %d\n", curTime - lastWorkDone);
        if ((curTime - lastWorkDone) < 2)
        {
            // Wait 1 ms
            usleep (1000);
        }
        else
        {
            // Wait 1 sec
            sleep (1);
        }
    }

    DBOUT (DBG_LF_EXIT, "ohcitest: EventProcess() exit\n");

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// HandleBusReset()
//
//      Handles a bus reset event
//
// Entry:
//
//      NONE
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void HandleBusReset (void)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    PMyDmaT dma;
    int i;

    DBOUT (DBG_LF_ENTRY, "ohcitest: HandleBusReset()\n");

    // NOTE:  for some reason, if the bus reset occurred from
    // a manual insertion, the SelfIdDma does not update.  In
    // this case we must induce another bus reset
    if (!(gSharedMem->ParentPausedOnEvent.EventWaitingFor & busReset_BIT))
    {
        DBOUT (DBG_LF_INTERNAL, "ohcitest: manual insertion\n");

        // Induce bus reset
        DoBusReset (TRUE);

        // Wait 1 ms for everything to settle down
        usleep (1000);
    }

    // Show topolgy map
    dma = (PMyDmaT)gOhciDmaMemMap;
    printf ("timeStamp = %08X\n", dma->SelfIds.timeStamp);
    printf ("selfIDGeneration = %08X\n", dma->SelfIds.selfIDGeneration);
    for (i = 0; dma->SelfIds.Packet[i].reg[0] != 0; i++)
    {
        printf ("%d %08X ", i, (int)dma->SelfIds.Packet[i].reg);
        printf ("phy_ID %02X, ", dma->SelfIds.Packet[i].Packet0.phy_ID);
        printf ("L %X, ", dma->SelfIds.Packet[i].Packet0.L);
        printf ("gap_cnt %02X, ", dma->SelfIds.Packet[i].Packet0.gap_cnt);
        printf ("sp %X, ", dma->SelfIds.Packet[i].Packet0.sp);
        printf ("del %X, ", dma->SelfIds.Packet[i].Packet0.del);
        printf ("c %X, ", dma->SelfIds.Packet[i].Packet0.c);
        printf ("pwr %X, ", dma->SelfIds.Packet[i].Packet0.pwr);
        printf ("p0 %X, ", dma->SelfIds.Packet[i].Packet0.p0);
        printf ("p1 %X, ", dma->SelfIds.Packet[i].Packet0.p1);
        printf ("p2 %X", dma->SelfIds.Packet[i].Packet0.p2);
        printf ("\n");
    }

    DBOUT (DBG_LF_EXIT, "ohcitest: HandleBusReset() exit\n");

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DoBusReset()
//
//      Does actual bus reset code
//
// Entry:
//
//      WantToBeRoot    Indicate whether or not to set the Root Holdoff Bit
//                      during the bus reset.
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DoBusReset (BOOLEAN WantToBeRoot)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    QUADLET q;
    PHYRegister1_REG phyData;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DoBusReset()\n");

    // Initialize self id dma
    InitSelfId ();

    // Set the Initiate Bus Reset (IBR) bit in PHY Register 1
    phyData.reg = 0;
    phyData.fld.IBR = 1;
    if (WantToBeRoot == TRUE)
        phyData.fld.RHB = 1;    // We want to be root
    q = 0;
    ((PHYControl_REG *)&q)->fld.wrData = phyData.reg;
    ((PHYControl_REG *)&q)->fld.wrReg = 1;
    ((PHYControl_REG *)&q)->fld.regAddr = PHYRegister1;
    REG_WRITE (PHYControl, q);

    DBOUT (DBG_LF_EXIT, "ohcitest: DoBusReset() exit\n");
    return;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// SignalEvent
//
//      The child process calls this to signal that an event
//		has occurred.
//
// Entry:
//
//      sig     signal
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void SignalEvent (__u32 sig)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    DBOUT (DBG_LF_ENTRY, "ohcitest: SignalEvent()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: sig = %08X\n", sig);

    DBOUT1(DBG_LF_INTERNAL, "ohcitest: EventWaitingFor = %08X\n", gSharedMem->ParentPausedOnEvent.EventWaitingFor);

	// Is the parent waiting for this event?
    if (gSharedMem->ParentPausedOnEvent.EventWaitingFor & sig)
    {
        // Clear event
        gSharedMem->ParentPausedOnEvent.EventWaitingFor &= ~sig;

        // Signal parent
        kill (getppid(), SIGUSR1);
    }

    DBOUT (DBG_LF_EXIT, "ohcitest: SignalEvent() exit\n");

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ParentSignalHandler
//
//      Handles signals from the child event process.
//      Does nothing.
//
// Entry:
//
//      sig     signal
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void ParentSignalHandler (int sig)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    DBOUT (DBG_LF_ENTRY, "ohcitest: ParentSignalHandler()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: sig = %08X\n", sig);

    DBOUT (DBG_LF_EXIT, "ohcitest: ParentSignalHandler() exit\n");

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// PauseOnEvent
//
//      Pauses the parent waiting for the specified event to occur
//
// Entry:
//
//      Event       Event to wait for
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN PauseOnEvent (__u32 Event, __u32 TimeOutSec)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    time_t curTime;
    time_t startTime;

    DBOUT (DBG_LF_ENTRY, "ohcitest: PauseOnEvent()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: Event = %08X\n", Event);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: TimeOutSec = %08X\n", TimeOutSec);

    // Assume timeout
    retcode = FALSE;

    if (TimeOutSec == 0)
    {
        while ((gSharedMem->ParentPausedOnEvent.EventWaitingFor & Event) == Event)
        {
            DBOUT2(DBG_LF_INTERNAL, "ohcitest: EventWaitingFor = %08X, Event = %08X\n", gSharedMem->ParentPausedOnEvent.EventWaitingFor, Event);
            pause ();
        }
    }
    else
    {
        curTime = time (NULL);
        startTime = curTime;
        while ((gSharedMem->ParentPausedOnEvent.EventWaitingFor & Event) == Event)
        {
            DBOUT2(DBG_LF_INTERNAL, "ohcitest: EventWaitingFor = %08X, Event = %08X\n", gSharedMem->ParentPausedOnEvent.EventWaitingFor, Event);

            // Get current time
            curTime = time (NULL);

            // Check for timeout
            if (curTime - startTime >= TimeOutSec)
                goto Exit;

            // Sleep for 1 sec
            sleep (1);
        }
    }

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: PauseOnEvent() exit, retcode = %s\n", BoolToStr(retcode));
    return (retcode);

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DisplayMenuAndExecute()
//
//      Displays a menu, fetches a command and executes the command
//
// Entry:
//
//      NONE
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayMenuAndExecute (void)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN done;
    BOOLEAN ccode;
    CommandT cmd;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DisplayMenuAndExecute()\n");

    // Process until done
    for (done = FALSE; done != TRUE; )
    {
        // Get command
        ccode = GetCommand (&cmd);
        if (ccode != TRUE)
        {
            // Tell user he goofed
            printf ("Invalid command or bad parameter\n");

            // Display menu
            DisplayMenu ();

            // Get the next command
            continue;
        }
        DBOUT1 (DBG_LF_INTERNAL, "ohcitest: cmd.CmdType = %08X\n", cmd.CmdType);

        // Execute command
        if (cmd.CmdType == CMD_EXIT)
            done = TRUE;
        else
        {
            ccode = ExecuteCommand (&cmd);
            if (ccode != TRUE)
            {
                printf ("Error executing command\n");
            }
        }
    }

    DBOUT (DBG_LF_EXIT, "ohcitest: DisplayMenuAndExecute() exit\n");

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DisplayMenu()
//
//      Displays a menu of commands to execute
//
// Entry:
//
//      NONE
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayMenu (void)
{
    printf ("\n");
    printf ("\n");
    printf ("Please choose a command:\n");
    printf ("\n");
    printf ("[#|;] <comment>                  - Comment\n");
    printf ("Echo <string>                    - Echo a string\n");
    printf ("Pause [s|m|u] <seconds|ms|us>    - Pauses for the specified number of seconds, milliseconds or microseconds\n");
    printf ("[Exit|e|Quit|q]                  - Exit\n");
    printf ("[Read|r] <reg offset>            - Read the specified OHCI register offset\n");
    printf ("[Write|w] <reg offset> <value>   - Write a value to the specified OHCI register offset\n");
    printf ("rdm <dma mem offset>             - Read a QUADLET from the specified DMA memory offset\n");
    printf ("wdm <dma mem offset> <value>     - Write a QUADLET value to the specified DMA memory offset\n");
    printf ("InitCard                         - Resets and initializes the OHCI device\n");
    printf ("InitSelfId                       - Set up the OHCI card to receive self ID packets\n");
    printf ("BusReset                         - Reset the 1394 bus\n");
    printf ("[ReadQuadlet|rq] <node id>:<addr offset high>:<addr offset low>\n");
    printf ("                                 - Reads a QUADLET at the specified address\n");
    printf ("[DumpRom|dr] <node id>           - Decodes the configuration ROM of the specified node\n");
#ifdef DEBUG
    printf ("DebugFlags                       - Display/Modify DebugFlags\n");
#endif
    printf ("\n");
    printf ("Not working yet:\n");
    printf ("InitArDma                        - Set up the OHCI card to receive asynchronous packets\n");
    printf ("InitAtDma                        - Set up the OHCI card to transmit asynchronous packets\n");
    printf ("\n");
    printf ("\n");
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GetCommand()
//
//      Prompts the user to enter a command
//
// Entry:
//
//      cmd     -> command structure
//
// Exit:
//
//      TRUE    command fetched
//      FALSE   user entered invalid command
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN GetCommand (PCommandT cmd)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    BOOLEAN cmdValid;
    #define CMD_STR_MAX_SIZE    500
    char *cmdStr;

    DBOUT (DBG_LF_ENTRY, "ohcitest: GetCommand()\n");
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: cmd = %08X\n", cmd);

    // Assume failure (user want's to exit)
    retcode = FALSE;
    cmdStr = NULL;

    // Allocate command string
    cmdStr = malloc (CMD_STR_MAX_SIZE);
    if (cmdStr == NULL)
        goto Exit;

    // Get the command string
    do
    {
        // Display prompt
        printf ("> ");

        // Get command string
        fgets (cmdStr, 500, stdin);

    } while (cmdStr[0] == '\0' || cmdStr[0] == '\n');
        
    // Strip new line character
    if (cmdStr[strlen(cmdStr)-1] == '\n')
        cmdStr[strlen(cmdStr)-1] = '\0';
    DBOUT1 (DBG_LF_INTERNAL, "ohcitest: cmdStr = \"%s\"\n", cmdStr);

    // Parse command string
    cmdValid = ParseCommandString (cmdStr, cmd);
    if (cmdValid != TRUE)
        goto Exit;

    // OK
    retcode = TRUE;

    DBOUT1 (DBG_LF_EXIT_PARAMS, "ohcitest: cmd->CmdType = %08X\n", cmd->CmdType);

Exit:

    // Dealloc resources
    if (cmdStr != NULL)
        free (cmdStr);

    DBOUT1 (DBG_LF_EXIT, "ohcitest: GetCommand() exit, retcode = %s\n", BoolToStr(retcode));
    return (retcode);

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ParseCommandString()
//
//      Parses and validates the specified command string
//
// Entry:
//
//      str     command string
//
//      cmd     -> command structure
//
// Exit:
//
//      TRUE    command string parsed and validated
//      FALSE   invalid command string
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN ParseCommandString (char *CmdStr, PCommandT cmd)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    int ccode;
    char *strtemp;
    char *str;
    char *scmd;
    int numAssociations;
    int i;
    char *delimiters = " \t";
	char chr;

    DBOUT (DBG_LF_ENTRY, "ohcitest: ParseCommandString()\n");
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: CmdStr = %s\n", CmdStr);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: cmd = %08X\n", cmd);

    // Assume failure
    retcode = FALSE;
    str = CmdStr;
    strtemp = NULL;

    // Alloc temp string and copy
    strtemp = malloc (strlen(CmdStr)+1);
    if (strtemp == NULL)
        goto Exit;
    strcpy (strtemp, CmdStr);

    // Get the command
    scmd = strtok (strtemp, delimiters);
    if (scmd == NULL)
        goto Exit;
    DBOUT1 (DBG_LF_INTERNAL, "ohcitest: scmd = \"%s\"\n", scmd);

    // Is this a comment?
    if (scmd[0] == ';' || scmd[0] == '#')
    {
        cmd->CmdType = CMD_COMMENT;
    }
    else
    {
        // Search for the command string and associated command
        numAssociations = sizeof(gCmdAssoc)/sizeof(CommandAssociationT); 
        DBOUT1 (DBG_LF_INTERNAL, "ohcitest: numAssociations = %d\n", numAssociations);
        for (i=0; i<numAssociations; i++)
        {
            DBOUT2 (DBG_LF_INTERNAL, "ohcitest: CmdString[%d] = \"%s\"\n", i, gCmdAssoc[i].CmdString);
            if (strcasecmp (gCmdAssoc[i].CmdString, scmd) == 0)
                break;
        }

        // Did we find the command?
        if (i >= numAssociations)
            goto Exit;

        // Save the command type
        cmd->CmdType = gCmdAssoc[i].CmdType;
    }

    // Parse the command parameters
    switch (cmd->CmdType)
    {
        case CMD_EXIT:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_EXIT\n");
            break;

        case CMD_HELP:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_HELP\n");
            break;

        case CMD_COMMENT:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_COMMENT\n");
            break;

        case CMD_ECHO:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_ECHO\n");

            // Copy the rest of the command string into the echo buffer
            str = strchr (CmdStr, ' ');
            str += 1;
            strncpy (cmd->u.EchoCmd.EchoBuffer, str, MAX_ECHO_BUFFER);
            cmd->u.EchoCmd.EchoBuffer[MAX_ECHO_BUFFER-1] = '\0';

            break;

        case CMD_READ:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_READ\n");

            // Parse the register offset
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%x", &cmd->u.ReadCmd.Offset);
            if (ccode != 1)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: Offset = %08X\n", cmd->u.ReadCmd.Offset);

            break;

        case CMD_WRITE:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_WRITE\n");

            // Parse the register offset
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%x", &cmd->u.WriteCmd.Offset);
            if (ccode != 1)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: Offset = %08X\n", cmd->u.WriteCmd.Offset);

            // Parse the value to write
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%x", &cmd->u.WriteCmd.ValueToWrite);
            if (ccode != 1)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: ValueToWrite = %08X\n", cmd->u.WriteCmd.ValueToWrite);

            break;

        case CMD_READ_DMA_MEM:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_READ_DMA_MEM\n");

            // Parse the register offset
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%x", &cmd->u.ReadDmaMemCmd.Offset);
            if (ccode != 1)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: Offset = %08X\n", cmd->u.ReadDmaMemCmd.Offset);

            break;

        case CMD_WRITE_DMA_MEM:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_WRITE_DMA_MEM\n");

            // Parse the register offset
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%x", &cmd->u.WriteDmaMemCmd.Offset);
            if (ccode != 1)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: Offset = %08X\n", cmd->u.WriteDmaMemCmd.Offset);

            // Parse the value to write
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%x", &cmd->u.WriteDmaMemCmd.ValueToWrite);
            if (ccode != 1)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: ValueToWrite = %08X\n", cmd->u.WriteDmaMemCmd.ValueToWrite);

            break;

        case CMD_INIT_CARD:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_INIT_CARD\n");
            break;

        case CMD_INIT_SELF_ID:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_INIT_SELF_ID\n");
            break;

        case CMD_INIT_AR_DMA:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_INIT_AR_DMA\n");
            break;

        case CMD_INIT_AT_DMA:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_INIT_AR_DMA\n");
            break;

        case CMD_READ_QUADLET:

            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_READ_QUADLET\n");

            // Parse the node id:addr offset hi:addr offset low
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%x:%x:%x", 
                        &cmd->u.ReadQuadletCmd.NodeId,
                        &cmd->u.ReadQuadletCmd.AddrOffsetHigh,
                        &cmd->u.ReadQuadletCmd.AddrOffsetLow);
            if (ccode != 3)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: NodeId = %08X\n", cmd->u.ReadQuadletCmd.NodeId);
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: AddrOffsetHigh = %08X\n", cmd->u.ReadQuadletCmd.AddrOffsetHigh);
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: AddrOffsetLow = %08X\n", cmd->u.ReadQuadletCmd.AddrOffsetLow);

            break;

        case CMD_DUMP_ROM:

            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_DUMP_ROM\n");

            // Parse the node id
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%x", &cmd->u.DumpRomCmd.NodeId);
            if (ccode != 1)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: NodeId = %08X\n", cmd->u.DumpRomCmd.NodeId);

            break;

        case CMD_PAUSE:

            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_PAUSE\n");

            // Parse the seconds type
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%c", &chr);
            if (ccode != 1)
                goto Exit;
			switch (toupper(chr))
			{
				case 'S':
					cmd->u.PauseCmd.SecondsType = PAUSE_SECONDS;
					break;
				case 'M':
					cmd->u.PauseCmd.SecondsType = PAUSE_MILLISECONDS;
					break;
				case 'U':
					cmd->u.PauseCmd.SecondsType = PAUSE_MICROSECONDS;
					break;
				default:
					goto Exit;
					break;
			}
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: SecondsType = %08X\n", cmd->u.PauseCmd.SecondsType);

            // Parse the seconds
            str = strtok (NULL, delimiters);
            if (str == NULL)
                goto Exit;
            ccode = sscanf (str, "%d", &cmd->u.PauseCmd.Seconds);
            if (ccode != 1)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: Seconds = %08X\n", cmd->u.PauseCmd.Seconds);

            break;

        case CMD_BUS_RESET:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_BUS_RESET\n");
            break;

#ifdef DEBUG
        case CMD_DEBUG_FLAGS:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_DEBUG_FLAGS\n");

            // Parse the DebugFlags value
            cmd->u.DebugFlagsCmd.PromptForValue = FALSE;
            str = strtok (NULL, delimiters);
            if (str == NULL)
            {
                // Indicate that we should prompt during command execution
                cmd->u.DebugFlagsCmd.PromptForValue = TRUE;
                break;
            }
            ccode = sscanf (str, "%x", &cmd->u.DebugFlagsCmd.DebugFlagsValue);
            if (ccode != 1)
                goto Exit;
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: DebugFlagsValue = %08X\n", cmd->u.DebugFlagsCmd.DebugFlagsValue);

            break;
#endif

    }

    DBOUT1 (DBG_LF_EXIT, "ohcitest: cmd->CmdType = %08X\n", cmd->CmdType);

    // OK
    retcode = TRUE;

Exit:

    // Dealloc resources
    if (strtemp != NULL)
        free (strtemp);

    DBOUT1 (DBG_LF_EXIT, "ohcitest: ParseCommandString() exit, retcode = %s\n", BoolToStr(retcode));

    return retcode;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ExecuteCommand()
//
//      Executes the specified command
//
// Entry:
//
//      cmd     -> command structure
//
// Exit:
//
//      TRUE    command executed without error
//      FALSE   error executing command
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN ExecuteCommand (PCommandT cmd)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    BOOLEAN ccode;
    QUADLET quad;

    DBOUT (DBG_LF_ENTRY, "ohcitest: ExecuteCommand()\n");
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: cmd = %08X\n", cmd);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: cmd->CmdType = %08X\n", cmd->CmdType);

    // Asusme failure
    retcode = FALSE;

    // Execute the command
    switch (cmd->CmdType)
    {

        case CMD_EXIT:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_EXIT\n");
            break;

        case CMD_HELP:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_HELP\n");
            DisplayMenu();
            break;

        case CMD_COMMENT:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_COMMENT\n");
            break;

        case CMD_ECHO:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_COMMENT\n");
            printf ("%s\n", cmd->u.EchoCmd.EchoBuffer);
            break;

        case CMD_READ:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_READ\n");
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: Offset = %08X\n", cmd->u.ReadCmd.Offset);
            quad = QUADLET_READ (gOhciRegMemMap, cmd->u.ReadCmd.Offset);
            printf ("%08X <-- %08X\n", (int)quad, (int)gOhciRegMemMap + cmd->u.ReadCmd.Offset);
            break;

        case CMD_WRITE:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_WRITE\n");
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: Offset = %08X\n", cmd->u.WriteCmd.Offset);
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: ValueToWrite = %08X\n", cmd->u.WriteCmd.ValueToWrite);
            QUADLET_WRITE (gOhciRegMemMap, cmd->u.WriteCmd.Offset, cmd->u.WriteCmd.ValueToWrite);
            printf ("%08X --> %08X\n", (int)cmd->u.WriteCmd.ValueToWrite, (int)gOhciRegMemMap + cmd->u.WriteCmd.Offset);
            break;

        case CMD_READ_DMA_MEM:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_READ_DMA_MEM\n");
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: Offset = %08X\n", cmd->u.ReadDmaMemCmd.Offset);
            quad = QUADLET_READ (gOhciDmaMemMap, cmd->u.ReadDmaMemCmd.Offset);
            printf ("%08X <-- reg %08X\n", (int)quad, (int)gOhciDmaMemMap + cmd->u.ReadDmaMemCmd.Offset);
            break;

        case CMD_WRITE_DMA_MEM:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_WRITE_DMA_MEM\n");
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: Offset = %08X\n", cmd->u.WriteDmaMemCmd.Offset);
            DBOUT1 (DBG_LF_INTERNAL, "ohcitest: ValueToWrite = %08X\n", cmd->u.WriteDmaMemCmd.ValueToWrite);
            QUADLET_WRITE (gOhciDmaMemMap, cmd->u.WriteDmaMemCmd.Offset, cmd->u.WriteDmaMemCmd.ValueToWrite);
            printf ("%08X --> reg %08X\n", (int)cmd->u.WriteDmaMemCmd.ValueToWrite, (int)gOhciDmaMemMap + cmd->u.WriteDmaMemCmd.Offset);
            break;

        case CMD_INIT_CARD:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_INIT_CARD\n");
            ccode = InitCard ();
            if (ccode != TRUE)
                printf ("InitCard() failed\n");
            break;

        case CMD_INIT_SELF_ID:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_INIT_SELF_ID\n");
            ccode = InitSelfId ();
            if (ccode != TRUE)
                printf ("InitSelfId() failed\n");
            break;

        case CMD_INIT_AR_DMA:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_INIT_AR_DMA\n");
            ccode = InitArDma ();
            if (ccode != TRUE)
                printf ("InitArDma() failed\n");
            break;

        case CMD_INIT_AT_DMA:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_INIT_AT_DMA\n");
            ccode = InitAtDma ();
            if (ccode != TRUE)
                printf ("InitAtDma() failed\n");
            break;

        case CMD_READ_QUADLET:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_READ_QUADLET\n");
            ccode = ReadQuadlet (
                        cmd->u.ReadQuadletCmd.NodeId, 
                        cmd->u.ReadQuadletCmd.AddrOffsetHigh, 
                        cmd->u.ReadQuadletCmd.AddrOffsetLow,
                        &quad);
            if (ccode != TRUE)
                printf ("ReadQaudlet() failed\n");
            else
                printf ("%08X <-- 1394 addr %02X:%04X:%08X\n", 
                        (int)quad, 
                        cmd->u.ReadQuadletCmd.NodeId, 
                        cmd->u.ReadQuadletCmd.AddrOffsetHigh, 
                        cmd->u.ReadQuadletCmd.AddrOffsetLow);
            break;

        case CMD_DUMP_ROM:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_DUMP_ROM\n");

            // Read the configuration ROM into a buffer
            ccode = ReadConfigRom (
                cmd->u.DumpRomCmd.NodeId, 
                gConfigRomBuf, 
                CONFIG_ROM_MAX_SIZE, 
                &gConfigRomBufSize);
            if (ccode != TRUE)
            {
                printf ("Cannot read configuration ROM\n");
                break;
            }

            // Dump configuration ROM
            ccode = DumpConfigRom (gConfigRomBuf, gConfigRomBufSize);
            if (ccode != TRUE)
            {
                printf ("Cannot dump configuration ROM\n");
                break;
            }
        
            // Decode configuration ROM
            ccode = DecodeConfigRom (gConfigRomBuf, gConfigRomBufSize);
            if (ccode != TRUE)
            {
                printf ("Cannot decode configuration ROM\n");
                break;
            }

            break;

        case CMD_PAUSE:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_PAUSE\n");
			MyPause (cmd->u.PauseCmd.SecondsType, cmd->u.PauseCmd.Seconds);
            break;

        case CMD_BUS_RESET:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_BUS_RESET\n");
            ccode = BusReset ();
            if (ccode != TRUE)
                printf ("BusReset() failed\n");
            break;

#ifdef DEBUG
        case CMD_DEBUG_FLAGS:
            DBOUT (DBG_LF_INTERNAL, "ohcitest: CMD_DEBUG_FLAGS\n");
            DisplayModifyDebugFlags (cmd->u.DebugFlagsCmd.PromptForValue, cmd->u.DebugFlagsCmd.DebugFlagsValue);
            break;
#endif

        default:
            goto Exit;
            break;
    }

    // OK
    retcode = TRUE;


Exit:

    DBOUT1 (DBG_LF_EXIT, "ohcitest: ExecuteCommand() exit, retcode = %s\n", BoolToStr(retcode));
    return (retcode);

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

#ifdef DEBUG
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DisplayModifyDebugFlags()
//
//      Displays and/or modifies DebugFlags
//
// Entry:
//
//      PromptForValue      Prompt for the value from the user
//                          TRUE: Display current DebugFlags and prompt the user.
//                          FALSE: Set DebugFlags to DebugFlagsValue
//
//      DebugFlagsValue     New value of DebugFlags (only if PromptForValue == FALSE)
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DisplayModifyDebugFlags (BOOLEAN PromptForValue, int DebugFlagsValue)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    int promptedValue;
    int ccode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DisplayModifyDebugFlags()\n");
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: PromptForValue = %s\n", BoolToStr(PromptForValue));
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DebugFlagsValue = %08X\n", DebugFlagsValue);

    // Do we need to display and prompt?
    if (PromptForValue == TRUE)
    {
        // Display (and set to current value)
        DBGInitialize (DebugFlags);

        // Prompt for new value
        printf ("Please enter a hex DebugFlags value: ");
        ccode = scanf ("%x", &promptedValue);
        if (ccode != 1)
            goto Exit;

        // Set DebugFlags and display new values
        DBGInitialize (promptedValue);
    }
    else
    {
        // Set DebugFlags and display new values
        DBGInitialize (DebugFlagsValue);
    }

Exit:
    DBOUT (DBG_LF_EXIT, "ohcitest: DisplayModifyDebugFlags() exit\n");
    return;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}
#endif

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// InitCard()
//
//      Initialize the OHCI board
//
// Entry:
//
//      NONE
//
// Exit:
//
//      TRUE    success
//      FALSE   error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN InitCard (void)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    QUADLET q;

    DBOUT (DBG_LF_ENTRY, "ohcitest: InitCard()\n");

    // Assume failure
    retcode = FALSE;

    // Soft reset
    q = 0;
    ((HCControlSet_REG *)&q)->fld.softReset = 1;
    REG_WRITE (HCControlSet, q);

    // Wait 1 ms for everything to settle down
    usleep (1000);

    // Set Link Power Status bit in HCControl
    q = 0;
    ((HCControlSet_REG *)&q)->fld.LPS = 1;
    REG_WRITE (HCControlSet, q);

    // Wait 1 ms for everything to settle down
    usleep (1000);

    // Link enable  q = 0;
    ((HCControlSet_REG *)&q)->fld.LPS = 1;
    ((HCControlSet_REG *)&q)->fld.linkEnable = 1;
    REG_WRITE (HCControlSet, q);

    // Wait 1 ms for everything to settle down
    usleep (1000);

    // Enable cycle timer
    q = 0;
    ((LinkControl_REG *)&q)->fld.cycleTimerEnable = 1;
    ((LinkControl_REG *)&q)->fld.cycleMaster = 1;
    REG_WRITE (LinkControl, q);

    // Wait 1 ms for everything to settle down
    usleep (1000);

    // Were we assigned a node number?
    q = REG_READ (NodeID);
    if (q == -1)
        goto NodeIdNotAssigned;

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: InitCard() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

NodeIdNotAssigned:
    printf ("ERROR: InitCard(): NodeID not assigned after link enable\n");
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// InitSelfId()
//
//      Initializes the asynchronous reveive dma
//
// Entry:
//
//      NONE
//
// Exit:
//
//      TRUE    success
//      FALSE   error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN InitSelfId (void)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    QUADLET q;

    DBOUT (DBG_LF_ENTRY, "ohcitest: InitSelfId()\n");

    // Assume failure
    retcode = FALSE;

    // Clear Self ID DMA memory
    memset (((BYTE*)gOhciDmaMemMap) + offsetof(MyDmaT,SelfIds), 0, sizeof(SelfIDReceive));

    // Setup Receive Seld ID Buffer
    q = gOhciDmaMemPhysAddr + offsetof(MyDmaT,SelfIds);
    REG_WRITE (SelfIDBuffer, q);

    // Enable cycle timer and enable Receive Self ID
    q = 0;
    ((LinkControl_REG *)&q)->fld.rcvSelfID = 1;
    REG_WRITE (LinkControl, q);

    // OK
    retcode = TRUE;

    DBOUT1 (DBG_LF_EXIT, "ohcitest: InitSelfId() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// BusReset()
//
//      Initializes the asynchronous reveive dma
//
// Entry:
//
//      NONE
//
// Exit:
//
//      TRUE    success
//      FALSE   error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN BusReset (void)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    BOOLEAN ccode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: BusReset()\n");

    // Assume failure
    retcode = FALSE;

    // Prevent event process from processing any events
    // until we are ready to wait for it
    gSharedMem->IgnoreEvents = TRUE;

    // Tell event process we are waiting for a bus reset
    gSharedMem->ParentPausedOnEvent.EventWaitingFor = busReset_BIT;

    // Set the Initiate Bus Reset (IBR) bit in PHY Register 1
    DoBusReset (TRUE);

    // Wait for the event process to signal
    gSharedMem->IgnoreEvents = FALSE;
    ccode = PauseOnEvent (busReset_BIT, 2);
    if (ccode != TRUE)
        goto BusResetError;

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: BusReset() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

BusResetError:
    printf ("ERROR: Bus reset did not occur\n");
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// InitArDma()
//
//      Initializes the asynchronous receive DMA.
//
// Entry:
//
//      NONE
//
// Exit:
//
//      TRUE    success
//      FALSE   error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN InitArDma (void)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    PDmaNodeT dmaNodeAllocated;
    PAsync_INPUT_MORE_descriptor dmaDesc;
    QUADLET q;
    BOOLEAN ccode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: InitArDma()\n");

    // Assume failure
    retcode = FALSE;
    dmaNodeAllocated = NULL;

    // Allocate a DMA node for a INPUT_MORE descriptor and receive buffer
    ccode = DmaAllocAndAddDmaNodeToList (
                &gAsyncResponseReceiveDmaList,
                sizeof (Async_INPUT_MORE_descriptor),
                sizeof (Block_read_response_receive) + 100,         // 104 byte data block for now 
                                                                    // change later when we get more 
                                                                    // robust dma paging working with driver
                &dmaNodeAllocated);
    if (ccode != TRUE)
        goto AllocError;

    // Fill in INPUT_MORE descriptor
    dmaDesc = (PAsync_INPUT_MORE_descriptor)dmaNodeAllocated->DescBlkHandle->BlockAddr;
    dmaDesc->reqCount = sizeof (Block_read_response_receive) + 100; // 104 byte data block for now 
                                                                    // change later when we get more 
                                                                    // robust dma paging working with driver

    dmaDesc->b = BRANCH_ALWAYS;
    dmaDesc->i = INTERRUPT_NEVER;
    dmaDesc->key = KEY_ADDRESS;
    dmaDesc->s = STATUS_CTRL_WRITE;
    dmaDesc->cmd = CMD_INPUT_MORE;
    dmaDesc->dataAddress = dmaNodeAllocated->DataBlkHandle->PhysAddr >> 2;
    dmaDesc->Z = Z_LAST;
    dmaDesc->branchAddress = 0;

    // Stop the async response receive dma
    q = 0;
    ((AsyncResponseReceiveContextControl_REG *)&q)->fld.run = 1;
    REG_WRITE (AsyncResponseReceiveContextControlClear, q);
    usleep (1000);  // wait for dma to stop

    // Set the async response receive dma command ptr
    q = 0;
    ((AsyncResponseReceiveCommandPtr_REG *)&q)->fld.descriptorAddress = dmaNodeAllocated->DescBlkHandle->PhysAddr >> 4;
    ((AsyncResponseReceiveCommandPtr_REG *)&q)->fld.Z = 1;
    REG_WRITE (AsyncResponseReceiveCommandPtr, q);

    // Start the async response receive dma
    q = 0;
    ((AsyncResponseReceiveContextControl_REG *)&q)->fld.run = 1;
    REG_WRITE (AsyncResponseReceiveContextControlSet, q);

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: InitArDma() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

AllocError:
    printf ("ERROR: Unable to allocate dma memory\n");
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// InitAtDma()
//
//      Initializes the asynchronous transmit DMA.
//
// Entry:
//
//      NONE
//
// Exit:
//
//      TRUE    success
//      FALSE   error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN InitAtDma (void)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    PDmaNodeT dmaNodeAllocated;
    PAsync_OUTPUT_LAST_descriptor dmaDesc;
    QUADLET q;
    BOOLEAN ccode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: InitAtDma()\n");

    // Assume failure
    retcode = FALSE;
    dmaNodeAllocated = NULL;

    // Allocate a DMA node for a OUTPUT_LAST descriptor and Transmit buffer
    ccode = DmaAllocAndAddDmaNodeToList (
                &gAsyncRequestTransmitDmaList,
                sizeof (Async_OUTPUT_LAST_descriptor),
                sizeof (Block_read_request_transmit) + 100,         // 104 byte data block for now 
                                                                    // change later when we get more 
                                                                    // robust dma paging working with driver
                &dmaNodeAllocated);
    if (ccode != TRUE)
        goto AllocError;

    // Fill in OUTPUT_LAST descriptor
    dmaDesc = (PAsync_OUTPUT_LAST_descriptor) dmaNodeAllocated->DescBlkHandle->BlockAddr;
    dmaDesc->reqCount = sizeof (Block_read_request_transmit) + 100; // 104 byte data block for now 
                                                                    // change later when we get more 
                                                                    // robust dma paging working with driver

    dmaDesc->b = BRANCH_ALWAYS;
    dmaDesc->i = INTERRUPT_NEVER;
    dmaDesc->key = KEY_ADDRESS;
    dmaDesc->cmd = CMD_OUTPUT_LAST;
    dmaDesc->dataAddress = dmaNodeAllocated->DataBlkHandle->PhysAddr;
    dmaDesc->Z = Z_LAST;
    dmaDesc->branchAddress = 0;

    // Stop the async request transmit dma
    q = 0;
    ((AsyncRequestTransmitContextControl_REG *)&q)->fld.run = 1;
    REG_WRITE (AsyncRequestTransmitContextControlClear, q);
    usleep (1000);  // wait for dma to stop

    // Set the async request transmit command ptr
    q = 0;
    ((AsyncRequestTransmitCommandPtr_REG *)&q)->fld.descriptorAddress = dmaNodeAllocated->DescBlkHandle->PhysAddr >> 4;
    ((AsyncRequestTransmitCommandPtr_REG *)&q)->fld.Z = 2;
    REG_WRITE (AsyncRequestTransmitCommandPtr, q);

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: InitAtDma() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

AllocError:
    printf ("ERROR: Unable to allocate dma memory\n");
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaAllocateBlock()
//
//      Allocates a block from the list of free memory
//
// Entry:
//
//      BlockSize       block size to allocate
//
//      DmaBlockHandle  -> to structure to be filled in
//
// Exit:
//
//      TRUE    success
//      FALSE   error
//
//      DmaBlockHandle  if return code is TRUE, has been allocated from application space
//                      with the following fields set:
//                      BlockAddr         block address allocated.
//                      PhysAddr          physical address allocated.
//                      BlockSize         block size requested
//                      RealAllocedSize   true block size allocated
//                      IsFree            set to FALSE.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaAllocateBlock (__u32 BlockSize, PDmaBlockHandleT *DmaBlockHandle)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    PDmaBlockHandleT freeBlock;
    PDmaBlockHandleT newBlock;
    __u32 realBlockSize;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaAllocateBlock()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: BlockSize = %08X\n", BlockSize);

    // Assume failure
    retcode = FALSE;
    *DmaBlockHandle = NULL;
    newBlock = NULL;

    // Check size
    if (BlockSize == 0)
        goto InvalidSizeError;

    // Calculate the real block size jumping up to next multiple of 16
    realBlockSize = (BlockSize + 0xF) & 0xFFFFFFF0;

    // Search through the free list for a first fit block
    freeBlock = gDynamicDmaMemoryFreeHead;
    while (freeBlock && freeBlock->BlockSize < realBlockSize)
        freeBlock = freeBlock->Next;
    if (freeBlock == NULL)
        goto NotFoundError;

    // Found a free block

    // Allocate a new DMA mem man block to be put in the "alloced" list
    newBlock = malloc (sizeof (DmaBlockHandleT));
    if (newBlock == NULL)
        goto AllocError;

    // Initialize the new "alloced" structure
    newBlock->Next = NULL;
    newBlock->IsFree = FALSE;
    newBlock->BlockAddr = freeBlock->BlockAddr;
    newBlock->PhysAddr = freeBlock->PhysAddr;
    newBlock->BlockSize = BlockSize;
    newBlock->RealAllocedSize = realBlockSize;

    // Now sub allocate the free block
    freeBlock->BlockSize -= realBlockSize;
    freeBlock->BlockAddr += realBlockSize;
    freeBlock->PhysAddr += realBlockSize;
    freeBlock->RealAllocedSize = 0;

    // Delete the free memory management structure?
    if (freeBlock->BlockSize == 0)
    {
        // Free the block structure
        free (freeBlock);

        // Reset the free head pointer
        if (freeBlock == gDynamicDmaMemoryFreeHead)
            gDynamicDmaMemoryFreeHead = NULL;
    }

    // Add the new structure to the head of the alloced list
    newBlock->Next = gDynamicDmaMemoryAllocedHead;
    gDynamicDmaMemoryAllocedHead = newBlock;

    // Clear the new memory block
    memset ((BYTE *)newBlock->BlockAddr, 0, newBlock->RealAllocedSize);

    // OK
    retcode = TRUE;
    *DmaBlockHandle = newBlock;
    DBOUT1 (DBG_LF_EXIT_PARAMS, "ohcitest: *DmaBlockHandle = %08X\n", *DmaBlockHandle);
    DBOUT1 (DBG_LF_EXIT_PARAMS, "ohcitest: (*DmaBlockHandle)->BlockAddr = %08X\n", (*DmaBlockHandle)->BlockAddr);
    DBOUT1 (DBG_LF_EXIT_PARAMS, "ohcitest: (*DmaBlockHandle)->PhysAddr = %08X\n", (*DmaBlockHandle)->PhysAddr);
    DBOUT1 (DBG_LF_EXIT_PARAMS, "ohcitest: (*DmaBlockHandle)->RealAllocedSize = %08X\n", (*DmaBlockHandle)->RealAllocedSize);

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaAllocateBlock() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

InvalidSizeError:
    printf ("ERROR: BlockSize %08X is invalid\n", BlockSize);
    goto Exit;

NotFoundError:
    printf ("ERROR: No blocks found with enough space in free list\n");
    goto Exit;

AllocError:
    printf ("ERROR: Cannot allocate a new DMA Block Handle for the allocated list\n");
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaCreateBlock()
//
//      Allocates and adds a new free block to dynamic dma free list
//
// Entry:
//
//      BlockAddr   Address of dma block
//
//      BlockSize   size of dma block
//
//      PhysAddr    Physical Address of block
//
// Exit:
//
//      TRUE    success
//      FALSE   error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaCreateBlock (__u32 BlockAddr, __u32 BlockSize, __u32 PhysAddr)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    PDmaBlockHandleT newBlock;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaCreateBlock()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: BlockAddr = %08X\n", BlockAddr);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: BlockSize = %08X\n", BlockSize);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: PhysAddr = %08X\n", PhysAddr);

    // Assume failure
    retcode = FALSE;

    // Is the new block address 16 byte aligned?
    if (BlockAddr & 0xF)
        goto AlignError;

    // Is the new block size in multiples of 16
    if (BlockSize & 0xF)
        goto SizeError;

    // Allocate a memory management block handle
    newBlock = malloc (sizeof (DmaBlockHandleT));
    if (newBlock == NULL)
        goto AllocError;

    // Initialize block info
    newBlock->Next = NULL;
    newBlock->IsFree = TRUE;
    newBlock->BlockAddr = BlockAddr;
    newBlock->BlockSize = BlockSize;
    newBlock->RealAllocedSize = 0;          // Only uses in alloced blocks
    newBlock->PhysAddr = PhysAddr;

    // Has the free list head been initialized?
    if (gDynamicDmaMemoryFreeHead == NULL)
    {
        // Initialize list head
        gDynamicDmaMemoryFreeHead = newBlock;
    }
    else
    {
        // Insert this block into the list
        // and try to coalesce according to address

        // Do we need a new list head?
        if (gDynamicDmaMemoryFreeHead->BlockAddr > newBlock->BlockAddr)
        {
            // Put the new block at the head
            newBlock->Next = gDynamicDmaMemoryFreeHead;
            gDynamicDmaMemoryFreeHead = newBlock;

            // Try to coalesce with next
            DmaCoalesceWithNextBlock (newBlock);
        }
        else
        {
            PDmaBlockHandleT prev;
            PDmaBlockHandleT cur;

            // Find where we belong in the list
            prev = gDynamicDmaMemoryFreeHead;
            cur = gDynamicDmaMemoryFreeHead->Next;
            while (cur != NULL && cur->BlockAddr < newBlock->BlockAddr)
            {
                prev = cur;
                cur = cur->Next;
            }

            // Add to the list
            prev->Next = newBlock;
            newBlock->Next = cur;

            // Try to coalesce with next and previous
            DmaCoalesceWithNextBlock (newBlock);
            DmaCoalesceWithNextBlock (prev);
        }
    }

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaCreateBlock() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

AlignError:
    printf ("ERROR: BlockAddress %08X must be aligned on a 16-byte boundary!\n", BlockSize);
    goto Exit;

SizeError:
    printf ("ERROR: BlockSize %08X must be in multiples of 16 bytes!\n", BlockSize);
    goto Exit;

AllocError:
    printf ("ERROR: Unable to allocate system memory for a new DMA Block Handle in the free list\n");
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaFreeBlock()
//
//      Adds a free block to dynamic dma memory
//
// Entry:
//
//      DmaBlockHandle  DMA memory management structure with
//                      the following fields set:
//
//                      IsFree          FALSE
//
//                      BlockAddr       Address of dma block
//
//                      BlockSize       size of dma block
//
//                      RealAllocedSize actual size of dma block
//
//                      PhysAddr        Physical Address of block
//
// Exit:
//
//      TRUE    success
//      FALSE   error
//
//      If return code is TRUE, then DmaBlockHandle has been put on the
//      alloced list and IsFree is set to FALSE.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaFreeBlock (PDmaBlockHandleT DmaBlockHandle)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    PDmaBlockHandleT prev;
    PDmaBlockHandleT cur;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaFreeBlock()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle = %08X\n", DmaBlockHandle);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->BlockAddr = %08X\n", DmaBlockHandle->BlockAddr);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->BlockSize = %08X\n", DmaBlockHandle->BlockSize);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->PhysAddr = %08X\n", DmaBlockHandle->PhysAddr);

    // Assume failure
    retcode = FALSE;

    // Check if the block has been allocated
    if (DmaBlockHandle->IsFree == TRUE)
        goto AlreadyFreeError;

    // Find the block on the alloced list
    prev = NULL;
    cur = gDynamicDmaMemoryAllocedHead;
    while (cur != NULL && cur != DmaBlockHandle)
    {
        prev = cur;
        cur = cur->Next;
    }
    if (cur != DmaBlockHandle)
        goto NotFoundError;

    // Delete the block from the alloced list
    if (gDynamicDmaMemoryAllocedHead == DmaBlockHandle)
        gDynamicDmaMemoryAllocedHead = DmaBlockHandle->Next;
    else
        prev->Next = DmaBlockHandle->Next;
    DmaBlockHandle->Next = NULL;

    // Change the status of the block to "FREE"
    DmaBlockHandle->IsFree = TRUE;

    // Set the real size of the block
    DmaBlockHandle->BlockSize = DmaBlockHandle->RealAllocedSize;
    DmaBlockHandle->RealAllocedSize = 0;

    // Add the block onto the free list

    // Is the free list empty?
    if (gDynamicDmaMemoryFreeHead == NULL)
    {
        // Initialize list head
        gDynamicDmaMemoryFreeHead = DmaBlockHandle;
    }
    else
    {
        // Insert this block into the list
        // Try to coalesce according to address

        // Do we need a new list head?
        if (gDynamicDmaMemoryFreeHead->BlockAddr > DmaBlockHandle->BlockAddr)
        {
            // Put the new block at the head
            DmaBlockHandle->Next = gDynamicDmaMemoryFreeHead;
            gDynamicDmaMemoryFreeHead = DmaBlockHandle;

            // Try to coalesce with next block
            DmaCoalesceWithNextBlock (DmaBlockHandle);
        }
        else
        {
            // Find where we belong in the list
            prev = gDynamicDmaMemoryFreeHead;
            cur = gDynamicDmaMemoryFreeHead->Next;
            while (cur != NULL && cur->BlockAddr < DmaBlockHandle->BlockAddr)
            {
                prev = cur;
                cur = cur->Next;
            }

            // Add to the list
            prev->Next = DmaBlockHandle;
            DmaBlockHandle->Next = cur;

            // Try to coalesce with next and previous blocks
            DmaCoalesceWithNextBlock (DmaBlockHandle);
            DmaCoalesceWithNextBlock (prev);
        }
    }

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaFreeBlock() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

AlreadyFreeError:
    printf ("ERROR: DMA Block Handle %08X is already free\n", (int)DmaBlockHandle);
    goto Exit;

NotFoundError:
    printf ("ERROR: Unable to find DMA Block Handle %08X on the allocated list\n", (int)DmaBlockHandle);
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaCoalesceWithNextBlock()
//
//      Will attemp to coalesce the specified block with it's next block.
//
// Entry:
//
//      Block   Block to coalesce with Block->Next
//
// Exit:
//
//      TRUE    success
//      FALSE   error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaCoalesceWithNextBlock (PDmaBlockHandleT DmaBlockHandle)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    PDmaBlockHandleT BlockToDelete;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaCoalesceWithNextBlock()\n");
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle = %08X\n", DmaBlockHandle);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->BlockAddr = %08X\n", DmaBlockHandle->BlockAddr);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->BlockSize = %08X\n", DmaBlockHandle->BlockSize);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->PhysAddr = %08X\n", DmaBlockHandle->PhysAddr);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->Next = %08X\n", DmaBlockHandle->Next);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->Next->BlockAddr = %08X\n", DmaBlockHandle->Next->BlockAddr);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->Next->BlockSize = %08X\n", DmaBlockHandle->Next->BlockSize);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaBlockHandle->Next->PhysAddr = %08X\n", DmaBlockHandle->Next->PhysAddr);

    // Assume failure
    retcode = FALSE;

    // Can the blocks be coalesced?
    // NOTE:  has to match both with linear space and physical

    // Adjacent linear space?
    if (DmaBlockHandle->BlockAddr+DmaBlockHandle->BlockSize != DmaBlockHandle->Next->BlockAddr)
        goto Exit;

    // Adjacent physical space?
    if (DmaBlockHandle->PhysAddr+DmaBlockHandle->BlockSize != DmaBlockHandle->Next->PhysAddr)
        goto Exit;

    // Coalesce the block with the next
    BlockToDelete = DmaBlockHandle->Next;
    DmaBlockHandle->BlockSize += BlockToDelete->BlockSize;
    DmaBlockHandle->Next = BlockToDelete->Next;

    // Free the old next block
    free (BlockToDelete);

    DBOUT2 (DBG_LF_INTERNAL, "ohcitest: block %08X coalesced with next %08X\n", DmaBlockHandle, BlockToDelete);
    DBOUT1 (DBG_LF_INTERNAL, "ohcitest: coalesced block addr = %08X\n", DmaBlockHandle->BlockAddr);
    DBOUT1 (DBG_LF_INTERNAL, "ohcitest: coalesced phys addr = %08X\n", DmaBlockHandle->PhysAddr);
    DBOUT1 (DBG_LF_INTERNAL, "ohcitest: coalesced block size = %08X\n", DmaBlockHandle->BlockSize);
    DBOUT1 (DBG_LF_INTERNAL, "ohcitest: coalesced block next = %08X\n", DmaBlockHandle->Next);

    // OK
    retcode = TRUE;

Exit:

    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaCoalesceWithNextBlock() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaAllocDmaNode()
//
//      Allocates a new DMA Node.
//
// Entry:
//
//      DmaListHead         Head of the list
//
//      DescriptorSize      size of the descriptor to allocate in DMA memory
//
//      DataSize            size of the data buffer to allocate in DMA memory
//
// Exit:
//
//      TRUE                success
//      FALSE               error
//
//      DmaNode             if return code is TRUE, then this holds the
//                          pointer to the DmaNodeT structure allocated
//                          and put at the end of the DMA List.
//
//      DmaListHead         if return code is TRUE and if DmaListHead == NULL 
//                          on entry, then DmaListHead = DmaNode
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaAllocDmaNode (
            __u32 DescriptorSize,
            __u32 DataSize,
            PDmaNodeT *DmaNode)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    int ccode;
    PDmaNodeT newDmaNode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaAllocDmaNode()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DescriptorSize = %08X\n", DescriptorSize);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DataSize = %08X\n", DataSize);

    // Assume failure
    retcode = FALSE;
    newDmaNode = NULL;

    // We must at least have a descriptor
    if (DescriptorSize == 0)
        goto DescSizeError;

    // Allocate the Dma Node from application space
    newDmaNode = malloc (sizeof(DmaNodeT));
    if (newDmaNode == NULL)
        goto AllocError;

    // Initialize node strucrture
    newDmaNode->DataBlkHandle = NULL;
    newDmaNode->DescBlkHandle = NULL;
    newDmaNode->Next = NULL;

    // Allocate descriptor from dynamic dma memory
    ccode = DmaAllocateBlock (
                DescriptorSize, 
                &(newDmaNode->DescBlkHandle));
    if (ccode != TRUE)
        goto AllocDescError;

    // Allocate data from dynamic dma memory
    if (DataSize != 0)
    {
        ccode = DmaAllocateBlock (
                    DataSize, 
                    &(newDmaNode->DataBlkHandle));
        if (ccode != TRUE)
            goto AllocDataError;
    }

    // OK
    *DmaNode = newDmaNode;
    retcode = TRUE;

    DBOUT1(DBG_LF_EXIT_PARAMS, "ohcitest: *DmaNode = %08X\n", *DmaNode);

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaAllocDmaNode() exit, retcode = %s\n", BoolToStr(retcode));
    return (retcode);

DescSizeError:
    printf ("ERROR: Descriptor size %08X is invalid\n", DescriptorSize);
    goto ExitWithError;

AllocError:
    printf ("ERROR: Unable to allocate dma node memory\n");
    printf ("Last error = %08x = %s\n", errno, strerror(errno));
    goto ExitWithError;

AllocDescError:
    printf ("ERROR: Unable to allocate descriptor in dma memory\n");
    goto ExitWithError;

AllocDataError:
    printf ("ERROR: Unable to allocate data in dma memory\n");
    goto ExitWithError;

ExitWithError:

    // Deallocate anything already allocated
    if (newDmaNode != NULL)
    {
        DmaFreeDmaNode (newDmaNode);
        newDmaNode = NULL;
    }
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaAddDmaNodeToList()
//
//      Adds a DMA Node to the end of the specified list.
//
// Entry:
//
//      DmaListHead         Head of the list
//
//      DmaNode             DMA Node to add
//
// Exit:
//
//      TRUE                success
//      FALSE               error
//
//      DmaListHead         if return code is TRUE and if DmaListHead == NULL 
//                          on entry, then DmaListHead = DmaNode
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaAddDmaNodeToList (
            PDmaNodeT *DmaListHead,
            PDmaNodeT DmaNode)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaAddDmaNodeToList()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DmaListHead = %08X\n", DmaListHead);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DmaNode = %08X\n", DmaNode);

    // Assume failure
    retcode = FALSE;

    // Add node to the end of the list
    if (*DmaListHead == NULL)
    {
        // Create a new list head
        *DmaListHead = DmaNode;
    }
    else
    {
        PDmaNodeT cur;
        PDmaNodeT prev;

        // Find the end of the list
        prev = *DmaListHead;
        cur = prev->Next;
        while (cur != NULL)
        {
            prev = cur;
            cur = cur->Next;
        }

        // Add node to end
        prev->Next = DmaNode;
        DmaNode->Next = NULL;
    }

    // OK
    retcode = TRUE;

    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaAddDmaNodeToList() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaAllocAndAddDmaNodeToList()
//
//      Allocates a new DMA Node and adds it to the end of the
//      specified list.
//
// Entry:
//
//      DmaListHead         Head of the list
//
//      DescriptorSize      size of the descriptor to allocate in DMA memory
//
//      DataSize            size of the data buffer to allocate in DMA memory
//
// Exit:
//
//      TRUE                success
//      FALSE               error
//
//      DmaNode             if return code is TRUE, then this holds the
//                          pointer to the DmaNodeT structure allocated
//                          and put at the end of the DMA List.
//
//      DmaListHead         if return code is TRUE and if DmaListHead == NULL 
//                          on entry, then DmaListHead = DmaNode
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaAllocAndAddDmaNodeToList (
            PDmaNodeT *DmaListHead,
            __u32 DescriptorSize,
            __u32 DataSize,
            PDmaNodeT *DmaNode)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    BOOLEAN ccode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaAllocAndAddDmaNodeToList()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DmaListHead = %08X\n", DmaListHead);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DescriptorSize = %08X\n", DescriptorSize);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: DataSize = %08X\n", DataSize);

    // Assume failure
    retcode = FALSE;
    *DmaNode = NULL;

    // Allocate a new node
    ccode = DmaAllocDmaNode (DescriptorSize, DataSize, DmaNode);
    if (ccode != TRUE)
        goto AllocError;

    // Add node to the end of the list
    ccode = DmaAddDmaNodeToList (DmaListHead, *DmaNode);
    if (ccode != TRUE)
        goto ListError;

    // OK
    retcode = TRUE;

    DBOUT1(DBG_LF_EXIT_PARAMS, "ohcitest: *DmaNode = %08X\n", *DmaNode);

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaAllocAndAddDmaNodeToList() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

AllocError:
    printf ("ERROR: Unable to allocate dma node\n");
    goto ExitWithError;

ListError:
    printf ("ERROR: Unable to add dma node %08X to list %08X\n", (__u32)DmaNode, (__u32)*DmaListHead);
    goto ExitWithError;

ExitWithError:

    // Deallocate anything already allocated
    if (*DmaNode != NULL)
    {
        // free the node
        DmaFreeDmaNode (*DmaNode);
        *DmaNode = NULL;
    }
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaFreeDmaNode()
//
//      Frees the DMA node.
//
// Entry:
//
//      DmaNode         Node to free
//
// Exit:
//
//      TRUE            success
//      FALSE           error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaFreeDmaNode (
            PDmaNodeT DmaNode)

{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaFreeDmaNode()\n");
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaNode = %08X\n", DmaNode);

    // Assume failure
    retcode = FALSE;

    // Is this a valid node?
    if (DmaNode == NULL)
        goto InvalidNodeError;

    // Free the descriptor
    if (DmaNode->DescBlkHandle)
        DmaFreeBlock (DmaNode->DescBlkHandle);

    // Free the data
    if (DmaNode->DataBlkHandle)
        DmaFreeBlock (DmaNode->DataBlkHandle);

    // Free the node
    free (DmaNode);

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaFreeDmaNode() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

InvalidNodeError:
    printf ("ERROR: DMA Node %08X is invalid\n", (__u32)DmaNode);
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaRemoveDmaNodeFromList()
//
//      Removes the specified node from the specified list
//
// Entry:
//
//      DmaListHead     pointer to the list head variable
//
//      DmaNode         Node to delete from the list
//
// Exit:
//
//      TRUE            success
//      FALSE           error
//
//      DmaListHead     if return code is TRUE and if DmaListHead == DmaNode
//                      on entry, then DmaListHead = DmaListHead->Next
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaRemoveDmaNodeFromList (
            PDmaNodeT *DmaListHead,
            PDmaNodeT DmaNode)

{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaRemoveDmaNodeFromList()\n");
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: *DmaListHead = %08X\n", *DmaListHead);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaNode = %08X\n", DmaNode);

    // Assume failure
    retcode = FALSE;

    // Is this a valid node?
    if (DmaNode == NULL)
        goto InvalidNodeError;

    // Is it the head?
    if (*DmaListHead == DmaNode)
    {
        // Remove the head node and replace with next node
        *DmaListHead = (*DmaListHead)->Next;
        DmaNode->Next = NULL;
    }
    else
    {
        PDmaNodeT prev;
        PDmaNodeT cur;

        // Search for the node
        prev = *DmaListHead;
        cur = prev->Next;
        while (prev != NULL && cur != NULL && cur != DmaNode)
        {
            // Advance to next node
            prev = cur;
            cur = cur->Next;
        }
        if (cur != DmaNode)
            goto NotFoundError;

        // Remove the node from the list
        prev->Next = cur->Next;
        cur->Next = NULL;
    }

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaRemoveDmaNodeFromList() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

InvalidNodeError:
    printf ("ERROR: DMA Node %08X is invalid\n", (__u32)DmaNode);
    goto Exit;

NotFoundError:
    printf ("ERROR: Cannot find DMA Node %08X in DMA List %08X\n", (__u32)DmaNode, (__u32)*DmaListHead);
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaRemoveAndFreeDmaNodeFromList()
//
//      Removes the specified node from the specified list
//      and then frees the node.
//
// Entry:
//
//      DmaListHead     pointer to the list head variable
//
//      DmaNode         Node to delete from the list
//
// Exit:
//
//      TRUE            success
//      FALSE           error
//
//      DmaListHead     if return code is TRUE and if DmaListHead == DmaNode
//                      on entry, then DmaListHead = DmaListHead->Next
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaRemoveAndFreeDmaNodeFromList (
            PDmaNodeT *DmaListHead,
            PDmaNodeT DmaNode)

{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    BOOLEAN ccode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaRemoveAndFreeDmaNodeFromList()\n");
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: *DmaListHead = %08X\n", *DmaListHead);
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaNode = %08X\n", DmaNode);

    // Assume failure
    retcode = FALSE;

    // Remove the node from the list
    ccode = DmaRemoveDmaNodeFromList (DmaListHead, DmaNode);
    if (ccode != TRUE)
        goto RemoveError;

    // Free the node
    ccode = DmaFreeDmaNode (DmaNode);
    if (ccode != TRUE)
        goto FreeError;

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaRemoveAndFreeDmaNodeFromList() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

RemoveError:
    printf ("ERROR: Unable to remove DMA Node %08X from List %08X\n", (__u32)DmaNode, (__u32)*DmaListHead);
    goto Exit;

FreeError:
    printf ("ERROR: Unable to free DMA Node %08X\n", (__u32)DmaNode);
    goto Exit;


    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// DmaDestroyNodeList()
//
//      Deallocates an entire DMA node list
//
// Entry:
//
//      DmaListHead     pointer to the list head variable
//
// Exit:
//
//      TRUE            success
//      FALSE           error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DmaDestroyNodeList (
            PDmaNodeT DmaListHead)

{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    PDmaNodeT cur;
    PDmaNodeT next;
    BOOLEAN ccode;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DmaDestroyNodeList()\n");
    DBOUT1 (DBG_LF_ENTRY_PARAMS, "ohcitest: DmaListHead = %08X\n", DmaListHead);

    // Assume failure
    retcode = FALSE;

    // Are we free already
    if (DmaListHead == NULL)
    {
        retcode = TRUE;
        goto Exit;
    }

    // Free all of the nodes from the list
    cur = DmaListHead;
    while (cur != NULL)
    {
        // Remember the next
        next = cur->Next;

        // Free the current
        ccode = DmaRemoveAndFreeDmaNodeFromList (&DmaListHead, cur);
        if (ccode != TRUE)
            goto FreeError;

        // Advance to next node
        cur = next;
    }

    // OK
    retcode = TRUE;

Exit:
    DBOUT1 (DBG_LF_EXIT, "ohcitest: DmaDestroyNodeList() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

FreeError:
    printf ("ERROR: Cannot free DMA node %08X on list %08X\n", (__u32)cur, (__u32)DmaListHead);
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// ReadQuadlet()
//
//      Reads a quadlet from the specified address (NodeID:AddrOffsetHigh:AddrOffsetLow)
//
// Entry:
//
//      NodeId          device ID of the node on the bus from which to read
//
//      AddrOffsetHigh  Address offset high of the quadlet to read from NodeId
//
//      AddrOffsetLow   Address offset low of the quadlet to read from NodeId
//
// Exit:
//
//      TRUE            success
//      FALSE           error
//
//      QuadletValue    if return value is TRUE, then this is the QUADLET 
//                      value read from the node
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN ReadQuadlet (
            __u32 NodeId, 
            __u32 AddrOffsetHigh, 
            __u32 AddrOffsetLow,
            QUADLET *QuadletValue)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    BOOLEAN ccode;
    QUADLET q;
    PDmaNodeT requestDmaNode;
    PDmaNodeT responseDmaNode;
    PAsync_OUTPUT_LAST_Immediate_descriptor rqDesc;
    PAsync_INPUT_MORE_descriptor rpDesc;
    PQuadlet_read_request_transmit rqPkt;
    PQuadlet_read_response_receive rpPkt;

    DBOUT (DBG_LF_ENTRY, "ohcitest: ReadQuadlet()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: NodeId = %08X\n", NodeId);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: AddrOffsetHigh = %08X\n", AddrOffsetHigh);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: AddrOffsetLow = %08X\n", AddrOffsetLow);

    // Assume failure
    retcode = FALSE;
    *QuadletValue = 0xFFFFFFFF;
    responseDmaNode = NULL;
    requestDmaNode = NULL;

    // Allocate a DMA node for a INPUT_MORE descriptor and receive buffer
    ccode = DmaAllocDmaNode (
                sizeof (Async_INPUT_MORE_descriptor),
                sizeof (Quadlet_read_response_receive),
                &responseDmaNode);
    if (ccode != TRUE)
        goto AllocError;

    // Fill in response INPUT_MORE descriptor
    rpDesc = (PAsync_INPUT_MORE_descriptor)responseDmaNode->DescBlkHandle->BlockAddr;
    rpDesc->reqCount = sizeof (Quadlet_read_response_receive);
    rpDesc->b = BRANCH_ALWAYS;
    rpDesc->i = INTERRUPT_NEVER;
    rpDesc->key = KEY_ADDRESS;
    rpDesc->s = STATUS_CTRL_WRITE;
    rpDesc->cmd = CMD_INPUT_MORE;
    rpDesc->dataAddress = responseDmaNode->DataBlkHandle->PhysAddr >> 2;
    rpDesc->Z = Z_LAST;
    rpDesc->branchAddress = 0;
    #ifdef DEBUG
    {
        if (DEBUGFLAGS & DBG_LF_INTERNAL)
        {
            DBGDumpBuffer (
                "Async_INPUT_MORE_descriptor", 
                (BYTE *)rpDesc, 
                sizeof(Async_INPUT_MORE_descriptor), 
                sizeof(QUADLET),
                4);
        }
    }
    #endif

    // Stop the async response receive dma
    q = 0;
    ((AsyncResponseReceiveContextControl_REG *)&q)->fld.run = 1;
    REG_WRITE (AsyncResponseReceiveContextControlClear, q);
    usleep (1000);  // wait for dma to stop

    // Make sure Rx response dma run, active and dead are 0
    q = REG_READ (AsyncResponseReceiveContextControlSet);
    if (    ((AsyncResponseReceiveContextControlSet_REG *)&q)->fld.run  ||
            ((AsyncResponseReceiveContextControlSet_REG *)&q)->fld.active ||
            ((AsyncResponseReceiveContextControlSet_REG *)&q)->fld.dead)
        goto RxNotReadyError;

    // Set the async response receive dma command ptr
    q = 0;
    ((AsyncResponseReceiveCommandPtr_REG *)&q)->fld.descriptorAddress = responseDmaNode->DescBlkHandle->PhysAddr >> 4;
    ((AsyncResponseReceiveCommandPtr_REG *)&q)->fld.Z = 1;
    REG_WRITE (AsyncResponseReceiveCommandPtr, q);

    // Start the async response receive dma
    q = 0;
    ((AsyncResponseReceiveContextControl_REG *)&q)->fld.run = 1;
    REG_WRITE (AsyncResponseReceiveContextControlSet, q);

    // Allocate a DMA node for a OUTPUT_LAST_immediate descriptor and Transmit buffer
    ccode = DmaAllocDmaNode (
                sizeof (Async_OUTPUT_LAST_Immediate_descriptor),
                0,
                &requestDmaNode);
    if (ccode != TRUE)
        goto AllocError;

    // Fill in OUTPUT_LAST_immediate descriptor
    rqDesc = (PAsync_OUTPUT_LAST_Immediate_descriptor) requestDmaNode->DescBlkHandle->BlockAddr;
    rqDesc->reqCount = sizeof (Quadlet_read_request_transmit);
    rqDesc->b = BRANCH_ALWAYS;
    rqDesc->i = INTERRUPT_NEVER;
    rqDesc->key = KEY_IMMEDIATE;
    rqDesc->cmd = CMD_OUTPUT_LAST;
    rqDesc->Z = Z_LAST;
    rqDesc->branchAddress = 0;

    // Set up the request packet (contained in the OUTPUT_LATE_Immedate descriptor)
    rqPkt = (PQuadlet_read_request_transmit) &(rqDesc->first_quadlet);
    rqPkt->tCode = tCode_READ_QUADLET_REQUEST;
    rqPkt->rt = retry_X;
    rqPkt->tLabel = GetNextTLabel();
    rqPkt->spd = spd_100;
    rqPkt->srcBusID = srcBusID_DEFAULT;
    rqPkt->destinationOffsetHigh = AddrOffsetHigh;
    rqPkt->destinationID = NodeId | (0x3FF<<6);
    rqPkt->destinationOffsetLow = AddrOffsetLow;
    #ifdef DEBUG
    {
        if (DEBUGFLAGS & DBG_LF_INTERNAL)
        {
            DBGDumpBuffer (
                "Async_OUTPUT_LAST_Immediate_descriptor", 
                (BYTE *)rqDesc, 
                sizeof(Async_OUTPUT_LAST_Immediate_descriptor), 
                sizeof(QUADLET),
                4);
        }
    }
    #endif

    // Stop the async request transmit dma
    q = 0;
    ((AsyncRequestTransmitContextControl_REG *)&q)->fld.run = 1;
    REG_WRITE (AsyncRequestTransmitContextControlClear, q);
    usleep (1000);  // wait for dma to stop

    // Make sure Tx request dma run, active and dead are 0
    q = REG_READ (AsyncRequestTransmitContextControlSet);
    if (    ((AsyncRequestTransmitContextControlSet_REG *)&q)->fld.run  ||
            ((AsyncRequestTransmitContextControlSet_REG *)&q)->fld.active ||
            ((AsyncRequestTransmitContextControlSet_REG *)&q)->fld.dead)
        goto TxNotReadyError;

    // Set the async request transmit command ptr
    q = 0;
    ((AsyncRequestTransmitCommandPtr_REG *)&q)->fld.descriptorAddress = requestDmaNode->DescBlkHandle->PhysAddr >> 4;
    ((AsyncRequestTransmitCommandPtr_REG *)&q)->fld.Z = 2;
    REG_WRITE (AsyncRequestTransmitCommandPtr, q);

    // Prevent event process from processing any events
    // until we are ready to wait for it
    gSharedMem->IgnoreEvents = TRUE;

    // Tell event process we are waiting for a ARRS or RSPkt interrupt
    gSharedMem->ParentPausedOnEvent.EventWaitingFor = ARRS_BIT | RSPkt_BIT;

    // Transmit the quadlet read request
    q = 0;
    ((AsyncRequestTransmitContextControl_REG *)&q)->fld.run = 1;
    REG_WRITE (AsyncRequestTransmitContextControlSet, q);

    // Wait for Response
    gSharedMem->IgnoreEvents = FALSE;
    ccode = PauseOnEvent (ARRS_BIT | RSPkt_BIT, 2);
    if (ccode != TRUE)
        goto ResponseError;

    // Extract the quadlet value from the response packet
    rpPkt = (PQuadlet_read_response_receive) responseDmaNode->DataBlkHandle->BlockAddr;
    #ifdef DEBUG
    {
        if (DEBUGFLAGS & DBG_LF_INTERNAL)
        {
            DBGDumpBuffer (
                "Quadlet_read_response_receive", 
                (BYTE *)rpPkt, 
                sizeof(Quadlet_read_response_receive), 
                sizeof(QUADLET),
                4);
        }
    }
    #endif
    *QuadletValue = ntohl(rpPkt->quadlet_data);

    // OK
    retcode = TRUE;

    DBOUT1 (DBG_LF_EXIT_PARAMS, "ohcitest: *QuadletValue = %08X\n", *QuadletValue);

Exit:

    // Dealloc local resources
    if (responseDmaNode)
        DmaFreeDmaNode (responseDmaNode);
    if (requestDmaNode)
        DmaFreeDmaNode (requestDmaNode);

    // Stop DMAs -- just in case
    q = 0;
    ((AsyncResponseReceiveContextControl_REG *)&q)->fld.run = 1;
    REG_WRITE (AsyncResponseReceiveContextControlClear, q);
    REG_WRITE (AsyncRequestTransmitContextControlClear, q);
    usleep (1000);  // wait for dma to stop

    DBOUT1 (DBG_LF_EXIT, "ohcitest: ReadQuadlet() exit, retcode = %s\n", BoolToStr(retcode));
    return retcode;

AllocError:
    printf ("ERROR: Unable to allocate DMA node or descriptor or data\n");
    goto Exit;

RxNotReadyError:
    printf ("ERROR: Receive DMA is not ready\n");
    goto Exit;

TxNotReadyError:
    printf ("ERROR: Transmit DMA is not ready\n");
    goto Exit;

ResponseError:
    printf ("ERROR: Timeout waiting for QUADLET read response from node %08X\n", NodeId);
    q = REG_READ (AsyncRequestTransmitContextControlSet);
    printf ("Async Request Transmit Context Control = %08X\n", q);
    q = REG_READ (AsyncRequestTransmitCommandPtr);
    printf ("Async Request Transmit Command Ptr  = %08X\n", q);
    printf ("Request xferStatus = %04X\n", rqDesc->xferStatus);
    printf ("Request timeStamp = %04X\n", rqDesc->timeStamp);
    q = REG_READ (AsyncResponseReceiveContextControlSet);
    printf ("Async Response Receive Context Control = %08X\n", q);
    q = REG_READ (AsyncResponseReceiveCommandPtr);
    printf ("Async Response Receive Command Ptr  = %08X\n", q);
    printf ("Response xferStatus = %04X\n", rpDesc->xferStatus);
    printf ("Response resCount = %04X\n", rpDesc->resCount);
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// GetNextTLabel()
//
//      Gets the next tlable which we can use to wait for a response
//
// Entry:
//
//      NONE
//
// Exit:
//
//      Next tlabel
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
__u32 GetNextTLabel (void)
{
    static __u32 nextTLabel = MAX_TLABEL-1;

    nextTLabel = (nextTLabel + 1) % MAX_TLABEL;
    return (nextTLabel);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// HostToNetU32Buffer()
//
//      Converts an buffer of __u32s to network order.
//
// Entry:
//
//      Buffer      Buffer of __u32s to convert
//
//      Size        Size of Buffer in bytes
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void HostToNetU32Buffer (__u32 *Buffer, int Size)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    __u32 *p; 
    int i; 

    DBOUT (DBG_LF_ENTRY, "ohcitest: HostToNetU32Buffer()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: Buffer = %08X\n", Buffer);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: Size = %08X\n", Size);


    for (p=Buffer, i=0; i < (Size/sizeof(__u32)); p++, i++)
    {
        #ifdef DEBUG
        __u32 temp;
        temp = *p;
        #endif
        *p = htonl(*p);
        DBOUT2 (DBG_LF_INTERNAL, "ohcitest: byteswap %08X to %08X\n", temp, *p);
    }

    DBOUT (DBG_LF_EXIT, "ohcitest: HostToNetU32Buffer() exit\n");

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// NetToHostU32Buffer()
//
//      Converts an buffer of __u32s to host order.
//
// Entry:
//
//      Buffer      Buffer of __u32s to convert
//
//      Size        Size of Buffer in bytes
//
// Exit:
//
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void NetToHostU32Buffer (__u32 *Buffer, int Size)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    __u32 *p; 
    int i; 

    DBOUT (DBG_LF_ENTRY, "ohcitest: NetToHostU32Buffer()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: Buffer = %08X\n", Buffer);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: Size = %08X\n", Size);

    for (p=Buffer, i=0; i < (Size/sizeof(__u32)); p++, i++)
    {
        #ifdef DEBUG
        __u32 temp;
        temp = *p;
        #endif
        *p = ntohl(*p);
        DBOUT2 (DBG_LF_INTERNAL, "ohcitest: byteswap %08X to %08X\n", temp, *p);
    }

    DBOUT (DBG_LF_EXIT, "ohcitest: NetToHostU32Buffer() exit\n");

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  ReadConfigRom
//
//      Reads the configuration ROM of the specified 1394 device into
//      the specified buffer.
//
//  Entry:
//
//      szDeviceName        name of device to open and read config rom
//      ConfigRomBuf        place to put QUADLETs read from device
//      ConfigRomBufMaxSize the allocated size of ConfigRomBuf
//
//  Exit:
//
//      TRUE                Configuration ROM has been read
//
//      FALSE               Unabled to read the configuration ROM
//
//      if the return value is TRUE:
//
//      BytesRead           how many bytes were read into ConfigRomBuf.
//                          This is less than or equal to ConfigRomBufMaxSize.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN ReadConfigRom (
            __u32 NodeId, 
            QUADLET *ConfigRomBuf, 
            __u32 ConfigRomBufMaxSize, 
            __u32 *BytesRead)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    BOOLEAN ccode;
    QUADLET q;
    int i;
    __u32 numQuadsLeft;
    ADDRESS_OFFSET curAddr;

    DBOUT (DBG_LF_ENTRY, "ohcitest: ReadConfigRom()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: NodeId = %08X\n", NodeId);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: ConfigRomBuf\n", ConfigRomBuf);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: ConfigRomBufMaxSize\n", ConfigRomBufMaxSize);

    // Assume failure
    retcode = FALSE;
    *BytesRead = 0;

    // Let's setup a pointer to our address range variable for dialog box
    i = 0;
    curAddr.Off_High = CONFIG_ROM_ADDR.Off_High;
    curAddr.Off_Low = CONFIG_ROM_ADDR.Off_Low;
    do
    {
        // Read the next QUADLET
        ccode = ReadQuadlet (NodeId, curAddr.Off_High, curAddr.Off_Low+(i*sizeof(QUADLET)), &q);
        if (ccode != TRUE)
            goto ReadQuadletError;
        DBOUT4 (DBG_LF_INTERNAL, "ohcitest: Quadlet @ %02X:%04X:%08X = %08X\n", NodeId, curAddr.Off_High, curAddr.Off_Low+(i*sizeof(QUADLET)), q); 

        // Store the quadlet in the config rom buffer
        *ConfigRomBuf = q;

        // Is this the first quadlet?
        if (i == 0)
        {
            // Is this a valid 1394 cfg rom?
            // NOTE: minimal cfg rom will have 0x01xxxxxx.
            // Since we already set numQuadsLeft to 0, there
            // is no reason for a special case.
            if ((q & 0xFF000000) == 0x04000000)
            {
                numQuadsLeft = (q & 0x00FF0000) >> 16;
                DBOUT1 (DBG_LF_INTERNAL, "ohcitest: numQuadsLeft = %08X\n", numQuadsLeft); 
            }
        }
        else
        {
            // One less quadlet to worry about
            if (numQuadsLeft != 0)
                numQuadsLeft--;
        }

        // Advance
        ConfigRomBuf++;
        i++;
        *BytesRead = *BytesRead + sizeof(QUADLET);

    } while (ccode == TRUE && numQuadsLeft && *BytesRead <= ConfigRomBufMaxSize);

    // OK
    retcode = TRUE;
    DBOUT1(DBG_LF_EXIT_PARAMS, "ohcitest: *BytesRead = %08X\n", *BytesRead);

Exit:

    DBOUT1(DBG_LF_EXIT, "ohcitest: ReadConfigRom() exit, retcode = %s\n", BoolToStr(retcode));
    return (retcode);

ReadQuadletError:
    printf ("ERROR:  Cannot read quadlet at %02X:%04X:%08X\n", NodeId, 0xFFFF, 0xF0000400+(i*sizeof(QUADLET)));
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  DumpConfigRom
//
//      Displays the raw quadlet data contained in the specified 
//      configuration ROM buffer.
//
//  Entry:
//
//      CfgRomBuf           quadlet buffer containing the configuration ROM image
//
//      CfgRomBufSize       size in bytes of CFGRomBuf
//
//  Exit:
//
//      TRUE                operation succeeded
//
//      FALSE               operation failed
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DumpConfigRom (
            QUADLET *CfgRomBuf, 
            __u32 CfgRomBufSize)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    BOOLEAN retcode;
    __u32 i;
    QUADLET *pq;

    DBOUT (DBG_LF_ENTRY, "ohcitest: DumpConfigRom()\n");
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: CfgRomBuf = %08X\n", CfgRomBuf);
    DBOUT1(DBG_LF_ENTRY_PARAMS, "ohcitest: CfgRomBufSize = %08X\n", CfgRomBufSize);

    // Assume failure
    retcode = FALSE;

    // Check if the size is in quadlets
    if (CfgRomBufSize % sizeof(QUADLET))
        goto InvalidSize;

    // RAW Dump of buffer
    printf ("\n");
    printf ("Raw Data Dump of the Configuration ROM\n\n");
    printf ("1394 Addr Off    Data\n");
    printf ("-------------    --------  --------  --------  -------- ");
    for (i=0, pq=CfgRomBuf; i<CfgRomBufSize; i+=4, pq++)
    {
        // Display the quadlet
        if ((i % 16) == 0)
        {
            printf ("\n%04X:%08X    ", 0xFFFF, (int)(0xF0000400+i));
        }
        printf ("%08X  ", *pq);
    }
    printf ("\n");

    // OK
    retcode = TRUE;

Exit:

    DBOUT1 (DBG_LF_EXIT, "DumpConfigRom(), retcode = %s\n", BoolToStr(retcode));
    return (retcode);

InvalidSize:
    printf ("\n\nERROR:  Buffer size %d is not divisible by %d\n\n", (int)CfgRomBufSize, sizeof(QUADLET));
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  DecodeConfigRom
//
//      Decodes the specified configuration ROM buffer according to the 
//      1212, 1394 and SBP-2 specifications.
//
//  Entry:
//
//      CfgRomBuf           quadlet buffer containing the configuration ROM image
//
//      CfgRomBufSize       size in bytes of CFGRomBuf
//
//  Exit:
//
//      TRUE                operation succeeded
//
//      FALSE               operation failed
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BOOLEAN DecodeConfigRom (
            QUADLET * CfgRomBuf, 
            __u32 CfgRomBufSize)
{
//    #undef DEBUGFLAGS
//    #define DEBUGFLAGS DebugThisRoutine

    // Block Types
    #define BT_NONE         0
    #define BT_DIRECTORY    1
    #define BT_LEAF         2

    // Directory Types
    // NOTE:  Use the 1212 key_value definitions (e.g. Unit_Directory) and
    // define the directory types not defined in 1212
    #define DT_UNKNOWN      0
    #define DT_ROOT         -1

    // Leaf Types
    // NOTE:  Use the 1212 key_value definitions (e.g. Textual_Descriptor) and
    // define the directory types not defined in 1212
    #define LT_UNKNOWN      0

    // Block Information Structure
    //
    // For every quadlet entry in the CFG ROM there will be a corresponding entry
    // in the Block Information Table.
    //
    // Used to determine if the given CFG ROM entry is the beginning of 
    // a new block.  The BlockType will be BT_NONE if the entry is just an
    // entry in the currrent block.  Otherwise it will specify what type 
    // of block it is.  BT_DIRECTORY indicates that this is a directory block.
    // BT_LEAF indicates that this is a leaf block.
    //
    //  ReferringDirEntry is a copy of the quadlet (directory entry) in a previous block
    // that refers to this block.
    //
    // ReferringDirEntryIdx the the index of the directory entry in a previous block
    // that refered to this block.  Need to this to show where this was referred from.
    //
    typedef struct BlockInfoStruct
    {
        __u32       BlockType;
        DirEntryT   ReferringDirEntry;
        __u32       ReferringDirEntryIdx;
    } BlockInfoT, *PBlockInfoT;

    // Helper macros for repeated operations.  Could make these functions, but
    // I like to hear the purists gripe.  Plus it's easier to change if they
    // are defined right here.
    #define INC_PLACE_HOLDERS {cr++; bi++; curTabIdx++;}
    #define INC_PLACE_HOLDERS2 (cr++, bi++, curTabIdx++)
    #define PRINT_ADDR_OFF_VALUE printf ("%04X:%08X %08X ", CONFIG_ROM_ADDR.Off_High, CONFIG_ROM_ADDR.Off_Low+(curTabIdx*sizeof(QUADLET)), *cr);
    #define CALC_LOW_PART_CFG_ROM_INDIRECT_ADDR_OFF(io) CONFIG_ROM_ADDR.Off_Low+(((curTabIdx+io)<maxTabIdx)?((curTabIdx+io)*sizeof(QUADLET)):0x0FFFFFFF)
    #define CALC_LOW_PART_CSR_BASE_ADDR_OFF(co) CSR_BASE_ADDR.Off_Low+(co*sizeof(QUADLET))
    #define SET_BLOCK_INFO_INDIRECT(io,bt,de,dei) if((bi+io)<biEnd){bi[io].BlockType=bt;bi[io].ReferringDirEntry.u.q=de;bi[io].ReferringDirEntryIdx=dei;}
    #define PRINT_HEADER_INFO printf ("length=%04X, crc=%04X", ((PBlockHeaderT)cr)->length, ((PBlockHeaderT)cr)->crc);

    BOOLEAN retcode;
    __u32 curTabIdx;
    __u32 maxTabIdx;
    QUADLET * cr;
    QUADLET * crEnd;
    PBlockInfoT blockInfo;
    PBlockInfoT bi;
    PBlockInfoT biEnd;
    PCfgRomHeaderT romHdr;
    __u32 curBlockType;
    PDirEntryT dirEntry;
    __u32 curUnitSpecId;
    __u32 curUnitSwVersion;

    DBOUT (DBG_LF_ENTRY, "DecodeConfigRom()\n");

    // Assume failure
    retcode = FALSE;

    // Init vars
    curTabIdx = 0;
    maxTabIdx = CfgRomBufSize/sizeof(QUADLET);
    blockInfo = NULL;
    cr = CfgRomBuf;
    crEnd = cr + maxTabIdx;
    bi = NULL;
    biEnd = NULL;
    curUnitSpecId = 0;
    curUnitSwVersion = 0;

    // Check if the size is in quadlets
    if (CfgRomBufSize % sizeof(QUADLET))
        goto InvalidSize;

    // Allocate a helper block type array which has a 1:1 correspondence
    // to each quadlet in the CFG ROM.
    // NOTE:  because the CFG ROM is hierarchical, we must
    // save information somewhere about blocks that are
    // referred to from other blocks during parsing.  Otherwise, when
    // we get to that block, we will not be able to distinquish the 
    // block's type and how to decode it.
    blockInfo = (PBlockInfoT)calloc (sizeof(BlockInfoT), maxTabIdx);
    if (blockInfo == NULL)
        goto AllocError;
    bi = blockInfo;
    biEnd = bi + maxTabIdx;

    // Print header
    printf ("\n");
    printf ("Decode of the Configuration ROM\n\n");
    printf ("1394 Addr Off Quadlet  Meaning\n");
    printf ("------------- -------- --------------------------------------------------\n");

    // Is this a minimal CFG ROM?
    romHdr = (PCfgRomHeaderT)cr;
    if (romHdr->u.mcr.info_length == 0x01)
        goto MinimalCfgRom;

    // Is this a valid general 1394 CFG ROM?
    if (romHdr->u.gcr.info_length != 0x04)
        goto InvalidRomHdr;

    // Display CFG ROM Header
    printf ("\nConfiruation ROM Header\n");
    PRINT_ADDR_OFF_VALUE
    printf ("info_length=%02X, crc_length=%02X, rom_crc_value=%04X\n", 
        romHdr->u.gcr.info_length, romHdr->u.gcr.crc_length, romHdr->u.gcr.rom_crc_value);
    INC_PLACE_HOLDERS

    // Check the bus_name in the bus_info_block
    if (romHdr->u.gcr.bus_info_block.bus_name != BUS_NAME_1394)
        goto InvalidRomHdr;

    // Display bus_info_block
    printf ("\nBus_Info_Block\n");
    PRINT_ADDR_OFF_VALUE
    printf ("bus_name=%08X (\"", romHdr->u.gcr.bus_info_block.bus_name);
    DumpQuadletAsChars (romHdr->u.gcr.bus_info_block.bus_name);
    printf ("\")\n");
    INC_PLACE_HOLDERS
    PRINT_ADDR_OFF_VALUE
    printf ("irmc=%X, cmc=%X, isc=%X, bmc=%X, cyc_clk_acc=%02X, max_rec=%X\n",
        romHdr->u.gcr.bus_info_block.irmc,
        romHdr->u.gcr.bus_info_block.cmc,
        romHdr->u.gcr.bus_info_block.isc,
        romHdr->u.gcr.bus_info_block.bmc,
        romHdr->u.gcr.bus_info_block.cyc_clk_acc,
        romHdr->u.gcr.bus_info_block.max_rec);
    INC_PLACE_HOLDERS
    PRINT_ADDR_OFF_VALUE
    printf ("node_vendor_id=%06X, chip_id_hi=%02X\n",
        romHdr->u.gcr.bus_info_block.node_vendor_id,
        romHdr->u.gcr.bus_info_block.chip_id_hi);
    INC_PLACE_HOLDERS
    PRINT_ADDR_OFF_VALUE
    printf ("chip_id_lo=%08X\n", romHdr->u.gcr.bus_info_block.chip_id_lo);
    INC_PLACE_HOLDERS

    // We should be at the begining of the root_directory.
    // It is possible that the CFG ROM consist only of
    // a ROM Header and BIB.
    if (cr >= crEnd)
        goto OK;

    // Decode the rest of the CFG ROM starting at the root_directory
    for (   bi->BlockType = BT_DIRECTORY, bi->ReferringDirEntry.u.q = DT_ROOT, curBlockType = BT_DIRECTORY;
            cr < crEnd;
            INC_PLACE_HOLDERS2)
    {
        // Are we at the start of a new block?
        switch (bi->BlockType)
        {
            // It's a element of a block
            case BT_NONE:

                // Print the current address and quadlet value
                PRINT_ADDR_OFF_VALUE

                // Are currently parsing a directory?
                // If so, we need to parse the directory entry
                if (curBlockType == BT_DIRECTORY)
                {
                    // Process according to the key_type
                    dirEntry = (PDirEntryT)cr;
                    //printf ("key=%02X ", dirEntry->general.key);
                    printf ("%s ", 
                            KeyToString(
                                dirEntry, 
                                curUnitSpecId, 
                                curUnitSwVersion));
                    switch (dirEntry->u.iv.key_type)
                    {
                        case KT_IMMEDIATE_VALUE:

                            // Remember the last Unit_Spec_Id and Unit_Sw_Version
                            // Used to differentiate protocol specific key_values
                            if (dirEntry->u.iv.key_value == Unit_Spec_Id)
                                curUnitSpecId = dirEntry->u.iv.value;
                            else if (dirEntry->u.iv.key_value == Unit_Sw_Version)
                                curUnitSwVersion = dirEntry->u.iv.value;

                            // Decode the immediate value
                            printf ("%s", 
                                    ImmediateValueToString
                                        (dirEntry, 
                                        curUnitSpecId, 
                                        curUnitSwVersion));

                            break;

                        case KT_CSR_OFFSET:
                            printf ("crc_offset=%06X (%04X:%08X)", 
                                dirEntry->u.co.csr_offset, 
                                CSR_BASE_ADDR.Off_High, 
                                CALC_LOW_PART_CSR_BASE_ADDR_OFF(dirEntry->u.co.csr_offset));
                            break;

                        case  KT_LEAF_INDIRECT_OFFSET:
                            printf ("ind_off=%06X (%04X:%08X)", 
                                dirEntry->u.io.indirect_offset, 
                                CONFIG_ROM_ADDR.Off_High, 
                                CALC_LOW_PART_CFG_ROM_INDIRECT_ADDR_OFF(dirEntry->u.io.indirect_offset));

                            // Set the referenced block info table entry to LEAF type
                            SET_BLOCK_INFO_INDIRECT(dirEntry->u.io.indirect_offset,BT_LEAF,dirEntry->u.q,curTabIdx)
                            break;

                        case KT_DIRECTORY_INDIRECT_OFFSET:
                            printf ("ind_off=%06X (%04X:%08X)", 
                                dirEntry->u.io.indirect_offset, 
                                CONFIG_ROM_ADDR.Off_High, 
                                CALC_LOW_PART_CFG_ROM_INDIRECT_ADDR_OFF(dirEntry->u.io.indirect_offset));

                            // Set the referenced block info table entry to DIRECTORY type
                            SET_BLOCK_INFO_INDIRECT(dirEntry->u.io.indirect_offset,BT_DIRECTORY,dirEntry->u.q,curTabIdx)
                            break;
                    }
                }
                // It's a leaf, just print the leaf data as characters
                else
                {
                    DumpQuadletAsChars (*cr);
                }
                break;

            // We're entering a LEAF or DIRECTORY block
            case BT_LEAF:
            case BT_DIRECTORY:

                // Print the name of the leaf or directory type
                printf ("\n%s", 
                        KeyToString(
                        &bi->ReferringDirEntry,
                        curUnitSpecId, 
                        curUnitSwVersion));

                // Print the reference address offset (if there is one)
                // NOTE:  If we're at the root, we don't have any
                // refernce to worry about.
                if (bi->ReferringDirEntry.u.q != DT_ROOT)
                {
                    printf (" referenced from %04X:%08X", 
                        CONFIG_ROM_ADDR.Off_High, 
                        CONFIG_ROM_ADDR.Off_Low+bi->ReferringDirEntryIdx*sizeof(QUADLET));
                }
                printf ("\n");

                // Print current address offset and value
                PRINT_ADDR_OFF_VALUE
                PRINT_HEADER_INFO

                // Set the context for the current block (DIRECTORY or LEAF)
                // This will let the parsing loop know that all of the quadlets
                // that follow this header will be either in a DIRECTORY or LEAF.
                curBlockType = bi->BlockType;
                break;
        }
        printf ("\n");
    }

OK:
    printf ("\n\n");

    // OK
    retcode = TRUE;

Exit:

    // Free resources
    if (blockInfo != NULL)
        free (blockInfo);

    DBOUT1 (DBG_LF_EXIT, "DecodeConfigRom, retcode = %08X\n", retcode);
    return (retcode);

InvalidSize:
    printf ("\n\nERROR:  Buffer size %d is not divisible by %d\n\n", CfgRomBufSize, sizeof(QUADLET));
    goto Exit;

AllocError:
    printf ("ERROR:  Cannot allocate memory.\n");
    goto Exit;

MinimalCfgRom:

    // Print minimal CFG ROM header and exit OK
    PRINT_ADDR_OFF_VALUE
    printf ("Minimal Configuration ROM, GUID = %06X\n", romHdr->u.mcr.vendor_id);
    retcode = TRUE;
    goto Exit;

InvalidRomHdr:
    PRINT_ADDR_OFF_VALUE
    printf ("Invalid Configuration ROM Header\n");
    for (INC_PLACE_HOLDERS2; cr < crEnd; INC_PLACE_HOLDERS2)
    {
        PRINT_ADDR_OFF_VALUE
        printf ("?????\n");
    }
    goto Exit;

    #undef DEBUGFLAGS
    #define DEBUGFLAGS DebugFlags
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  DumpQuadletAsChars
//
//      Prints the specified quadlet value as an array of characters.
//
//  Entry:
//
//      q       quadlet value to print
//
//  Exit:
//      NONE
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void DumpQuadletAsChars (
            QUADLET q)
{
    int i;
    BYTE ba[sizeof(QUADLET)];

    // Copy the quadlet to a byte array.
    // NOTE:  If little Endian, we must swap the quadlet 
    // because we will now treat as a byte array
    *(QUADLET *)ba = htonl (q);

    // Print the byte array as characters
    for (i=0; i<sizeof(QUADLET); i++)
    {
        if (MyIsPrint(ba[i]))
            printf ("%c", ba[i]);
        else
            printf (".");
    }
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  KeyValueToString
//
//      Translate the specified directory entry's key_value into a printable
//      string.  The translation is dependent on the specified UnitSpecId
//      and UnitSwVersion.
//
//  Entry:
//
//      DirEntry            quadlet value (directory entry) to translate
//
//      UnitSpecId      Unit_Spec_Id value of the directory entry
//
//      UnitSwVersion   Unit_Sw_Version value of the directory entry
//
//  Exit:
//
//      pointer to a character string specifying the translation of
//      the key_value in the directory entry.  This is a static string
//      and should not be modified by the calling routine.  Also, the
//      string is subject to change on the next call to this routine.
//      If you need to retain the string value, you must copy it
//      before calling this routine again.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
char * KeyToString (
            PDirEntryT DirEntry, 
            __u32 UnitSpecId, 
            __u32 UnitSwVersion)
{
    //static char bad_key_value_str[128];
    static char ret_str[128];

    // Initialize static return string
    ret_str[0] = '\0';

    // Root_Directory?
    if (DirEntry->u.q == DT_ROOT)
        return ("Root_Directory");

    switch (DirEntry->u.iv.key_value)
    {
        case Textual_Descriptor:
            strcat (ret_str, "Textual_Descriptor");
            break;
        case Bus_Dependent_Info:
            strcat (ret_str, "Bus_Dependent_Info");
            break;
        case Module_Vendor_Id:
            strcat (ret_str, "Module_Vendor_Id");
            break;
        case Module_Hw_Version:
            strcat (ret_str, "Module_Hw_Version");
            break;
        case Module_Spec_Id:
            strcat (ret_str, "Module_Spec_Id");
            break;
        case Module_Sw_Version:
            strcat (ret_str, "Module_Sw_Version");
            break;
        case Module_Dependent_Info:
            strcat (ret_str, "Module_Dependent_Info");
            break;
        case Node_Vendor_Info:
            strcat (ret_str, "Node_Vendor_Info");
            break;
        case Node_Hw_Version:
            strcat (ret_str, "Node_Hw_Version");
            break;
        case Node_Spec_Id:
            strcat (ret_str, "Node_Spec_Id");
            break;
        case Node_Sw_Version:
            strcat (ret_str, "Node_Sw_Version");
            break;
        case Node_Capabilities:
            strcat (ret_str, "Node_Capabilities");
            break;
        case Node_Unique_Id:
            strcat (ret_str, "Node_Unique_Id");
            break;
        case Node_Units_Extent:
            strcat (ret_str, "Node_Units_Extent");
            break;
        case Node_Memory_Extent:
            strcat (ret_str, "Node_Memory_Extent");
            break;
        case Node_Dependent_Info:
            strcat (ret_str, "Node_Dependent_Info");
            break;
        case Unit_Directory:
            strcat (ret_str, "Unit_Directory");
            break;
        case Unit_Spec_Id:
            strcat (ret_str, "Unit_Spec_Id");
            break;
        case Unit_Sw_Version:
            strcat (ret_str, "Unit_Sw_Version");
            break;
        case Unit_Dependent_Info:           // 0x14
        //case Logical_Unit_Number:     // 0x14 SBP-2
        //case Logical_Unit_Directory:  // 0xD4 SBP-2
        //case Management_Agent:            // 0x54 SBP-2
            if (UnitSpecId == SBP2_Unit_Spec_Id && UnitSwVersion == SBP2_Unit_Sw_Version)
            {
                switch (DirEntry->u.iv.key_type)
                {
                    case KT_IMMEDIATE_VALUE:
                        strcat (ret_str, "Logical_Unit_Number");
                        break;
                    case KT_DIRECTORY_INDIRECT_OFFSET:
                        strcat (ret_str, "Logical_Unit_Directory");
                        break;
                    case KT_CSR_OFFSET:
                        strcat (ret_str, "Management_Agent");
                        break;
                    case KT_LEAF_INDIRECT_OFFSET:
                    default:
                        break;

                }
            }
            else
            {
                strcat (ret_str, "Unit_Dependent_Info");
            }
            break;
        case Unit_Location:
            strcat (ret_str, "Unit_Location");
            break;
        case Unit_Poll_Mask:
            strcat (ret_str, "Unit_Poll_Mask");
            break;
        case Unit_Power_Requirements:
            strcat (ret_str, "Unit_Power_Requirements");
            break;
        case Model_Id:
            strcat (ret_str, "Model_Id");
            break;
        case Command_Set_Spec_Id:
            strcat (ret_str, "Command_Set_Spec_Id");
            break;
        case Command_Set:
            strcat (ret_str, "Command_Set");
            break;
        case Unit_Characteristics:
            strcat (ret_str, "Unit_Characteristics");
            break;
        case Command_Set_Revision:
            strcat (ret_str, "Commanmd_Set_Revision");
            break;
        case Firmware_Revision:
            strcat (ret_str, "Firmware_Revision");
            break;
//      case :
//          strcat (ret_str, "");
//          break;
        default:
            sprintf (ret_str+strlen(ret_str), "key=%02X (UNKNOWN)", DirEntry->u.general.key);
            break;
    }
    switch (DirEntry->u.iv.key_type)
    {
        case KT_LEAF_INDIRECT_OFFSET:
            strcat (ret_str, " leaf");
            break;
        case KT_DIRECTORY_INDIRECT_OFFSET:
            strcat (ret_str, " directory");
            break;
    }
    return (ret_str);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//  ImmediateValueToString
//
//      Translate the specified directory entry's value into a printable
//      string.  The translation is dependent on the specified UnitSpecId
//      and UnitSwVersion.
//
//  Entry:
//
//      DirEntry            quadlet value (directory entry) to translate
//
//      UnitSpecId      Unit_Spec_Id value of the directory entry
//
//      UnitSwVersion   Unit_Sw_Version value of the directory entry
//
//  Exit:
//
//      pointer to a character string specifying the translation of
//      the value in the directory entry.  This is a static string
//      and should not be modified by the calling routine.  Also, the
//      string is subject to change on the next call to this routine.
//      If you need to retain the string value, you must copy it
//      before calling this routine again.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
char * ImmediateValueToString (
            PDirEntryT DirEntry, 
            __u32 UnitSpecId, 
            __u32 UnitSwVersion)
{
    static char ret_str[128];
    __u32 tempvalue;

    // Initialize vars
    ret_str[0] = '\0';
    tempvalue = DirEntry->u.iv.value;

    // The meaning of the immediate value is dependent on the key_value
    switch (DirEntry->u.iv.key_value)
    {
        case Textual_Descriptor:
        case Bus_Dependent_Info:
        case Module_Dependent_Info:
        case Node_Unique_Id:
        case Node_Dependent_Info:
        case Unit_Directory:
        case Unit_Location:
            sprintf(ret_str, "ERROR:  No immediate value possible");
            break;
        case Module_Vendor_Id:
        case Module_Hw_Version:
        case Module_Spec_Id:
        case Module_Sw_Version:
        case Node_Vendor_Info:
        case Node_Hw_Version:
        case Node_Spec_Id:
        case Node_Sw_Version:
        case Unit_Spec_Id:
        case Unit_Sw_Version:
        case Command_Set_Spec_Id:
        case Command_Set:
        case Unit_Characteristics:
        case Command_Set_Revision:
        case Firmware_Revision:
            sprintf(ret_str, "%06X", DirEntry->u.iv.value);
            break;
        case Unit_Dependent_Info:       // 0x14
        //case Logical_Unit_Number:     // 0x14 SBP-2
            if (UnitSpecId == SBP2_Unit_Spec_Id && UnitSwVersion == SBP2_Unit_Sw_Version)
            {
                PLogicalUnitNumberT p;
                p = (PLogicalUnitNumberT)&tempvalue;
                sprintf(ret_str, "o=%X, device_type=%02X, lun=%04X", p->o, p->device_type, p->lun);
            }
            else
                sprintf(ret_str, "ERROR:  No immediate value possible");
            break;
        case Node_Capabilities:
            {
                PNodeCapabilitiesT p;
                p = (PNodeCapabilitiesT)&tempvalue;
                if (p->spt)
                    strcat (ret_str, "spt ");
                if (p->ms)
                    strcat (ret_str, "ms ");
                if (p->_int)
                    strcat (ret_str, "int ");
                if (p->ext)
                    strcat (ret_str, "ext ");
                if (p->bas)
                    strcat (ret_str, "bas ");
                if (p->prv)
                    strcat (ret_str, "prv ");
                if (p->_64)
                    strcat (ret_str, "64 ");
                if (p->fix)
                    strcat (ret_str, "fix ");
                if (p->lst)
                    strcat (ret_str, "lst ");
                if (p->drg)
                    strcat (ret_str, "drg ");
                if (p->elo)
                    strcat (ret_str, "elo ");
                if (p->atn)
                    strcat (ret_str, "atn ");
                if (p->off)
                    strcat (ret_str, "off ");
                if (p->ded)
                    strcat (ret_str, "ded ");
                if (p->init)
                    strcat (ret_str, "init ");
            }
            break;
        case Node_Units_Extent:
        case Node_Memory_Extent:
            {
                PNodeUnitsExtentT p;
                p = (PNodeUnitsExtentT)&tempvalue;
                sprintf (ret_str, "align=%02X, number=%04X", p->align, p->number);
            }
            break;
        case Unit_Poll_Mask:
            {
                PUnitPollMaskT p;
                p = (PUnitPollMaskT)&tempvalue;
                sprintf (ret_str, "poll_mask=%04X", p->poll_mask);
            }
            break;
        case Unit_Power_Requirements:
            sprintf(ret_str, "%06X deciwatts", DirEntry->u.iv.value);
            break;
        case KV_UNKNOWN:
        default:
            sprintf(ret_str, "value = %06X", DirEntry->u.iv.value);
            break;
    }

    return (ret_str);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// MyPause()
//
//		Paused for the specified number of seconds, milliseconds 
//		or microseconds
//
// Entry:
//
//		SecondsType		Describes the type of seconds
//						PAUSE_SECONDS		Seconds
//						PAUSE_MILLISECONDS	Milliseconds
//						PAUSE_MICROSECONDS	Microseconds
//
//		Seconds			The number of seconds, milliseconds or
//						microseconds to pause.
//
// Exit:
//
//      TRUE			success
//      FALSE			error
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
void MyPause (__u32 SecondsType, __u32 Seconds)
{
	if (Seconds == 0)
		return;

	switch (SecondsType)
	{
		case PAUSE_SECONDS:
		default:
			sleep (Seconds);
			break;

		case PAUSE_MILLISECONDS:
			usleep (Seconds*1000);
			break;

		case PAUSE_MICROSECONDS:
			usleep (Seconds);
			break;
	}
}

////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//// XXX()
////
////        XXX
////
//// Entry:
////
////        XXX
////
//// Exit:
////
////        TRUE    success
////        FALSE   error
////
////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//BOOLEAN XXX (void)
//{
////    #undef DEBUGFLAGS
////    #define DEBUGFLAGS DebugThisRoutine
//
//  BOOLEAN retcode;
//
//  DBOUT (DBG_LF_ENTRY, "ohcitest: XXX()\n");
//
//  // Assume failure
//  retcode = FALSE;
//
//  // OK
//  retcode = TRUE;
//
//Exit:
//  DBOUT1 (DBG_LF_EXIT, "ohcitest: XXX() exit, retcode = %s\n", BoolToStr(retcode));
//  return retcode;
//
//XXXError:
//  printf ("ERROR: XXX\n");
//  printf ("Last error = %08x = %s\n", errno, strerror(errno));
//  goto Exit;
//
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugFlags
//}

////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//// XXX()
////
////        XXX
////
//// Entry:
////
////        XXX
////
//// Exit:
////
////        TRUE    success
////        FALSE   error
////
////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//BOOLEAN XXX (void)
//{
////    #undef DEBUGFLAGS
////    #define DEBUGFLAGS DebugThisRoutine
//
//  BOOLEAN retcode;
//
//  DBOUT (DBG_LF_ENTRY, "ohcitest: XXX()\n");
//
//  // Assume failure
//  retcode = FALSE;
//
//  // OK
//  retcode = TRUE;
//
//Exit:
//  DBOUT1 (DBG_LF_EXIT, "ohcitest: XXX() exit, retcode = %s\n", BoolToStr(retcode));
//  return retcode;
//
//XXXError:
//  printf ("ERROR: XXX\n");
//  printf ("Last error = %08x = %s\n", errno, strerror(errno));
//  goto Exit;
//
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugFlags
//}

////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//// XXX()
////
////        XXX
////
//// Entry:
////
////        XXX
////
//// Exit:
////
////        TRUE    success
////        FALSE   error
////
////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//BOOLEAN XXX (void)
//{
////    #undef DEBUGFLAGS
////    #define DEBUGFLAGS DebugThisRoutine
//
//  BOOLEAN retcode;
//
//  DBOUT (DBG_LF_ENTRY, "ohcitest: XXX()\n");
//
//  // Assume failure
//  retcode = FALSE;
//
//  // OK
//  retcode = TRUE;
//
//Exit:
//  DBOUT1 (DBG_LF_EXIT, "ohcitest: XXX() exit, retcode = %s\n", BoolToStr(retcode));
//  return retcode;
//
//XXXError:
//  printf ("ERROR: XXX\n");
//  printf ("Last error = %08x = %s\n", errno, strerror(errno));
//  goto Exit;
//
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugFlags
//}

////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//// XXX()
////
////        XXX
////
//// Entry:
////
////        XXX
////
//// Exit:
////
////        TRUE    success
////        FALSE   error
////
////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//BOOLEAN XXX (void)
//{
////    #undef DEBUGFLAGS
////    #define DEBUGFLAGS DebugThisRoutine
//
//  BOOLEAN retcode;
//
//  DBOUT (DBG_LF_ENTRY, "ohcitest: XXX()\n");
//
//  // Assume failure
//  retcode = FALSE;
//
//  // OK
//  retcode = TRUE;
//
//Exit:
//  DBOUT1 (DBG_LF_EXIT, "ohcitest: XXX() exit, retcode = %s\n", BoolToStr(retcode));
//  return retcode;
//
//XXXError:
//  printf ("ERROR: XXX\n");
//  printf ("Last error = %08x = %s\n", errno, strerror(errno));
//  goto Exit;
//
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugFlags
//}

////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//// XXX()
////
////        XXX
////
//// Entry:
////
////        XXX
////
//// Exit:
////
////        TRUE    success
////        FALSE   error
////
////~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//BOOLEAN XXX (void)
//{
////    #undef DEBUGFLAGS
////    #define DEBUGFLAGS DebugThisRoutine
//
//  BOOLEAN retcode;
//
//  DBOUT (DBG_LF_ENTRY, "ohcitest: XXX()\n");
//
//  // Assume failure
//  retcode = FALSE;
//
//  // OK
//  retcode = TRUE;
//
//Exit:
//  DBOUT1 (DBG_LF_EXIT, "ohcitest: XXX() exit, retcode = %s\n", BoolToStr(retcode));
//  return retcode;
//
//XXXError:
//  printf ("ERROR: XXX\n");
//  printf ("Last error = %08x = %s\n", errno, strerror(errno));
//  goto Exit;
//
//  #undef DEBUGFLAGS
//  #define DEBUGFLAGS DebugFlags
//}

