package tecgraf.objpool;

import java.util.*;

/**
 *	This class represents a manager of a pool of objects of a given class
 *	present in an ObjectPool. It is responsible for allocating new objects
 *	when needed, clearing free objects and retrieving available objects
 *	when possible.
 *
 *	<P> The abstract method <code>newInstance</code> must be overriden by
 *	subclasses to allocate new instances of the class managed in the Java
 *	heap.
 *
 *	@author Alexandre Sieira Vilar (asv@tecgraf.puc-rio.br)
 *	@version $Revision: 1.5 $
 */
public abstract class ClassManager
{
	/** The list of available instances. */
	List avail;

	/** If true, debugging output and runtime checks are made. */
	boolean debug=false;

	/** The number of objects requested. */
	long requests=0;

	/** The number of objects allocated from the heap with 'new'. */
	long misses=0;

	/** 
	 *	Creates a new pool of objects of a given class.
	 */
	public ClassManager()
	{
		avail = new ArrayList(64);
	}

	/**
	 *	Retrieves a previously allocated object of the given
	 *	type if available. If none are available, allocates a
	 *	new one and returns it.
	 *	
	 *	@return an usable object
	 */
	public synchronized Object allocate()
	{
		int size = avail.size();

		requests++;

		if (size > 0) {
			return avail.remove(size-1);
		} else {
			misses++;
			return newInstance();
		}
	}

	/**
	 *	Adds the given object to the available list, indicating
	 *	that it is no longer being used.
	 *
	 *	@param obj the object to add to the available list
	 */
	public synchronized void release( Object obj )
	{
		if (debug) {
			for (int i=0; i<avail.size(); i++) {
				if (avail.get(i) == obj) {
					throw new IllegalArgumentException("object already on free list");
				}
			}
		}

		avail.add(obj);
	}

	/**
	 *	Returns the number of available objects in the class manager.
	 *	If more than this number of objects is allocated then new
	 *	instances will be created.
	 *
	 *	@return the number of available objects in the class manager
	 */
	public synchronized int getSize()
	{
		return avail.size();
	}

	/**
	 *	Releases all objects on the available list, to allow the system
	 *	to collect them.
	 */
	public synchronized void clear()
	{
		avail.clear();
	}

	/**
	 *	The finalizer method. Calls the superclass finalizer and
	 *	sets all references to <code>null</code>.
	 */
	public void finalize()
	throws Throwable
	{
		avail.clear();
		avail = null;
		
		if (debug) System.err.println(this);

		super.finalize();
	}

	/**
	 *	This method must be overriden to allocate a new instance of
	 *	the class managed.
	 *
	 *	@return a new instance of the class managed
	 */
	protected abstract Object newInstance();

	/**
	 *	Returns a string representation of the status of the object.
	 */
	public String toString()
	{
		double hitrate = (double) (requests - misses)  * 100.0 / (double) requests;

		return "ClassManager ("
		       + avail.size() + " available objects, "
		       + requests + " requests, " 
		       + hitrate + "% hit rate"
		       + ")";
	}
}
