package tecgraf.objpool;

import java.util.*;

/**
 *	The object pool is a repository for objects of a commonly instantiated
 *	class. It allows for the reuse of previously allocated objects,
 *	reducing the impact of garbage collection.
 *
 *	@author Alexandre Sieira Vilar (asv@tecgraf.puc-rio.br)
 *	@version $Revision: 1.6 $
 */
public final class ObjectPool
{
	/** This variable maps class names to a list of available objects.
	    All elements of this map are of the type ClassManager. */
	static Map typeMap = Collections.synchronizedMap(new HashMap());

	/**
	 *	Sets the class manager of a given class to a user-defined
	 *	object.
	 *
	 *	@param className the name of the class pooled
	 *	@param cm the class manager
	 */
	public static void setClassManager( String className,
	                                    ClassManager cm )
	{
		typeMap.put(className, cm);
	}

	/**
	 *	Returns the ClassManager associated with the given className.
	 *	If none is available, an instance of ReflectionClassManager
	 *	is created for the default constructor (no parameters).
	 *
	 *	@param className the name of class desired
	 *	@return the object pool class manager
	 */
	private static ClassManager getClassManager ( String className )

	{
		ClassManager cm = (ClassManager) typeMap.get(className);

		// If no object manager is installed, try default one.
		if (cm == null) {
			try {
				cm = new ReflectionClassManager(className);
			} catch (Exception e) {
				e.printStackTrace();
				throw new NoClassManagerException(className);
			}
		
			typeMap.put(className, cm);
		}

		return cm;
	}

	/**
	 *	Returns a useable instance of the fully qualified class name
	 *	given. The instance returned will contain garbage in its
	 *	variables, since it might have been used previously.
	 *
	 *	@param className the fully qualified name of the class desired, such as "java.awt.geom.Point2D$Double"
	 *	@return an useable instance of the class requested
	 */
	public static Object allocate(String className)
	{
		return getClassManager(className).allocate();
	}

	/**
	 *	Releases an object which is no longer in use.
	 *
	 *	@param obj the object to release
	 */
	public static void release( Object obj )
	{
		getClassManager( obj.getClass().getName() ).release(obj);
	}

	/**
	 *	Returns the number of available objects of the given class name
	 *	currently in the pool. If more than the given number of objects
	 *	is allocated, the pool will automatically create as many new
	 *	instances as needed.
	 *
	 *	@param className the fully qualified name of class desired
	 *	@return the non-negative number of available instances
	 */
	public static int getAvailableCount( String className )
	{
		return getClassManager(className).getSize();
	}

	/**
	 *	Clear the pool of available objects belonging to the given
	 *	class name, allowing them to be collected.
	 *
	 *	@param className the fully qualified name of class desired
	 */
	 public static void clear( String className )
	 {
		 getClassManager(className).clear();
	 }
	 

	/**
	 *	Prints debugging information to <code>System.err</code>.
	 */
	public static void printDebugInfo()
	{
		Iterator it = typeMap.entrySet().iterator();

		while (it.hasNext()) {
			Map.Entry entry = (Map.Entry) it.next();

			System.err.println(entry.getKey());
			System.err.print("   ");
			System.err.println(entry.getValue());
		}
	}
}


