package tecgraf.objpool;

import java.lang.reflect.*;

/**
 *	This class is a subclass of ClassManager that uses the Java Reflection
 *	API to instance new objects.
 *
 *	@author Alexandre Sieira Vilar (asv@tecgraf.puc-rio.br)
 *	@version $Revision: 1.6 $
 */
public final class ReflectionClassManager
extends ClassManager
{
	/** The name of the class managed. */
	String className;
	
	/** The default constructor of the class managed. */
	Constructor constructor;

	/** Argument list for the constructor. */
	Object[] args;

	/**
	 *	Creates a new manager for the given class using a constructor
	 *	with no parameters.
	 *
	 *	@param className the name of class managed by this object
	 *	@throws ClassNotFoundException if the given class is not found
	 *	@throws NoSuchMethodException if the class does not have a default constructor
	 */
	public ReflectionClassManager( String className )
	throws ClassNotFoundException, NoSuchMethodException
	{
		super();

		this.className = className;
		this.args = null;
		Class c = Class.forName(className);
		constructor = c.getConstructor(null);
	}

	/**
	 *	Creates a new manager for the given class using a constructor
	 *	with the given list of parameters.
	 *
	 *	@param className the name of class managed by this object
	 *	@param argsClasses the classes of the arguments that must be provided to the class constructor
	 *	@param args the values of the arguments passed to the class constructor
	 *	@throws ClassNotFoundException if the given class is not found
	 *	@throws NoSuchMethodException if the class does not have a default constructor
	 */
	public ReflectionClassManager( String className, Class[] argsClasses,
	                                  Object[] args )
	throws ClassNotFoundException, NoSuchMethodException
	{
		super();

		this.className = className;
		this.args = args;
		Class c = Class.forName(className);
		constructor = c.getConstructor(argsClasses);
	}

	/**
	 *	Allocates a new instance of the class managed by calling its
	 *	constructor. If an exception occurs in the instantiation,
	 *	a RuntimeException is thrown to abort the program execution.
	 *
	 *	@return a newly allocated instance of the managed class
	 */
	protected Object newInstance()
	{
		try {
			return constructor.newInstance(args);
		} catch (Exception e) {
			e.printStackTrace();
			throw new RuntimeException("unable to instance class " + className);
		}
	}
}
